#include "gtopology.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <math.h>
#include "galias.h"
#include "ordering.h"
#include "gmatrix.h"
#include <time.h>

gtopology::gtopology ()
{
  //  number of nodes
  nn=0;
  //  number of elements
  ne=0;
  //  number of layered nodes
  nln=0;
  //  number of degrees of freedom in the problem
  ndof=0;
  //  number of internal DOFs
  nidof=0;
  //  number of boundary/interface DOFs
  nbdof=0;
  //  number of unknowns in saddle point problem
  nsad=0;
  //  number of nodes which are switched on
  //  it is used in problems with changing number of nodes and elements
  nnso=0;
  //  number of elements which are switched on
  //  it is used in problems with changing number of nodes and elements
  neso=0;
  //  number of subdomains/aggregates
  ns=0;
  //  number of phases
  nph=0;
  //  indicator of sequential topology reading
  //  sequential topology is not read
  rst=0;
  
  
  //  general nodes
  gnodes=NULL;
  //  general elements
  gelements=NULL;
  //  layered general nodes
  lgnodes=NULL;
  
  domsizes = NULL;
  
  nadjelnod=NULL;  adjelnod=NULL;
  nadjelel=NULL;  adjelel=NULL;
  bckcn=NULL;  unln=NULL;  unnl=NULL;
  cngtopcorr=NULL;

  gphases=NULL;
  
  
  
  //  DOF control
  //  default value - DOFs are independent on time, they do not change
  dofcontr=0;
  //  type of generator of code numbers
  //  default value - the classical generation of code numbers
  cngen=1;
  //  reordered node numbers
  ordering=NULL;
  //  type of renumbering
  nodren = no_renumbering;
  

  //
  //  TOOLS FOR PROBLEMS WITH CHANGING NUMBER OF ELEMENTS
  //
  //  state of code numbers
  cnstate=0;
  //  general functions
  gf=NULL;
  //  list of nodes which are switched on
  lnso=NULL;
  //  list of elements which are switched on
  leso=NULL;
  //  list of nodes used for storage of previous load case
  nm = NULL;
  
  
  
  //  number of general edges
  nged=0;
  //  number of edge series
  nser=0;
  //  series of edges
  edgeser=NULL;
  
  //  general edges
  gedges=NULL;
  
  //  number of end nodes
  nen=0;
  //  end nodes
  endnodes=NULL;
  
  //  sequential topology
  stop=NULL;
}

gtopology::~gtopology ()
{
  long i;
  delete [] gnodes;  
  delete [] gelements;
  delete [] lgnodes;
  
  delete [] domsizes;
  
  delete [] nadjelnod;  
  if (adjelnod)
  {
    for (i=0;i<nn;i++)
      delete [] adjelnod[i];
  }
  delete [] adjelnod;
  
  delete [] nadjelel;
  if (adjelel)
  {
    for (i=0;i<ne;i++)
      delete [] adjelel[i];
  }
  delete [] adjelel;
  
  delete [] bckcn;
  delete [] unln;  
  delete [] unnl;
  delete [] cngtopcorr;

  delete [] gphases;
  delete [] ordering;

  //
  //  TOOLS FOR PROBLEMS WITH CHANGING NUMBER OF ELEMENTS
  //
  delete [] gf;
  delete [] lnso;
  delete [] leso;
  delete [] nm;
  
  delete [] edgeser;
  delete [] gedges;
  
  delete [] endnodes;
  
  //  sequential toplogy
  delete [] stop;
}

/**
   function initiates class gtopology
   
   @param icn - array containing code numbers
   @param nc - array containing numbers of components
   @param m - number of generalized elements
   
   24.7.2002
*/
void gtopology::initiate (long **icn,long *nc,long m)
{
  long i;
  ne=m;
  alloc_elements (ne);
  for (i=0;i<ne;i++){
    gelements[i].initiate (icn[i],nc[i]);
  }
}

/**
   function allocates nodes
   
   @param m - number of nodes
*/
void gtopology::alloc_nodes (long m)
{
  nn=m;
  if (m<0){
    fprintf (stderr,"\n\n negative number of nodes in function alloc_nodes (file %s, line %d).\n",__FILE__,__LINE__);
  }
  gnodes = new gnode [nn];
}

/**
   function allocates elements
   
   @param m - number of elements
*/
void gtopology::alloc_elements (long m)
{
  ne=m;
  if (m<0){
    fprintf (stderr,"\n\n negative number of elements in function alloc_nodes (file %s, line %d).\n",__FILE__,__LINE__);
  }
  gelements = new gelement [ne];
}

/**
   function allocates layered nodes
   
   @param m - number of layered nodes
*/
void gtopology::alloc_lnodes (long m)
{
  nln=m;
  if (m<0){
    fprintf (stderr,"\n\n negative number of nodes in function alloc_nodes (file %s, line %d).\n",__FILE__,__LINE__);
  }
  lgnodes = new lgnode [nln];
}



/**
   function allocates phases
   
   @param m - number of nodes
   6.4.2005 TKr
*/
void gtopology::alloc_phases (long m)
{
  nph=m;
  if (m<0){
    fprintf (stderr,"\n\n negative number of phases in function alloc_phases (file %s, line %d).\n",__FILE__,__LINE__);
  }
  gphases = new gphase [nph];
}

/**
   function returns number of degrees of freedom of required node
   
   @param nid - node id (number of node)
*/
long gtopology::give_ndofn (long nid)
{
  long ndofn;
  ndofn = gnodes[nid].give_ndofn ();
  return ndofn;
}

/**
   function returns number of multipliers defined between two layers in one node
   
   @param lnid - layerd node id (number of layered node)
*/
long gtopology::give_nmult (long lnid)
{
  return lgnodes[lnid].nmult;
}

/**
   function returns number of degrees of freedom of required element
   
   @param eid - element id (number of element)
 */
long gtopology::give_ndofe (long eid)
{
  return gelements[eid].give_ndofe ();
}

/**
   function returns number of all degrees of freedom of required element
   it means with Lagrange multipliers defined on element
   
   @param eid - element id (number of element)
*/
long gtopology::give_gndofe (long eid)
{
  long ndofe,nne;
  nne = gelements[eid].give_nne ();
  ndofe = gelements[eid].give_ndofe ();
  ndofe+= nne*2*gelements[eid].give_nmult ();
  return ndofe;
}

/**
   function returns number of nodes on selected edge
   
   @param edid - edge id
*/
long gtopology::give_nnedge (long edid)
{
  return gedges[edid].nn;
}

/**
   function returns number of nodes on required element
   
   @param eid - element id (number of element)
*/
long gtopology::give_nne (long eid)
{
  return gelements[eid].give_nne ();
}

/**
   function returns indicator of code numbers defined on element
   
   @param eid - element id (number of element)
*/
long gtopology::give_cne (long eid)
{
  return gelements[eid].give_cne ();
}

/**
   function returns required degree of freedom of required node
   
   @param nid - node id
   @param n - number of degree of freedom
*/
long gtopology::give_dof (long nid,long n)
{
  return gnodes[nid].give_dof (n);
}

/**
   function sets one code number on node
   
   @param nid - node id
   @param n - number of degree of freedom
   @param num - value of code number
*/
void gtopology::save_dof (long nid,long n,long num)
{
  gnodes[nid].save_dof (n,num);
}

/**
   function returns numbers of nodes on element
   
   @param eid - element id
   @param nod - array containing node numbers
*/
void gtopology::give_nodes (long eid,ivector &nod)
{
  gelements[eid].give_nodes (nod);
}


/**
   function returns node coordinates of the appropriate element
   
   @param x - vectors containing node coordinates
   @param eid - element id
   
   16.7.2002
*/
void gtopology::give_node_coord1d (vector &x,long eid)
{
  long i,nne;
  nne = gelements[eid].give_nne ();
  if (x.n!=nne){
    fprintf (stderr,"\n\n wrong size of array is assumed in function give_node_coord1d (file %s, line %d).\n",__FILE__,__LINE__);
  }
  ivector nodes(nne);
  gelements[eid].give_nodes (nodes);
  
  for (i=0;i<nne;i++){
    x[i]=gnodes[nodes[i]].x;
  }
}


/**
   function returns node coordinates of the appropriate element
   
   @param x,y - vectors containing node coordinates
   @param eid - element id
   
   16.7.2002
*/
void gtopology::give_node_coord2d (vector &x,vector &y,long eid)
{
  long i,nne;
  nne = gelements[eid].give_nne ();
  if (x.n!=nne || y.n!=nne){
    fprintf (stderr,"\n\n wrong size of array is assumed in function gtopology::give_node_coord2d (file %s, line %d).\n",__FILE__,__LINE__);
  }
  ivector nodes(nne);
  gelements[eid].give_nodes (nodes);
  
  for (i=0;i<nne;i++){
    x[i]=gnodes[nodes[i]].x;
    y[i]=gnodes[nodes[i]].y;
  }
}

/**
   function returns node coordinates of the appropriate element
   
   @param x,z - vectors containing coordinates
   @param eid - element id
   
   16.7.2002
*/
void gtopology::give_node_coord2dxz (vector &x,vector &z,long eid)
{
  long i,nne;
  nne = gelements[eid].give_nne ();
  if (x.n!=nne || z.n!=nne){
    fprintf (stderr,"\n\n wrong size of array is assumed in function gtopology::give_node_coord2d (file %s, line %d).\n",__FILE__,__LINE__);
  }
  ivector nodes(nne);
  gelements[eid].give_nodes (nodes);
  
  for (i=0;i<nne;i++){
    x[i]=gnodes[nodes[i]].x;
    z[i]=gnodes[nodes[i]].z;
  }
}

/**
   function returns node coordinates of the appropriate element
   
   @param x,y,z - vectors containing coordinates
   @param eid - element id
   
   16.7.2002
*/
void gtopology::give_node_coord3d (vector &x,vector &y,vector &z,long eid)
{
  long i,nne;
  nne = gelements[eid].give_nne ();
  if (x.n!=nne || y.n!=nne || z.n!=nne){
    fprintf (stderr,"\n\n wrong size of array is assumed in function gtopology::give_node_coord2d (file %s, line %d).\n",__FILE__,__LINE__);
  }
  ivector nodes(nne);
  gelements[eid].give_nodes (nodes);
  
  for (i=0;i<nne;i++){
    x[i]=gnodes[nodes[i]].x;
    y[i]=gnodes[nodes[i]].y;
    z[i]=gnodes[nodes[i]].z;
  }
}

/**
   function extracts all code numbers of actual element
   
   @param eid - element id
   @param cn - array containing all code numbers on element
   
   16.7.2002
*/
void gtopology::give_code_numbers (long eid,long *cn)
{
  long i,j,k,nne,ndofn;
  
  if (cnstate==0){
    //fprintf (stderr,"\n\n code numbers are required but not available (file %s, line %d)\n",__FILE__,__LINE__);
  }
  
  if (gelements[eid].cne==1){
    //  code numbers are stored on elements
    for (i=0;i<give_ndofe (eid);i++){
      cn[i]=gelements[eid].cn[i];
    }
  }
  else{
    //  code numbers are obtained from nodes
    k=0;
    nne=give_nne (eid);
    ivector nod(nne);
    give_nodes (eid,nod);
    for (i=0;i<nne;i++){
      ndofn=give_ndofn (nod[i]);
      for (j=0;j<ndofn;j++){
	cn[k]=give_dof (nod[i],j);
	k++;
      }
    }
  }
  
}

/**
   function extracts all code numbers of actual element
   including code numbers of Lagrange multipliers
   
   @param eid - element id
   @param cn - array containing all code numbers on element
   
   16.7.2002
*/
void gtopology::give_gcode_numbers (long eid,long *cn)
{
  long i,j,k,nne,ndofe,ndofn,lnid,lid,nmult;
  ivector nod;

  if (cnstate==0){
    //fprintf (stderr,"\n\n code numbers are required but not available (file %s, line %d)\n",__FILE__,__LINE__);
  }

  ndofe = give_ndofe (eid);
  
  if (gelements[eid].cne==1){
    //  code numbers are stored on elements
    for (i=0;i<ndofe;i++){
      cn[i]=gelements[eid].cn[i];
    }
  }
  else{
    //  code numbers are obtained from nodes
    k=0;
    nne=give_nne (eid);
    allocv (nne,nod);
    give_nodes (eid,nod);
    for (i=0;i<nne;i++){
      ndofn=give_ndofn (nod[i]);
      for (j=0;j<ndofn;j++){
	cn[k]=give_dof (nod[i],j);
	k++;
      }
    }
    destrv (nod);
  }
  
  if (gelements[eid].nmult>0){
    k=ndofe;
    nne=give_nne (eid);
    allocv (nne,nod);
    give_nodes (eid,nod);
    for (i=0;i<nne;i++){
      lnid=unln[nod[i]];
      lid=unnl[nod[i]];
      nmult=give_nmult (lnid);
      give_mult_code_numbers (lnid,lid,cn+k);
      k+=nmult;
    }
    for (i=0;i<nne;i++){
      lnid=unln[nod[i]];
      lid=unnl[nod[i]]+1;
      nmult=give_nmult (lnid);
      give_mult_code_numbers (lnid,lid,cn+k);
      k+=nmult;
    }
    destrv (nod);
  }
  
}

/**
   function extracts code numbers of actual node
   
   @param nid - node id
   @param cn - array containing code numbers on element
   
   7.12.2002
*/
void gtopology::give_node_code_numbers (long nid,long *cn)
{
  long i,ndofn;
  
  if (cnstate==0){
    fprintf (stderr,"\n\n code numbers are required but not available (file %s, line %d)\n",__FILE__,__LINE__);
  }

  ndofn = give_ndofn (nid);
  for (i=0;i<ndofn;i++){
    cn[i]=gnodes[nid].cn[i];
  }
}

/**
   function extracts code numbers of multipliers of actual node
   
   @param nid - node id
   @param lid - layer id
   @param cn - array containing code numbers on element
   
   7.12.2002
*/
void gtopology::give_mult_code_numbers (long nid,long lid,long *cn)
{
  long i,nmult;
  
  if (cnstate==0){
    fprintf (stderr,"\n\n code numbers are required but not available (file %s, line %d)\n",__FILE__,__LINE__);
  }

  nmult = give_nmult (nid);
  for (i=0;i<nmult;i++){
    cn[i]=lgnodes[nid].cn[lid][i];
  }
}

/**
   function assembles code numbers of nodes and multipliers defined on selected end node
   
   @param enid - end node id
   @param ncn1 - array containing code numbers of the first node (on one side of edge)
   @param ncn2 - array containing code numbers of the second node (on the other side of edge)
   @param mcn - array containing code numbers of multipliers
   
   JK, 6.8.2008
*/
void gtopology::give_endnode_code_numbers (long enid,long *ncn1,long *ncn2,long *mcn)
{
  long ndofn,*nid;
  
  nid = new long [2];

  //  node number
  endnodes[enid].give_node_numbers (nid);
  
  //  number of DOFs on node
  ndofn = give_ndofn (nid[0]);
  
  //  code numbers of nodes
  give_node_code_numbers (nid[0],ncn1);
  give_node_code_numbers (nid[1],ncn2);
  
  //  code numbers of Lagrange multipliers
  endnodes[enid].give_mult_code_numbers (mcn);
  
  delete [] nid;
}

/**
   function assembles code numbers of nodes and multipliers defined on selected edge
   
   @param edid - edge id
   @param fln - first or last nodes indicator
          fln=1 - first nodes are used
	  fln=2 - last nodes are used
   @param ncn1 - array containing code numbers of the first node (on one side of edge)
   @param ncn2 - array containing code numbers of the second node (on the other isde of edge)
   @param mcn - array containing code numbers of multipliers
   
   JK, 6.8.2008
*/
void gtopology::give_edge_code_numbers (long edid,long fln,long *ncn1,long *ncn2,long *mcn)
{
  long ndofn,*nid;
  
  nid = new long [2];

  if (fln==1){
    gedges[edid].give_first_node_numbers (nid);
  }
  if (fln==2){
    gedges[edid].give_last_node_numbers (nid);
  }
  
  //  number of DOFs on node
  ndofn = give_ndofn (nid[0]);
  
  //  code numbers of nodes
  give_node_code_numbers (nid[0],ncn1);
  give_node_code_numbers (nid[1],ncn2);
  
  gedges[edid].give_mult_code_numbers (fln,mcn);

  delete [] nid;
}



/**
   function returns type of finite element
   
   @param eid - element id
   
   JK, 13.4.2007
*/
gelemtype gtopology::give_elem_type (long eid)
{
  return gelements[eid].get;
}

/**
   function selects end nodes on required 1D element
   
   @param eid - element id
   @param nodes - end node numbers
   
   JK, 12.10.2008
*/
void gtopology::give_end_nodes (long eid,long *nodes)
{
  long nne;
  gelemtype get;
  ivector nod;
  long *enod;
  
  enod=new long [2];

  //  number nodes on element
  nne = give_nne (eid);
  //  allocation of the array of node numbers
  allocv (nne,nod);
  //  node numbers
  give_nodes (eid,nod);
  //  type of finite element
  get = give_elem_type (eid);

  switch (get){
  case linbar:{
    linbar_endpoints (enod);
    nodes[0]=nod[enod[0]];
    nodes[1]=nod[enod[1]];
    break;
  }
  case quadbar:{
    quadbar_endpoints (enod);
    nodes[0]=nod[enod[0]];
    nodes[1]=nod[enod[1]];
    break;
  }

  default:{
    print_err("unknown element type is required", __FILE__, __LINE__, __func__);
  }
  }
  
  delete [] enod;
  destrv (nod);
}


/**
   function selects nodes on required edge
   
   @param eid - element id
   @param edid - edge id
   @param nodes - edge node numbers
   
   JK, 13.4.2007
*/
void gtopology::give_edge_nodes (long eid,long edid,long *nodes)
{
  long i,nne,nned;
  gelemtype get;
  ivector nod;
  long *edgenod;
  
  //  number nodes on element
  nne = give_nne (eid);
  //  allocation of the array of node numbers
  allocv (nne,nod);
  //  node numbers
  give_nodes (eid,nod);
  //  type of finite element
  get = give_elem_type (eid);
  // number of nodes on element edge
  nned = give_nned (eid);
  
  edgenod = new long[nned];
  switch (get){
  case lintriag:{
    lintriangle_edgnod(edgenod,edid);
    for(i = 0; i < nned; i++){
      nodes[i]=nod[edgenod[i]];
    }
    break;
  }
  case quadtriag:{
    quadtriangle_edgnod(edgenod,edid);
    for(i = 0; i < nned; i++){
      nodes[i]=nod[edgenod[i]];
    }
    break;
  }
  case linquad:{
    linquadrilat_edgnod (edgenod,edid);
    for(i = 0; i < nned; i++){
      nodes[i]=nod[edgenod[i]];
    }
    break;
  }
  case quadquad:{
    quadquadrilat_edgnod (edgenod,edid);
    for(i = 0; i < nned; i++){
      nodes[i]=nod[edgenod[i]];
    }
    break;
  }
  case lintetra:{
    lintetrahedral_edgnod(edgenod,edid);
    for(i = 0; i < nned; i++){
      nodes[i]=nod[edgenod[i]];
    }
    break;
  }
  case quadtetra:{
    quadtetrahedral_edgnod (edgenod,edid);
    for(i = 0; i < nned; i++){
      nodes[i]=nod[edgenod[i]];
    }
    break;
  }
  case linhexa:{
    linhexahedral_edgnod(edgenod,edid);
    for(i = 0; i < nned; i++){
      nodes[i]=nod[edgenod[i]];
    }
    break;
  }
  case quadhexa:{
    quadhexahedral_edgnod (edgenod,edid);
    for(i = 0; i < nned; i++){
      nodes[i]=nod[edgenod[i]];
    }
    break;
  }
  default:{
    print_err("unknown element type is required", __FILE__, __LINE__, __func__);
  }
  }
  
  destrv (nod);
  delete []edgenod;
}

/**
   function returns node numbers on required element surface
   
   @param eid - element id
   @param edid - edge id
   @param nodes - array containing edge nodes

   JB, 18.4.2007
*/
void gtopology::give_surf_nodes (long eid,long surfid,long *nodes)
{
  long i,nne,nnsurf;
  gelemtype get;
  ivector nod;
  long *surfnod;
  
  //  number nodes on element
  nne = give_nne (eid);
  //  allocation of the array of node numbers
  allocv (nne,nod);
  //  node numbers
  give_nodes (eid,nod);
  //  type of finite element
  get = give_elem_type (eid);
  //  number of nodes on surface
  nnsurf = give_nnsurf (eid);
  
  surfnod = new long[nnsurf];
  switch(get){
  case lintetra:{
    lintetrahedral_surfnod (surfnod,surfid);
    for(i = 0; i < nnsurf; i++){
      nodes[i]=nod[surfnod[i]];
    }
    break;
  }
  case quadtetra:{
    quadtetrahedral_surfnod (surfnod,surfid);
    for(i = 0; i < nnsurf; i++){
      nodes[i]=nod[surfnod[i]];
    }
    break;
  }
  case linhexa:{
    linhexahedral_surfnod (surfnod,surfid);
    for(i = 0; i < nnsurf; i++){
      nodes[i]=nod[surfnod[i]];
    }
    break;
  }
  case quadhexa:{
    quadhexahedral_surfnod (surfnod,surfid);
    for(i = 0; i < nnsurf; i++){
      nodes[i]=nod[surfnod[i]];
    }
    break;
  }
    
  default:{
    fprintf (stderr,"\n\n unknown element type is required in function give_surface_nodes (file %s, line %d).\n",__FILE__,__LINE__);
  }
  }
  destrv (nod);
  delete []surfnod;
}

/**
   function returns number of end nodes on the element
   
   @param eid - element id
   
   JK, 17.10.2008
*/
long gtopology::give_nen (long eid)
{
  long ned=0;
  gelemtype get;
  
  get = give_elem_type (eid);
  
  switch (get){
  case linbar:{     ned=2;    break;  }
  case quadbar:{    ned=2;    break;  }
  case lintriag:{   ned=0;    break;  }
  case quadtriag:{  ned=0;    break;  }
  case linquad:{    ned=0;    break;  }
  case quadquad:{   ned=0;    break;  }
  case lintetra:{   ned=0;    break;  }
  case quadtetra:{  ned=0;    break;  }
  case linhexa:{    ned=0;    break;  }
  case quadhexa:{   ned=0;    break;  }
  default:{
    print_err("unknown element type is required", __FILE__, __LINE__, __func__);
  }
  }
  return ned;
}

/**
   function returns number of edges of the element
   
   @param eid - element id
   
   JB, 18.4.2007
*/
long gtopology::give_ned (long eid)
{
  long ned=0;
  gelemtype get;
  
  get = give_elem_type (eid);
  
  switch (get){
  case linbar:{     ned=0;    break;  }
  case quadbar:{    ned=0;    break;  }
  case lintriag:{   ned=3;    break;  }
  case quadtriag:{  ned=3;    break;  }
  case linquad:{    ned=4;    break;  }
  case quadquad:{   ned=4;    break;  }
  case lintetra:{   ned=6;    break;  }
  case quadtetra:{  ned=6;    break;  }
  case linhexa:{    ned=12;   break;  }
  case quadhexa:{   ned=12;   break;  }
  default:{
    fprintf (stderr,"\n\n unknown element type is required in function give_ned (file %s, line %d).\n",__FILE__,__LINE__);
  }
  }
  return ned;
}

/**
   function returns number of nodes on one element edge
   
   @param eid - number of element
   
   JB, 18.4.2007
*/
long gtopology::give_nned (long eid)
{
  long nned=0;
  gelemtype get;

  get = give_elem_type (eid);
  
  switch (get){
  case linbar:{     nned=2;    break;  }
  case quadbar:{    nned=2;    break;  }
  case lintriag:{   nned=2;    break;  }
  case quadtriag:{  nned=3;    break;  }
  case linquad:{    nned=2;    break;  }
  case quadquad:{   nned=3;    break;  }
  case lintetra:{   nned=2;    break;  }
  case quadtetra:{  nned=3;    break;  }
  case linhexa:{    nned=2;    break;  }
  case quadhexa:{   nned=3;    break;  }
  default:{
    print_err("unknown element type is required", __FILE__, __LINE__, __func__);
  }
  }
  return nned;
}

/**
   function returns number of surfaces of the element
   
   @param eid - element id
   
   JB, 18.4.2004
*/
long gtopology::give_nsurf (long eid)
{
  long nsurf;
  gelemtype get;

  get = give_elem_type (eid);
  
  switch (get){
  case lintetra:{   nsurf=4;    break;  }
  case quadtetra:{  nsurf=4;    break;  }
  case linhexa:{    nsurf=6;    break;  }
  case quadhexa:{   nsurf=6;    break;  }
  default:{
    fprintf (stderr,"\n\n unknown element type is required in function give_nsurf (file %s, line %d).\n",__FILE__,__LINE__);
  }
  }
  return nsurf;
}

/**
   function returns number of nodes on element surface
   
   @param eid - element id
   
   JB, 18.4.2004
*/
long gtopology::give_nnsurf (long eid)
{
  long nnsurf;
  gelemtype get;
  
  get = give_elem_type (eid);
  
  switch (get){
  case lintetra:{         nnsurf=3;    break;  }
  case linhexa:{          nnsurf=4;    break;  }
  case quadtetra:{        nnsurf=6;    break;  }
  case quadhexa:{         nnsurf=8;    break;  }
  default:{
    fprintf (stderr,"\n\n unknown element type is required in function give_nnsurf (file %s, line %d).\n",__FILE__,__LINE__);
  }
  }
  return nnsurf;
}


/**
   function returns degree of polynomial approximation
   
   @param eid - number of element
   
   JB, 18.4.2007
*/
long gtopology::give_degree (long eid)
{
  long deg=0;
  gelemtype get;
  
  get = give_elem_type (eid);
  
  switch (get){
  case linbar:{     deg=1;    break;  }
  case quadbar:{    deg=2;    break;  }
  case lintriag:{   deg=1;    break;  }
  case quadtriag:{  deg=2;    break;  }
  case linquad:{    deg=1;    break;  }
  case quadquad:{   deg=2;    break;  }
  case lintetra:{   deg=1;    break;  }
  case quadtetra:{  deg=2;    break;  }
  case linhexa:{    deg=1;    break;  }
  case quadhexa:{   deg=2;    break;  }
  default:{
    fprintf (stderr,"\n\n unknown element type is required in function give_degree (file %s, line %d).\n",__FILE__,__LINE__);
  }
  }
  return deg;
}



/**
   function generates code numbers (ordering of unknowns and equations)
   function possibly reorders node numbers
   
   JK, 28.2.2007
*/
long gtopology::codenum_generation (FILE *out)
{
  long i,ndof;
  
  ordering = new long [nn];
  
  
  switch (nodren){
  case no_renumbering:{
    for (i=0;i<nn;i++){
      ordering[i]=i;
    }
    break;
  }
  case cuthill_mckee:
  case rev_cuthill_mckee:{
    cuthill_mckee_renumb (out);
    break;
  }
  case sloan:{
    sloan_renumb(out);
    break;
  }
  default:{
    print_err("unknown type of renumbering is required",__FILE__,__LINE__,__func__);
  }
  }
  
  //  generation of the code numbers
  if (cngen==1){
    //  the classical generation of code numbers
    ndof = gencodnum ();
  }
  if (cngen==2){
    //  internal unknowns are ordered first, interface unknowns are ordered last
    ndof = schur_ordering (stop->ltg);
  }
  if (cngen==3){
    //  primal unknowns are ordered first, dual unknowns are ordered last
    ndof = saddlepoint_ordering (stop->ns,stop->nnsd,stop->ltg);
  }
  
  delete [] ordering;
  ordering = NULL;
  
  return ndof;
}

/**
   function generates code numbers (ordering of unknowns and equations)
   
   JK, 25.6.2001
*/
long gtopology::gencodnum ()
{
  long i,j,k,ii,ndofn,*aux;
  
  //  searching of maximum code number
  ndof=0;
  for (i=0;i<nn;i++){
    ndofn=give_ndofn (i);
    for (j=0;j<ndofn;j++){
      k=give_dof (i,j);
      if (k>ndof)  ndof=k;
    }
  }
  ndof--;
  if (ndof<0)  ndof=0;
  aux = new long [ndof];
  for (i=0;i<ndof;i++){
    aux[i]=-1;
  }
  
  //  contributions from nodes
  ndof=1;
  for (i=0;i<nn;i++){
    ii=ordering[i];
    ndofn=give_ndofn (ii);
    for (j=0;j<ndofn;j++){
      k=give_dof (ii,j);
      if (k<0)  continue;
      if (k==0)  continue;
      if (k==1){
	gnodes[ii].cn[j]=ndof;  ndof++;
      }
      if (k>1){
	if (aux[k-2]==-1){
	  gnodes[ii].cn[j]=ndof;
	  aux[k-2]=ndof;
	  ndof++;
	}
	else{
	  gnodes[ii].cn[j]=aux[k-2];
	}
      }
    }
  }
  ndof--;
  
  delete [] aux;
  
  //  contributions from elements
  for (i=0;i<ne;i++){
    if (gelements[i].cne==1){
      for (j=0;j<give_ndofe (i);j++){
	if (gelements[i].cn[j]>ndof)  ndof=gelements[i].cn[j];
      }
    }
  }
  
  //  state of code numbers is changed
  cnstate=1;
  
  return ndof;
}

/**
   function generates code numbers (numbers of equations)
   with respect to Schur ordering
   
   internal nodes are ordered first, interface nodes are ordered at the end
   
   interface nodes are indicated by array ltg which has value 1
   
   JK, 22.10.2007
*/
long gtopology::schur_ordering (long **ltg)
{
  long i,j,k,ii,ndofn,*aux;
  
  //  searching of maximum code number
  ndof=0;
  for (i=0;i<nn;i++){
    ndofn=give_ndofn (i);
    for (j=0;j<ndofn;j++){
      k=give_dof (i,j);
      if (k>ndof)  ndof=k;
    }
  }
  ndof--;
  if (ndof<0)  ndof=0;
  aux = new long [ndof];
  for (i=0;i<ndof;i++){
    aux[i]=-1;
  }

  //  ordering of internal nodes
  ndof=1;
  for (i=0;i<nn;i++){
    if (ltg[0][i]==-1){
      ii=ordering[i];
      ndofn=give_ndofn (ii);
      for (j=0;j<ndofn;j++){
	k=give_dof (ii,j);
	if (k<0)  continue;
	if (k==0)  continue;
	if (k==1){
	  gnodes[ii].cn[j]=ndof;  ndof++;
	}
	if (k>1){
	  if (aux[k-2]==-1){
	    gnodes[ii].cn[j]=ndof;
	    aux[k-2]=ndof;
	    ndof++;
	  }
	  else{
	    gnodes[ii].cn[j]=aux[k-2];
	  }
	}
      }
    }
  }
  //  number of internal DOFs
  nidof=ndof-1;
  
  //  ordering of interface nodes
  for (i=0;i<nn;i++){
    if (ltg[0][i]==0){
      ii=ordering[i];
      ndofn=give_ndofn (ii);
      for (j=0;j<ndofn;j++){
	k=give_dof (ii,j);
	if (k<0)  continue;
	if (k==0)  continue;
	if (k==1){
	  gnodes[ii].cn[j]=ndof;  ndof++;
	}
	if (k>1){
	  if (aux[k-2]==-1){
	    gnodes[ii].cn[j]=ndof;
	    aux[k-2]=ndof;
	    ndof++;
	  }
	  else{
	    gnodes[ii].cn[j]=aux[k-2];
	  }
	}
      }
    }
  }
  ndof--;
  
  delete [] aux;
  
  //  contributions from elements
  for (i=0;i<ne;i++){
    if (gelements[i].cne==1){
      for (j=0;j<give_ndofe (i);j++){
	if (gelements[i].cn[j]>ndof)
	  ndof=gelements[i].cn[j];
      }
    }
  }
  
  //  number of boundary/interface DOFs
  nbdof=ndof-nidof;
  
  //  state of code numbers is changed
  cnstate=1;
  
  return ndof;
}



/**
   function generates code numbers (numbers of equations)
   with respect to saddle point problem
   
   internal nodes are ordered first, interface nodes are ordered at the end
   
   interface nodes are indicated by array ltg which has value 1
   
   @param ns - number of subdomains
   @param nnsd - number of nodes on subdomains
   @param ltg - local to global map
   
   JK, 13.12.2007
*/
long gtopology::saddlepoint_ordering (long ns,long *nnsd,long **ltg)
{
  long i,j,k,ii,jj,nid,ndofn,*aux;
  
  //  searching of maximum code number
  ndof=0;
  for (i=0;i<nn;i++){
    ndofn=give_ndofn (i);
    for (j=0;j<ndofn;j++){
      k=give_dof (i,j);
      if (k>ndof)  ndof=k;
    }
  }
  ndof--;
  if (ndof<0)  ndof=0;
  aux = new long [ndof];
  for (i=0;i<ndof;i++){
    aux[i]=-1;
  }

  //  ordering of internal nodes
  nid=0;
  ndof=1;
  for (jj=0;jj<ns;jj++){
    for (i=0;i<nnsd[jj];i++){
      if (ltg[jj][i]==-1){
	ii=ordering[nid];
	ndofn=give_ndofn (ii);
	for (j=0;j<ndofn;j++){
	  k=give_dof (ii,j);
	  if (k<0)  continue;
	  if (k==0)  continue;
	  if (k==1){
	    gnodes[ii].cn[j]=ndof;  ndof++;
	  }
	  if (k>1){
	    if (aux[k-2]==-1){
	      gnodes[ii].cn[j]=ndof;
	      aux[k-2]=ndof;
	      ndof++;
	    }
	    else{
	      gnodes[ii].cn[j]=aux[k-2];
	    }
	  }
	}
      }
      nid++;
    }
  }
  //  number of internal DOFs
  nidof=ndof-1;
  
  //  ordering of interface nodes
  nid=0;
  for (jj=0;jj<ns;jj++){
    for (i=0;i<nnsd[jj];i++){
      if (ltg[jj][i]!=-1){
	ii=ordering[nid];
	ndofn=give_ndofn (ii);
	for (j=0;j<ndofn;j++){
	  k=give_dof (ii,j);
	  if (k<0)  continue;
	  if (k==0)  continue;
	  if (k==1){
	    gnodes[ii].cn[j]=ndof;  ndof++;
	  }
	  if (k>1){
	    if (aux[k-2]==-1){
	      gnodes[ii].cn[j]=ndof;
	      aux[k-2]=ndof;
	      ndof++;
	    }
	    else{
	      gnodes[ii].cn[j]=aux[k-2];
	    }
	  }
	}
      }
      nid++;
    }
  }
  //ndof--;
  
  delete [] aux;
  
  //  contributions from elements
  for (i=0;i<ne;i++){
    if (gelements[i].cne==1){
      for (j=0;j<give_ndofe (i);j++){
	if (gelements[i].cn[j]>ndof)
	  ndof=gelements[i].cn[j];
      }
    }
  }
  
  //  number of boundary/interface DOFs
  nbdof=ndof-nidof;
  
  //  generation of code numbers of Lagrange multipliers
  //  nsad is the number of unknowns in saddle point problem
  //nsad = codenum_multip ();
  ndof = codenum_multip ();

  ndof--;

  //  state of code numbers is changed
  cnstate=1;
  
  fprintf (stdout,"\n number of DOFs            %ld",ndof);
  fprintf (stdout,"\n number of internal DOFs   %ld",nidof);
  fprintf (stdout,"\n number of interface DOFs  %ld",nbdof);
  fprintf (stdout,"\n number of unknowns in sadd. point. prob.  %ld",nsad);

  
  return ndof;
}
















/**
   function sorts nodes in the problems with changing number of nodes, DOFs and elements
   
   JK, 12.12.2006
*/
void gtopology::node_mark (long n)
{
  long i,j,k,ndofn,*aux;
  
  aux = new long [n];
  for (i=0;i<n;i++){
    aux[i]=0;
  }
  
  for (i=0;i<nn;i++){
    ndofn = give_ndofn (i);
    for (j=0;j<ndofn;j++){
      k=give_dof (i,j)-1;
      if (k>-1)
	aux[k]++;
    }
  }
  
  //nm = new long [nn];
  
  for (i=0;i<nn;i++){
    nm[i]=0;
    ndofn = give_ndofn (i);
    for (j=0;j<ndofn;j++){
      k=give_dof (i,j)-1;
      if (k>-1){
	if (aux[k]==-1)
	  nm[i]=0;
	if (aux[k]==1)
	  nm[i]=1;
	if (aux[k]>1){
	  nm[i]=1;
	  aux[k]=-1;
	}
      }
    }
  }
  
  delete [] aux;
}



/**
   function creates general topology from two identical topologies (meshes)
   with different unknowns defined in nodes
   
   function is also usefull for two different meshes used in accordance with
   Babuska-Brezzi condition; in such case, one mesh contains elements with
   quadratic approximation functions and the second mesh contains only elements
   with linear approximation functions; the first mesh contains therefore more
   nodes than the second mesh; user can add all nodes from the first mesh which
   are not defined in the second mesh to the list of nodes of the second mesh
   but they are not mentioned in element-node correspondence
   
   important notice: topology describing the mesh containing elements with
   quadratic approximation functions must be send to gt1 while topology
   describing the mesh containing elements with linear approximation functions
   must be send to gt2
   
   @param gt1 - pointer to the first topology
   @param gt2 - pointer to the second topology
   
   JK, 28.10.2002
*/
void gtopology::comptop (gtopology *gt1,gtopology *gt2)
{
  long i,j,k,l,nn1,nn2,ne1,ne2,ndofn1,nne;
  ivector nod;
  
  //  meshes checking - nodes
  nn1 = gt1->nn;
  nn2 = gt2->nn;
  if (nn1!=nn2){
    fprintf (stderr,"\n two topologies in function gtopology::linktop have different");
    fprintf (stderr,"\n numbers of nodes (file %s, line %d).\n",__FILE__,__LINE__);
    abort ();
  }
  
  nn = nn1;
  gnodes = new gnode [nn];
  
  //  assigment of coordinates to the unified mesh
  //  assigment of numbers of DOFs from the first mesh to the unified mesh
  for (i=0;i<nn;i++){
    ndofn1 = gt1->gnodes[i].ndofn;
    gnodes[i].ndofn = ndofn1;
    
    gnodes[i].x=gt1->gnodes[i].x;
    gnodes[i].y=gt1->gnodes[i].y;
    gnodes[i].z=gt1->gnodes[i].z;
  }
  
  //  meshes checking - elements
  ne1 = gt1->ne;
  ne2 = gt2->ne;
  if (ne1!=ne2){
    fprintf (stderr,"\n two topologies in function gtopology::linktop have different");
    fprintf (stderr,"\n numbers of elements (file %s, line %d).\n",__FILE__,__LINE__);
    abort ();
  }

  ne = ne1;
  gelements = new gelement [ne];
  
  //  assigment of numbers of DOFs from the second mesh to the unified mesh
  for (i=0;i<ne;i++){
    nne = gt2->give_nne (i);
    allocv (nne,nod);
    gt2->give_nodes (i,nod);
    for (j=0;j<nne;j++){
      if (gt2->gnodes[nod[j]].ai==1)  continue;
      gnodes[nod[j]].ndofn+=gt2->gnodes[nod[j]].ndofn;
      gt2->gnodes[nod[j]].ai=1;
    }
    destrv (nod);
  }
  
  for (i=0;i<nn;i++){
    gt2->gnodes[i].ai=0;
  }
  
  //  assigment of code numbers from the first mesh to the unified mesh
  for (i=0;i<nn;i++){
    gnodes[i].cn = new long [gnodes[i].ndofn];
    for (j=0;j<gt1->gnodes[i].ndofn;j++){
      gnodes[i].cn[j]=gt1->gnodes[i].cn[j];
    }
  }
  
  //  assigment of code numbers from the second mesh to the unified mesh
  for (i=0;i<ne;i++){
    nne = gt2->give_nne (i);
    allocv (nne,nod);
    gt2->give_nodes (i,nod);
    for (j=0;j<nne;j++){
      if (gt2->gnodes[nod[j]].ai==1)  continue;
      l=gt1->gnodes[nod[j]].ndofn;
      for (k=0;k<gt2->gnodes[nod[j]].ndofn;k++){
	gnodes[nod[j]].cn[l]=gt2->gnodes[nod[j]].cn[k];  l++;
      }
      gt2->gnodes[nod[j]].ai=1;
    }
    destrv (nod);
  }
  
  //  assigment of numbers of nodes on element to the unified mesh
  //  assigment of node numbers of elements to the unified mesh
  for (i=0;i<ne1;i++){
    gelements[i].ndofe = gt1->gelements[i].ndofe + gt2->gelements[i].ndofe;
    nne = gt1->gelements[i].nne;
    gelements[i].nne = nne;
    gelements[i].nodes = new long [nne];
    for (j=0;j<nne;j++){
      gelements[i].nodes[j] = gt1->gelements[i].nodes[j];
    }
  }

}

/**
   function distributes code numbers from unified topology to the
   local topologies
   
   function is also usefull for two different meshes used in accordance with
   Babuska-Brezzi condition; in such case, one mesh contains elements with
   quadratic approximation functions and the second mesh contains only elements
   with linear approximation functions; the first mesh contains therefore more
   nodes than the second mesh; user can add all nodes from the first mesh which
   are not defined in the second mesh to the list of nodes of the second mesh
   but they are not mentioned in element-node correspondence
   
   important notice: topology describing the mesh containing elements with
   quadratic approximation functions must be send to gt1 while topology
   describing the mesh containing elements with linear approximation functions
   must be send to gt2

   @param gt1 - pointer to first topology
   @param gt2 - pointer to second topology
   
   JK, 28.10.2002
*/
void gtopology::cndistr (gtopology *gt1,gtopology *gt2)
{
  long i,j,k,ndofn1,ndofn2;
  
  for (i=0;i<nn;i++){
    ndofn1=gt1->gnodes[i].ndofn;
    if (gnodes[i].ndofn>ndofn1){
      for (j=0;j<ndofn1;j++){
	gt1->gnodes[i].cn[j]=gnodes[i].cn[j];
      }
      k=ndofn1;
      ndofn2=gt2->gnodes[i].ndofn;
      for (j=0;j<ndofn2;j++){
	gt2->gnodes[i].cn[j]=gnodes[i].cn[k];
	k++;
      }
    }
    else{
      for (j=0;j<ndofn1;j++){
	gt1->gnodes[i].cn[j]=gnodes[i].cn[j];
      }
    }
  }
  
}

/**
   function asembles list of adjacent elements
   adjacent elements share a node, an edge or a surface
   
   @param out - output stream
*/
void gtopology::adjacelem (FILE *out)
{
  long i,j,k,min,ind,s,nne,last;
  long **aux;
  long *nodes;
  
  //  allocation of array of counts of adjacent elements to nodes
  nadjelnod = new long [nn];
  memset (nadjelnod,0,nn*sizeof(long));
  
  //  number of contributions
  for (i=0;i<ne;i++){
    nne = give_nne (i);
    nodes = gelements[i].nodes;
    for (j=0;j<nne;j++){
      nadjelnod[nodes[j]]++;
    }
  }
  
  //  allocation of array of numbers of adjacent elements to nodes
  adjelnod = new long* [nn];
  for (i=0;i<nn;i++){
    adjelnod[i] = new long [nadjelnod[i]];
    nadjelnod[i] = 0;
  }
  
  //  filling of array
  for (i=0;i<ne;i++){
    nne = give_nne (i);
    nodes = gelements[i].nodes;
    for (j=0;j<nne;j++){
      adjelnod[nodes[j]][nadjelnod[nodes[j]]++] = i;
    }
  }
  
  //  printing of array
  //fprintf (out,"\n\n\n LIST OF ELEMENTS ADJACENT TO NODES\n");
  //for (i=0;i<nn;i++){
  //fprintf (out,"\n node %4ld  ",i);
  //for (j=0;j<nadjelnod[i];j++){
  //fprintf (out,"  %ld",adjelnod[i][j]);
  //}
  //}
  //fprintf (out,"\n\n\n");
  
  
  //  allocation of array of counts of adjacent elements to elements
  nadjelel = new long [ne];
  memset (nadjelel,0,ne*sizeof(long));
  
  //  number of contributions
  for (i=0;i<ne;i++){
    nne=give_nne (i);
    nodes = gelements[i].nodes;
    for (j=0;j<nne;j++){
      nadjelel[i]+=nadjelnod[nodes[j]];
    }
  }
  
  //  allocation of the auxiliary array
  aux = new long* [ne];
  for (i=0;i<ne;i++){
    aux[i] = new long [nadjelel[i]];
  }
  
  
  //  filling of array
  memset (nadjelel,0,ne*sizeof(long));
  for (i=0;i<ne;i++){
    nne=give_nne (i);
    nodes = gelements[i].nodes;
    for (j=0;j<nne;j++){
      for (k=0;k<nadjelnod[nodes[j]];k++){
	aux[i][nadjelel[i]]=adjelnod[nodes[j]][k];
	nadjelel[i]++;
      }
    }
  }
  
  //  sorting of array
  for (i=0;i<ne;i++){
    last=-1;
    for (j=0;j<nadjelel[i];j++){
      min=LONG_MAX;
      for (k=j;k<nadjelel[i];k++){
	if (min>aux[i][k]){
	  min=aux[i][k];  ind=k;
	}
      }
      if (last==min){
	aux[i][ind]=aux[i][nadjelel[i]-1];
	nadjelel[i]--;  j--;
      }
      else{
	s=aux[i][j];
	aux[i][j]=min;
	aux[i][ind]=s;
	last=min;
      }
    }
  }
  
  //  allocation of array of numbers of adjacent elements to elements
  adjelel = new long* [ne];
  for (i=0;i<ne;i++){
    adjelel[i] = new long [nadjelel[i]];
    memset (adjelel[i],0,nadjelel[i]*sizeof(long));
  }
  
  for (i=0;i<ne;i++){
    for (j=0;j<nadjelel[i];j++){
      adjelel[i][j]=aux[i][j];
    }
  }
  
  //  printing of array
  //fprintf (out,"\n\n\n LIST OF ELEMENTS ADJACENT TO ELEMENTS\n");
  //for (i=0;i<ne;i++){
  //fprintf (out,"\n element %4ld  ",i);
  //for (j=0;j<nadjelel[i];j++){
  //fprintf (out,"  %ld",adjelel[i][j]);
  //}
  //}
  //fprintf (out,"\n\n\n");
  
  //  delete auxiliary array
  for (i=0;i<ne;i++){
    delete [] aux[i];
  }
  delete [] aux;
  
}




void gtopology::adjacnodes (FILE *out)
{
  long i,j,k,l,m,min,prev,nne;
  long **auxadjnodnod;
  long *nodes;
    
  //  allocation of array of counts of adjacent elements to nodes
  if (nadjelnod==NULL)
    nadjelnod = new long [nn];
  memset (nadjelnod,0,nn*sizeof(long));
  
  //  number of contributions
  for (i=0;i<ne;i++){
    nne = give_nne (i);
    nodes = gelements[i].nodes;
    for (j=0;j<nne;j++){
      nadjelnod[nodes[j]]++;
    }
  }
  
  //fprintf (out,"\n\n NADJELNOD");
  //for (i=0;i<nn;i++){
  //fprintf (out,"\n nadjelnod %5ld   %ld",i,nadjelnod[i]);
  //}
  
  
  //  allocation of array of numbers of adjacent elements to nodes
  if (adjelnod==NULL){
    adjelnod = new long* [nn];
    for (i=0;i<nn;i++){
      adjelnod[i] = new long [nadjelnod[i]];
    }
  }
  
  for (i=0;i<nn;i++){
    nadjelnod[i] = 0;
  }
  
  //  filling of array
  for (i=0;i<ne;i++){
    nne = give_nne (i);
    nodes = gelements[i].nodes;
    for (j=0;j<nne;j++){
      adjelnod[nodes[j]][nadjelnod[nodes[j]]++] = i;
    }
  }
  
  
  //fprintf (out,"\n\n ADJELNOD");
  //for (i=0;i<nn;i++){
  //  fprintf (out,"\n %ld    ",i);
  //  for (j=0;j<nadjelnod[i];j++){
  //    fprintf (out," %ld",adjelnod[i][j]);
  //  }
  //}
  
  
  //  array containing number of nodes adjacent to node before sorting
  nadjnodnod = new long [nn];
  for (i=0;i<nn;i++){
    nadjnodnod[i]=0;
    for (j=0;j<nadjelnod[i];j++){
      k=adjelnod[i][j];
      nadjnodnod[i]+=gelements[k].nne;
    }
  }
  
  //  array of adjacent nodes to each node before sorting
  auxadjnodnod = new long* [nn];
  for (i=0;i<nn;i++){
    auxadjnodnod[i] = new long [nadjnodnod[i]];
    nadjnodnod[i]=0;
  }
  
  //  filling of array
  for (i=0;i<nn;i++){
    for (j=0;j<nadjelnod[i];j++){
      l=adjelnod[i][j];
      nne=gelements[l].give_nne ();
      for (k=0;k<nne;k++){
	auxadjnodnod[i][nadjnodnod[i]]=gelements[l].nodes[k];
	nadjnodnod[i]++;
      }
    }
  }
  
  //fprintf (out,"\n\n auxadjnodnod");
  //for (i=0;i<nn;i++){
  //  fprintf (out,"\n %ld    ",i);
  //  for (j=0;j<nadjnodnod[i];j++){
  //    fprintf (out," %ld",auxadjnodnod[i][j]);
  //  }
  // }
  
  
  //fprintf (out,"\n\n\n nadjnodnod\n");
  //for (i=0;i<nn;i++)
  //  fprintf(out,"%ld %ld\n",i,nadjnodnod[i]);
  
  //  sorting
  for (i=0;i<nn;i++){
    prev=LONG_MAX;
    for (j=0;j<nadjnodnod[i];j++){
      min=LONG_MAX;
      for (k=j;k<nadjnodnod[i];k++){
	if (min>auxadjnodnod[i][k]){
	  min=auxadjnodnod[i][k];  l=k;
	}
      }
      if (min==prev){
	nadjnodnod[i]--;
	auxadjnodnod[i][l]=auxadjnodnod[i][nadjnodnod[i]];
	j--;
      }
      else{
	m=auxadjnodnod[i][j];
	auxadjnodnod[i][j]=min;
	auxadjnodnod[i][l]=m;
	prev=min;
      }
    }
  }
  
  //fprintf (out,"\n\n\n nadjnodnod\n");
  //for (i=0;i<nn;i++)
  //  fprintf(out,"%ld %ld\n",i,nadjnodnod[i]);
  
  adjnodnod = new long* [nn];
  for (i=0;i<nn;i++){
    adjnodnod[i] = new long [nadjnodnod[i]];
    for (j=0;j<nadjnodnod[i];j++){
      adjnodnod[i][j]=auxadjnodnod[i][j];
    }
    delete [] auxadjnodnod[i];
  }
  delete [] auxadjnodnod;
  
  //  printing of array
  //fprintf (out,"\n\n\n LIST OF NUMBERS OF NODES ADJACENT TO NODE\n");
  //for (i=0;i<nn;i++)
  //  fprintf(out,"%ld %ld\n",i,nadjnodnod[i]);
  //fprintf (out,"\n\n\n LIST OF NODES ADJACENT TO NODE\n");
  //for (i=0;i<nn;i++){
  //  fprintf (out,"\n node %4ld  ",i);
  // for (j=0;j<nadjnodnod[i];j++){
  //    fprintf (out,"  %ld",adjnodnod[i][j]);
  // }
  // }
  // fprintf (out,"\n\n\n");
  
}



/**
   function searches for adjacent elements to element,
   adjacent elements must share one edge
   
   
   nadjacedgeelem - array contains list of numbers of adjacent elements to edge elemet
   nadjacedgeelem[i][j] = k - the j-th edge on the i-th element has k neighboring elements
   
   adjacedgeelem - array contains list of adjacent elements to edge element
   adjacedgeelem[i][j][k] = l - the k-th neighbour element of the j-th edge on the i-th element has number l
   
   auxedgeelem - array contains node on edge on element
   auxedgeelem[i][j][k] = l - the k-th node on the j-th edge on the i-th element has number l
   
   for homogeneous mesh only
   
   @param out - output stream
   
   JB, 17.4.2007
*/
void gtopology::adjacelem_edge (FILE *out)
{
  long i,ii,j,jj,k,kk,m,a;
  long ned,nned,ned2,nned2,deg;
  long ***auxedgeelem,**aux;
  long *nodes;
  time_t t1,t2;
  // allocation of array  auxedgeelem
  auxedgeelem = new long** [ne];
  t1=time(NULL);
  for(i = 0; i < ne; i++){
    // number of edge  - gtopology
    ned = give_ned(i);
    // allocation of help array
    nodes = new long [ned];
    // numbers of nodes on one edge - gtopology
    nned = give_nned(i);
    
    auxedgeelem[i] = new long* [ned];
    for(j = 0; j < ned; j++){
      auxedgeelem[i][j] = new long[nned];
      give_edge_nodes(i,j,nodes);
      
      // filling of array  auxegdeelem
      for(k = 0; k < nned ; k++){
	auxedgeelem[i][j][k] = nodes[k];	
      }
    }
    delete []nodes;
  }
  t2=time(NULL);  
  fprintf (out,"time aux :%lf\n",double(t2-t1));
  
  t1=time(NULL);
  // sorting of array auxedgeelem
  for(i = 0; i < ne; i++){
    ned = give_ned(i);
    nned = give_nned(i);
    // degree of approximation function
    deg = give_degree(i);
    for(j =0; j < ned;j++ ){
      for(ii = 1; ii < nned; ii++){
	for(jj = nned-1; jj >= ii; jj--){
	  if(auxedgeelem[i][j][jj-1] >= auxedgeelem[i][j][jj]){
	    a = auxedgeelem[i][j][jj-1];
	    auxedgeelem[i][j][jj-1] = auxedgeelem[i][j][jj];
	    auxedgeelem[i][j][jj] = a;   
	  }
	}
      }
    }
  }
  t2=time(NULL);  
  fprintf (out,"time sorting :%lf\n",double(t2-t1));
   /*
   fprintf (out,"TEST PRINT OF AUXEDGEELEM\n");
   for(i = 0; i < ne; i++){
     fprintf (out,"Element number %ld\n",i+1);
     ned = give_ned(i);
     nned = give_nned(i);
     for(j =0; j < ned;j++ ){
       fprintf (out,"Edge number %ld :",j+1);
       for (k = 0; k < nned; k++){
	 fprintf (out,"     %ld",auxedgeelem[i][j][k]);
       }
       fprintf (out,"\n");
     }
   }
  */
   
  // allocation of array adjacedgeelem
  nadjacedgeelem = new long*[ne];
  t1=time(NULL);  
  // comparing  array  auxedgeelem[i][j][k] and array auxedgeelem[i+1][j][k] for filling of array  nadjacedgeelem
  for(i = 0; i < ne; i++){
    ned = give_ned(i);
    nned = give_nned(i);
    nadjacedgeelem[i] = new long[ned];
    for(j = 0; j < ned; j++ ){
      nadjacedgeelem[i][j] = -1;
      for(ii = 0; ii < ne; ii++){
	ned2 = give_ned(ii);
	nned2 = give_nned(ii);
	for(jj = 0; jj < ned2; jj++){
	  if(auxedgeelem[i][j][0] == auxedgeelem[ii][jj][0]){
	    m = 1;
	    // homogeneous mesh
	    if(nned == nned2){
	      for(k = 1; k < nned; k++){
		if(auxedgeelem[i][j][k] == auxedgeelem[ii][jj][k]){
		  m++;
		}
	      }
	      if(m == nned){
		nadjacedgeelem[i][j]++;
	      }
	    }
	    // inhomogeneous mesh
	    if(ned < ned2 || ned > ned2){
	      fprintf (stderr,"\n\n inhomogeneous mesh in function adjacelem_edge (file %s, line%d).\n",__FILE__,__LINE__); 
	    }
	  }
	}
      }
    }
  }
  t2=time(NULL);  
  fprintf (out,"time comparation :%lf\n",double(t2-t1));
  
  fprintf (out,"\n\n\n LIST OF NUMBERS OF EDGE ADJACENT TO EDGE ON ELEMENT\n");
  for(i = 0; i < ne; i++){
    fprintf (out,"Element number %ld\n",i+1);
    ned = give_ned(i);
    for(j = 0; j < ned; j++ ){
      fprintf (out,"Edge number %ld: %ld\n",j+1,nadjacedgeelem[i][j]);
    }
  }
  
  // allocation of array adjacedgeelem
  adjacedgeelem = new long** [ne];
  for(i = 0; i < ne; i++){
    // number of edge  - gtopology
    ned = give_ned(i);
    adjacedgeelem[i] = new long*[ned];
    for(j = 0; j < ned; j++){
      adjacedgeelem[i][j] = new long [nadjacedgeelem[i][j]];
    }
  }
  
  aux = new long*[ne];
  for(i = 0; i < ne; i++){
    ned = give_ned(i);
    aux[i] = new long[ned];
    for(j = 0; j < ned; j++){
      aux[i][j] = 0;
    }
  }
  
  t1=time(NULL);  
  // comparing  array  auxedgeelem[i][j][k] and array auxedgeelem[i+1][j][k] for filling of array	adjacedgeelem
  for(i = 0; i < ne; i++){
    ned = give_ned(i);
    nned = give_nned(i);
    for(j = 0; j < ned; j++ ){
      k = 0;
      for(ii = i+1; ii < ne; ii++){
	ned2 = give_ned(ii);
	nned2 = give_nned(ii);
	for(jj = 0; jj < ned2; jj++){
	  if(auxedgeelem[i][j][0] == auxedgeelem[ii][jj][0]){
	    m = 1;
	    // homogeneous mesh
	    if(nned == nned2){
	      for(kk = 1; kk < nned; kk++){
		if(auxedgeelem[i][j][kk] == auxedgeelem[ii][jj][kk]){
		  m++;
		}
	      }
	      if(m == nned){
		if(aux[i][j] == 0 ){
		  adjacedgeelem[i][j][k]  = ii;
		  adjacedgeelem[ii][jj][aux[ii][jj]] = i; 
		  aux[ii][jj]++;
		  k++;
		}
		else{
		  adjacedgeelem[i][j][aux[i][j]+k]  = ii;
		  adjacedgeelem[ii][jj][aux[ii][jj]] = i; 
		  aux[ii][jj]++;
		  k++;
		}
	      }
	    }
	    // inhomogeneous mesh
	    if(nned < nned2 || nned > nned2){
	      fprintf (stderr,"\n\n inhomogeneous mesh in function adjacelem_edge (file %s, line%d).\n",__FILE__,__LINE__);  
	    }
	  }
	}
      }
    }
  }
  t2=time(NULL);  
  fprintf (out,"time comparation 2 :%lf\n",double(t2-t1));
  
  fprintf (out,"\n\n\n LIST OF ELEMENT ADJACENT TO EDGE ON ELEMENT\n");
  for(i = 0; i < ne; i++){
    fprintf (out,"Element number %ld\n",i+1);
    ned = give_ned(i);
    nned = give_nned(i);
    for(j = 0; j < ned; j++){
      fprintf (out,"Edge number %ld: ",j+1);
      for(k = 0; k < nadjacedgeelem[i][j]; k++){
	fprintf (out,"     %ld",adjacedgeelem[i][j][k]+1);
      }
      fprintf (out,"\n");
    }
  }
  
  // clearing of array auxedgeelem
  for(i = 0; i < ne; i++){
    ned = give_ned(i);
    delete []aux[i];
    for(j =0; j < ned; j++){
      delete[]auxedgeelem[i][j];
    }
    delete[]auxedgeelem[i];
  }
  delete []auxedgeelem;
  delete []aux;
}



/**
   function asembles list of adjacent elements
   adjacent elements must share one surface
   
   
   nadjacsurfaceelem - array contains list of number of adjacent elements to surface element
   nadjacsurfelem[i][j] = k - the j-th surface on the i-th element has k neighboring elements
   
   adjacsurfaceelem - array contains list of adjacent elements to surface element
   adjacsurfaceelem[i][j][k] = l - the k-th neighbour element of the j-th surface on the i-th element has number l
   
   auxsurfaceelem - array contains node on edge on element
   auxsurfacesurf[i][j][k] = l - the k-th node on the j-th surface on the i-th element has number l
   
   @param out - output stream
   
   JB, 17.4.2007
*/
void gtopology::adjacelem_surf (FILE *out)
{
  long i,ii,j,jj,k,kk,l,m,a;
  long nsurf,nnsurf,nsurf2,nnsurf2;
  long **aux,***auxsurfelem;
  long *nodes;
  
  // allocation of array  auxsurfelem
  auxsurfelem = new long** [ne];
  
  for(i = 0; i < ne; i++){
    // number of surfaces  - gtopology
    nsurf = give_nsurf(i);
    // allocation of help array
    nodes = new long [nsurf];
    // numbers of nodes on one surface - gtopology
    nnsurf = give_nnsurf(i);
    
    auxsurfelem[i] = new long* [nsurf];
    for(j = 0; j < nsurf; j++){
      auxsurfelem[i][j] = new long[nnsurf];
      give_surf_nodes(i,j,nodes);
      
      // filling of array  auxsurfelem
      for(k = 0; k < nnsurf ; k++){
	auxsurfelem[i][j][k] = nodes[k];	
      }
    }
    delete []nodes;
  }
  
  // bubble sorting of array auxsurfelem
  for(i = 0; i < ne; i++){
    nsurf = give_nsurf(i);
    nnsurf = give_nnsurf(i);
    for(j = 0; j < nsurf;j++ ){
      for (ii = 1; ii < nnsurf; ii++){
	for(jj = nnsurf-1; jj >= ii; jj--){
	  if (auxsurfelem[i][j][jj-1] > auxsurfelem[i][j][jj]){ 
	    a = auxsurfelem[i][j][jj-1];
	    auxsurfelem[i][j][jj-1] = auxsurfelem[i][j][jj];
	    auxsurfelem[i][j][jj]= a;
	  }
	}
      }
    }
  }
  
  /*
  fprintf (out,"TEST PRINT OF AUXSURFELEM\n");
   for(i = 0; i < ne; i++){
     fprintf (out,"Element number %ld\n",i+1);
     nsurf = give_nsurf(i);
     nnsurf = give_nnsurf(i);
     for(j =0; j < nsurf;j++ ){
       fprintf (out,"Surf number %ld :",j+1);
       for (k = 0; k < nnsurf; k++){
	 fprintf (out,"     %ld",auxsurfelem[i][j][k]);
       }
       fprintf (out,"\n");
     }
   }
   */
  
  // allocation of array adjacedgeelem
  nadjacsurfelem = new long*[ne];
 
  // comparing  array  auxsurfelem[i][j][k] and array auxsurfelem[i+1][j][k] for  filling of array  nadjacsurfelem
  for(i = 0; i < ne; i++){
    nsurf = give_nsurf(i);
    nnsurf = give_nnsurf(i);
    nadjacsurfelem[i] = new long[nsurf];
    for(j = 0; j < nsurf; j++ ){
      nadjacsurfelem[i][j] = -1;
      for(ii = 0; ii < ne; ii++){
	nsurf2 = give_nsurf(ii);
	nnsurf2 = give_nnsurf(ii);
	for(jj = 0; jj < nsurf2; jj++){
	  if(auxsurfelem[i][j][0] == auxsurfelem[ii][jj][0]){
	    m = 1;
	    // homogeneous meshes
	    if(nnsurf == nnsurf2){
	      for(k = 1; k < nnsurf; k++){
		if(auxsurfelem[i][j][k] == auxsurfelem[ii][jj][k]){
		  m++;
		}
	      }
	      if(m == nnsurf){
		nadjacsurfelem[i][j]++;	
	      }
	    }
	    // inhomogeneous meshes
	    if(nnsurf > nnsurf2){
	      for(kk = 1; kk < nnsurf; kk++){
		for (l = 1; l < nnsurf2; l++ ){
		  if(auxsurfelem[i][j][kk] == auxsurfelem[ii][jj][l]){
		    m++;
		  }
		}
	      }
	      if(m == nnsurf2){
		nadjacsurfelem[i][j]++;		
	      }
	    }
	    // inhomogeneous meshes
	    if(nnsurf < nnsurf2){
	      for(kk = 1; kk < nnsurf2; kk++){
		for (l = 1; l < nnsurf; l++ ){
		  if(auxsurfelem[i][j][kk] == auxsurfelem[ii][jj][l]){
		    m++;
		  }
		}
	      }
	      if(m == nnsurf){
	     	nadjacsurfelem[i][j]++;	
	      }
	    }
	  }
	}
      }
    }
  }
  
  fprintf (out,"\n\n\n LIST OF NUMBERS OF SURFACE ADJACENT TO SURFACE ON ELEMENT\n");
  for(i = 0; i < ne; i++){
    fprintf (out,"Element number %ld\n",i+1);
    nsurf = give_nsurf(i);
    for(j = 0; j < nsurf; j++ ){
      fprintf (out,"Surface number %ld: %ld\n",j+1,nadjacsurfelem[i][j]);
    }
  }
  
  // allocation of array adjacsurfelem
  adjacsurfelem = new long** [ne];
  for(i = 0; i < ne; i++){
    // number of surfaces  - gtopology
    nsurf = give_nsurf(i);
    adjacsurfelem[i] = new long* [nsurf];
    for(j = 0; j < nsurf; j++){
      adjacsurfelem[i][j] = new long [nadjacsurfelem[i][j]];
    }
  }
  
  aux = new long*[ne];
  for(i = 0; i < ne; i++){
    nsurf = give_nsurf(i);
    aux[i] = new long[nsurf];
    for(j = 0; j < nsurf; j++){
      aux[i][j] = 0;
    }
  }
  
  // comparing  array  auxsurfelem[i][j][k] and array auxsurfelem[i+1][j][k]
  for(i = 0; i < ne; i++){
    nsurf = give_nsurf(i);
    nnsurf = give_nnsurf(i);
    for(j = 0; j < nsurf; j++ ){
      k = 0;
      for(ii = i+1; ii < ne; ii++){
	nsurf2 = give_nsurf(ii);
	nnsurf2 = give_nnsurf(ii);
	for(jj = 0; jj < nsurf2; jj++){
	  if(auxsurfelem[i][j][0] == auxsurfelem[ii][jj][0]){
	    m = 1;
	    // homogeneous meshes
	    if(nsurf == nsurf2){
	      for(kk = 1; kk < nnsurf; kk++){
		if(auxsurfelem[i][j][kk] == auxsurfelem[ii][jj][kk]){
		  m++;
		}
	      }
	      if(m == nnsurf){
		if(aux[i][j] == 0 ){
		  adjacsurfelem[i][j][k]  = ii;
		  adjacsurfelem[ii][jj][aux[ii][jj]] = i; 
		  aux[ii][jj]++;
		  k++;
		}
		else{
		  adjacsurfelem[i][j][aux[i][j]+k]  = ii;
		  adjacsurfelem[ii][jj][aux[ii][jj]] = i; 
		  aux[ii][jj]++;
		  k++; 	 
		}
	      }
	    }	
	    // inhomogeneous meshes
	    if(nnsurf > nnsurf2){
	      for(kk = 1; kk < nnsurf; kk++){
		for (l = 1; l < nnsurf2; l++ ){
		  if(auxsurfelem[i][j][kk] == auxsurfelem[ii][jj][l]){
		    m++;
		  }
		}
	      }
	      if(m == nnsurf2){
		if(aux[i][j] == 0 ){
		  adjacsurfelem[i][j][k]  = ii;
		  adjacsurfelem[ii][jj][aux[ii][jj]] = i; 
		  aux[ii][jj]++;
		  k++;
		}
		else{
		  adjacsurfelem[i][j][aux[i][j]+k]  = ii;
		  adjacsurfelem[ii][jj][aux[ii][jj]] = i; 
		  aux[ii][jj]++;
		  k++; 	 
		}
	      }
	    }
	    if(nnsurf < nnsurf2){
	        for(kk = 1; kk < nnsurf2; kk++){
		for (l = 1; l < nnsurf; l++ ){
		  if(auxsurfelem[i][j][kk] == auxsurfelem[ii][jj][l]){
		    m++;
		  }
		}
	      }
	      if(m == nnsurf){
		if(aux[i][j] == 0 ){
		  adjacsurfelem[i][j][k]  = ii;
		  adjacsurfelem[ii][jj][aux[ii][jj]] = i; 
		  aux[ii][jj]++;
		  k++;
		}
		else{
		  adjacsurfelem[i][j][aux[i][j]+k]  = ii;
		  adjacsurfelem[ii][jj][aux[ii][jj]] = i; 
		  aux[ii][jj]++;
		  k++; 	 
		}
	      }
	    } 
	  }
	}
      }
    }
  }
 
  fprintf (out,"\n\n\n LIST OF ELEMENT ADJACENT TO SURFACE ON ELEMENT\n");
  for(i = 0; i < ne; i++){
    fprintf (out,"Element number %ld\n",i+1);
    nsurf = give_nsurf(i);
    nnsurf = give_nnsurf(i);
    for(j = 0; j < nsurf; j++){
      fprintf (out,"Surface number %ld: ",j+1);
      for(k = 0; k < nadjacsurfelem[i][j]; k++){
	fprintf (out,"     %ld",adjacsurfelem[i][j][k]+1);
      }
      fprintf (out,"\n");
    }
  } 
  
  // clearing of array auxsurfelem
  for(i = 0; i < ne; i++){
    nsurf = give_nsurf(i);
    delete []aux[i];
    for(j =0; j < nsurf; j++){
      delete[]auxsurfelem[i][j];
    }
    delete []auxsurfelem[i];
  }
  delete []auxsurfelem;
  delete []aux;
  
}


/**
   function creates list of adjacent nodes to nodes due to a common edge
   for each node, it creates a list of nodes connected by an edge
   
   JB
*/
void gtopology::adjacnodes_edge (FILE *out)
{
  long a,b,i,j,k;
  long ned,nned;
  long *nodes,*cursor,**help_nodes,***auxedgenodes,*auxnadjacnodesedge;
  time_t t1,t2;
    
  // allocation of array  auxedgenodes
  auxedgenodes = new long** [ne];
  
  auxnadjacnodesedge = new long[nn];
  cursor = new long[nn];
  for(i = 0; i < nn; i++){
    auxnadjacnodesedge[i] = 0;
    cursor[i] = 0;
  }
  // assembling of arrays auxedgenodes and auxnadjacnodesedge
  t1=time(NULL);
  for(i = 0; i < ne; i++){
    // number of edgeaces  - gtopology
    ned = give_ned(i);
    // allocation of help array
    nodes = new long [ned];
    // numbers of nodes on one edgeace - gtopology
    nned = give_nned(i);
    auxedgenodes[i] = new long* [ned];
    for(j = 0; j < ned; j++){
      auxedgenodes[i][j] = new long[nned];
      give_edge_nodes(i,j,nodes);
      // filling of array  auxedgenodes
      for(k = 0; k < nned ; k++){
	auxedgenodes[i][j][k] = nodes[k];	
	auxnadjacnodesedge[nodes[k]]++;
      }
    }
    delete []nodes;
  }
  t2=time(NULL);  
  fprintf (out,"time of assembling of arrays auxnadjnodesedge and auxedgenodes :%lf\n",double(t2-t1));
  
  //fprintf (out,"\n\n\n AUXNADJACNODESEDGE \n");
  //for(i = 0; i < nn; i++){
  //  fprintf (out,"%ld %ld\n",i+1,auxnadjacnodesedge[i]);
  //} 
  
  
  help_nodes = new long*[nn];
  for(i = 0; i < nn; i++){
    help_nodes[i] = new long [auxnadjacnodesedge[i]];
  }
  
  
  // fprintf (out,"TEST PRINT OF AUXEDGENODES\n");
  // for(i = 0; i < ne; i++){
  // fprintf (out,"Element number %ld\n",i+1);
  // ned = give_ned(i);
  // nned = give_nned(i);
  // for(j =0; j < ned;j++ ){
  // fprintf (out,"Edge number %ld :",j+1);
  // for (k = 0; k < nned; k++){
  // fprintf (out,"     %ld",auxedgenodes[i][j][k]+1);
  // }
  // fprintf (out,"\n");
  // }
  // }
  
  // assembling of array help_nodes
  t1=time(NULL); 
  for(i = 0; i < ne; i++){
    // number of edgeaces  - gtopology
    ned = give_ned(i);
    for(j = 0; j < ned; j++){
      // filling of array  help_nodes
      if (nned == 2){
	a = auxedgenodes[i][j][0];
	b = auxedgenodes[i][j][1];
	help_nodes[a][cursor[a]]=b;
	cursor[a]++;
	help_nodes[b][cursor[b]]=a;
	cursor[b]++;
      }
    }
  }
  t2=time(NULL);  
  fprintf (out,"time of assembling of array help_nodes :%lf\n",double(t2-t1));
  
  // for(i = 0; i < nn; i++){
  // fprintf (out,"Node number %ld:",i+1);
  // for (j = 0; j < auxnadjacnodesedge[i]; j++){
  // fprintf (out,"   %ld",help_nodes[i][j]+1);
  // }
  // fprintf (out,"\n");
  // }
  
  // deleting of auxedgenodes
  for(i = 0; i < ne; i++){
    ned = give_ned(i);
    for(j = 0; j < ned; j++){
      delete []auxedgenodes[i][j];
      }
    delete []auxedgenodes[i];
    }
  delete []auxedgenodes;
  delete []cursor;
  
  t1=time(NULL);
  nadjacnodesedge = new long[nn];
  adjacnodesedge = new long*[nn];
  for ( i = 0; i < nn; i++){
    // bubble sorting of array auxedgenodes
    for (j = 1; j < auxnadjacnodesedge[i]; j++){
      for (k = auxnadjacnodesedge[i]-1; k >=j ; k--){
	if( help_nodes[i][k-1] > help_nodes[i][k]){
	  a = help_nodes[i][k-1];
	  help_nodes[i][k-1] = help_nodes[i][k];
	  help_nodes[i][k] = a;
	}
      }
    }
    // numbers of adjacent nodes to node due to edge
    //fprintf (out,"Node : %ld\n",i+1);
    nadjacnodesedge[i] = 1;
    for (j = 1; j < auxnadjacnodesedge[i]; j++){
      if( help_nodes[i][j] != help_nodes[i][j-1]){
	nadjacnodesedge[i]++;
      }
    }
    // list of adjacent nodes to node due to edge
    //fprintf (out,"pocet sousedu : %ld\n",nadjacnodesedge[i]);
    adjacnodesedge[i] = new long[nadjacnodesedge[i]];
    adjacnodesedge[i][0] = help_nodes[i][0];
    a = 1;
    for (j = 1; j < auxnadjacnodesedge[i]; j++){
      if( help_nodes[i][j] != help_nodes[i][j-1]){
	adjacnodesedge[i][a]=help_nodes[i][j];
	//fprintf (out," %ld",adjacnodesedge[i][a]);
	a++;
      }
    }
    //fprintf (out,"\n");
  }
  t2=time(NULL);  
  fprintf (out,"time of assembling of arrays nadjacnodesedge and adjacnodesedge :%lf\n",double(t2-t1));
  
  fprintf (out,"\n\nArrays nadjacnodesedge and adjacnodesedge:\n");
  for(i = 0; i < nn; i++){
    fprintf (out,"Node number %ld nadjacnodesedge: %ld adjacnodesedge: ",i+1,nadjacnodesedge[i]);
    for (j = 0; j < nadjacnodesedge[i]; j++){
      fprintf (out,"   %ld",adjacnodesedge[i][j]+1);
    }
    fprintf (out,"\n");
  } 
  
  for(i = 0; i < nn; i++){
    delete []help_nodes[i];
  }
  delete []help_nodes;
  delete []auxnadjacnodesedge;
    
}

/*
  function searches for adjacent nodes which share a surface

  nadjacnodessurf - array of numbers of surfaces which are shared by nodes
  nadjacnodessurf[i]=j - the i-th node belongs to j surfaces
  
  n_nnsurf - array of numbers of nodes on particular surfaces
  n_nnsurf[i][j]=k - the j-th surface adjacent to the i-th node contains k nodes
  
  adjacnodessurf - list of adjacent nodes to nodes which share a surface
  adjacnodessurf[i][j][k]=l - the k-th node of the j-th surface adjacent to the i-th node has number l
  
  JB, 17.4.2007
*/
void gtopology::adjacnodes_surf (FILE *out)
{
  long i,ii,j,jj,k,l,a,n,m;
  long nsurf,nnsurf;
  long *auxnadjacnodessurf,**auxn_nnsurf,***auxadjacnodessurf,***auxsurfnodes;
  long *nodes;
  
  
  // allocation of array  auxsurfnodes
  auxsurfnodes = new long** [ne];
  
  for(i = 0; i < ne; i++){
    // number of surfaces  - gtopology
    nsurf = give_nsurf(i);
    // allocation of help array
    nodes = new long [nsurf];
    // numbers of nodes on one surface - gtopology
    nnsurf = give_nnsurf(i);
    auxsurfnodes[i] = new long* [nsurf];
    for(j = 0; j < nsurf; j++){
      auxsurfnodes[i][j] = new long[nnsurf];
      give_surf_nodes(i,j,nodes);
      
      // filling of array  auxsurfnodes
      for(k = 0; k < nnsurf ; k++){
	auxsurfnodes[i][j][k] = nodes[k];	
      }
    }
    delete []nodes;
  }
  
  // bubble sorting of array auxsurfnodes
  for(i = 0; i < ne; i++){
    nsurf = give_nsurf(i);
    nnsurf = give_nnsurf(i);
    for(j = 0; j < nsurf;j++ ){
      for (ii = 1; ii < nnsurf; ii++){
	for(jj = nnsurf-1; jj >= ii; jj--){
	  if (auxsurfnodes[i][j][jj-1] > auxsurfnodes[i][j][jj]){ 
	    a = auxsurfnodes[i][j][jj-1];
	    auxsurfnodes[i][j][jj-1] = auxsurfnodes[i][j][jj];
	    auxsurfnodes[i][j][jj]= a;
	  }
	}
      }
    }
  }
  /*
  fprintf (out,"TEST PRINT OF AUXSURFNODES\n");
  for(i = 0; i < ne; i++){
    fprintf (out,"Element number %ld\n",i+1);
    nsurf = give_nsurf(i);
    nnsurf = give_nnsurf(i);
    for(j =0; j < nsurf;j++ ){
      fprintf (out,"Surf number %ld :",j+1);
      for (k = 0; k < nnsurf; k++){
	fprintf (out,"     %ld",auxsurfnodes[i][j][k]+1);
      }
      fprintf (out,"\n");
    }
  }
  */
  

  
  auxnadjacnodessurf = new long[nn];
  for(i = 0; i < nn; i++){
    auxnadjacnodessurf[i] = 0;
    for(j = 0; j < ne; j++){
      nsurf = give_nsurf(j);
      nnsurf = give_nnsurf(j);
      for(k = 0; k < nsurf; k++){
	for(l = 0; l < nnsurf; l++){
	  if(i == auxsurfnodes[j][k][l]){
	   auxnadjacnodessurf[i]++;
	  }
	}
      }
    }
  }
  /*
  fprintf (out,"\n\n\n AUXNADJACNODESSURF \n");
  for(i = 0; i < nn; i++){
    fprintf (out,"%ld %ld\n",i+1,auxnadjacnodessurf[i]);
  } 
  */
  
  auxadjacnodessurf = new long**[nn];
  auxn_nnsurf = new long*[nn];
  for(i = 0; i < nn; i++){
    auxadjacnodessurf[i] = new long*[auxnadjacnodessurf[i]];
    auxn_nnsurf[i] = new long[auxnadjacnodessurf[i]];
    n = 0;
    for(j = 0; j < ne; j++){
      nsurf = give_nsurf(j);
      nnsurf = give_nnsurf(j);
      for(k = 0; k < nsurf; k++){
	for(l = 0; l < nnsurf; l++){
	  if(i == auxsurfnodes[j][k][l]){
	    auxadjacnodessurf[i][n] = new long[nnsurf];
	    auxn_nnsurf[i][n]= nnsurf;
	    for(ii = 0; ii < nnsurf; ii++){
	      auxadjacnodessurf[i][n][ii] = auxsurfnodes[j][k][ii];
	    }
	    n++;
	  }
	}
      }
    }
  }
  
  /*
  fprintf (out,"AUXN_NNSURF\n");
  for(i = 0; i < nn; i++){
    fprintf (out,"Node number %ld\n",i+1);
    for(j = 0; j < auxnadjacnodessurf[i]; j++ ){
      fprintf (out,"     %ld",auxn_nnsurf[i][j] );
    }
    fprintf (out,"\n"); 
  }
  */
  
  /*
  fprintf (out,"AUXADJACNODESSURF\n");
  for(i = 0; i < nn; i++){
    fprintf (out,"Node number %ld\n",i+1);
    for(j = 0; j < auxnadjacnodessurf[i]; j++ ){
      for (k = 0; k < auxn_nnsurf[i][j]; k++){
	fprintf (out,"     %ld", auxadjacnodessurf[i][j][k]+1);
      }
      fprintf (out,"\n");
    }
  }
  */
  
  // clearing of array auxsurfnodes
  for(i = 0; i < ne; i++){
    nsurf = give_nsurf(i);
    for(j =0; j < nsurf; j++){
      delete[]auxsurfnodes[i][j];
    }
    delete []auxsurfnodes[i];
  }
  delete []auxsurfnodes;
  
  
  nadjacnodessurf = new long[nn];
  for(i = 0; i < nn; i++){
    nadjacnodessurf[i] = auxnadjacnodessurf[i];
    for(j = 0; j < auxnadjacnodessurf[i]; j++ ){
      for(jj = j+1; jj < auxnadjacnodessurf[i]; jj++ ){
	if(auxadjacnodessurf[i][j][0] == auxadjacnodessurf[i][jj][0]){
	  m = 1;
	  // homogeneous mesh
	  if(auxn_nnsurf[i][j] == auxn_nnsurf[i][jj]){
	    for(k = 1; k < auxn_nnsurf[i][j]; k++){
	      if(auxadjacnodessurf[i][j][k] == auxadjacnodessurf[i][jj][k]){
		m++;
	      }
	    }
	    if(m == auxn_nnsurf[i][j]){
	      nadjacnodessurf[i]--; 
	    }
	  }
	  // inhomogeneus mesh 
	  else{
	    fprintf (stderr,"\n\n inhomogeneous mesh in function adjacnodes_surface (file %s, line%d).\n",__FILE__,__LINE__);  
	  }
	}
      }
    }
  }
  
  fprintf (out,"\n\n\n NADJACNODESSURF \n");
  for(i = 0; i < nn; i++){
    fprintf (out,"%ld %ld\n",i+1,nadjacnodessurf[i]);
  }
  
  adjacnodessurf = new long**[nn];
  n_nnsurf = new long*[nn];
  for(i = 0; i < nn; i++){
    adjacnodessurf[i] = new long*[nadjacnodessurf[i]];
    n_nnsurf[i] = new long[nadjacnodessurf[i]];
    n = 0;
    for(j = 0; j < auxnadjacnodessurf[i]; j++ ){
      a = 0;
      for(jj = j+1; jj < auxnadjacnodessurf[i]; jj++ ){
	if(auxadjacnodessurf[i][j][0] == auxadjacnodessurf[i][jj][0]){
	  m = 1;
	  // homogeneous mesh
	  if(auxn_nnsurf[i][j] == auxn_nnsurf[i][jj]){
	    for(k = 1; k < auxn_nnsurf[i][j]; k++){
	      if(auxadjacnodessurf[i][j][k] == auxadjacnodessurf[i][jj][k]){
		m++;
	      }
	    }
	    if(m == auxn_nnsurf[i][j]){
	      a++;
	    }
	  }
	  // inhomogeneus mesh
	  else{
	    fprintf (stderr,"\n\n inhomogeneous mesh in function adjacnodes_surface (file %s, line%d).\n",__FILE__,__LINE__);  
	  }
	}
      }
      if( a == 0){
	n_nnsurf[i][n] =  auxn_nnsurf[i][j];
	adjacnodessurf[i][n] = new long[n_nnsurf[i][n]];
	for(k = 0; k < 	n_nnsurf[i][n]; k++){
	  adjacnodessurf[i][n][k] = auxadjacnodessurf[i][j][k];
	}
	n++;
      }
    }
  }
   
  fprintf (out,"N_NNSURF\n");
  for(i = 0; i < nn; i++){
    fprintf (out,"Node number %ld\n",i+1);
    for(j = 0; j < nadjacnodessurf[i]; j++ ){
      fprintf (out,"     %ld",n_nnsurf[i][j] );
    }
    fprintf (out,"\n"); 
  }
  
  fprintf (out,"ADJACNODESSURF\n");
  for(i = 0; i < nn; i++){
    fprintf (out,"Node number %ld\n",i+1);
    for(j = 0; j < nadjacnodessurf[i]; j++ ){
      for (k = 0; k < n_nnsurf[i][j]; k++){
	fprintf (out,"     %ld", adjacnodessurf[i][j][k]+1);
      }
      fprintf (out,"\n");
    }
  }
  
  // clearing of arrays auxnadjacnodessurf,auxn_nnsurf and auxnadjacnodessurf 
  for(i = 0; i < nn; i++){
    for(j = 0; j < auxnadjacnodessurf[i]; j++ ){
      delete[]auxadjacnodessurf[i][j];
    }
    delete []auxadjacnodessurf[i];
    delete []auxn_nnsurf[i];
  }
  delete []auxadjacnodessurf;
  delete []auxn_nnsurf;
  delete []auxnadjacnodessurf;
}





/**
   function backups code numbers for later changes in the supports during problem solving

   3.12.2002
*/
void gtopology::backup_codnum (void)
{
  long i, j, ndofn;

  for (i = 0; i < nn; i++)
  {
    ndofn = give_ndofn (i);
    bckcn[i] = new long[ndofn];
    for (j = 0; j < ndofn; j++)
      bckcn[i][j]=give_dof (i,j);
  }
}



/**
   function backups code numbers for later changes in the supports during problem solving

   3.12.2002
*/
void gtopology::restore_codnum (void)
{
  long i, j, ndofn;

  for (i = 0; i < nn; i++)
  {
    ndofn = give_ndofn (i);
    for (j = 0; j < ndofn; j++)
      gnodes[i].cn[j] = bckcn[i][j];
  }
}

void gtopology::write_nodes(FILE *out)
{
  fprintf(out, "NODES\n");
  for (long i = 0; i < nn; i++)
    fprintf(out, "%ld %e %e %e\n", i, gnodes[i].x, gnodes[i].y, gnodes[i].z);
}

/**
   function assembles correspondation between usual nodes and layered nodes
   function creates array unln with nn components, unln[i] represents
   number of layered node which contains i-th usual node
   
   JK, 1.2.2003
*/
void gtopology::unodelnode ()
{
  long i,j;
  unln = new long [nn];
  unnl = new long [nn];
  for (i=0;i<nln;i++){
    for (j=0;j<lgnodes[i].nl;j++){
      unln[lgnodes[i].nodes[j]]=i;
      unnl[lgnodes[i].nodes[j]]=j;
    }
  }
}

/**
   function initiates number of multipliers on elements
   
   JK, 2.2.2003
*/
void gtopology::initiate_elemmult ()
{
  long i,nid,lnid;

  for (i=0;i<ne;i++){
    nid=gelements[i].nodes[0];
    lnid=unln[nid];
    gelements[i].nmult=give_nmult (lnid);
  }
}

/**
   function determines maximum sizes of solved domain
   
   6.10.2003, JK
*/
void gtopology::comp_domain_sizes ()
{
  long i;
  double x,y,z,minx,miny,minz,maxx,maxy,maxz;
  
  domsizes = new double [3];
  
  minx=gnodes[0].x;  maxx=gnodes[0].x;
  miny=gnodes[0].y;  maxy=gnodes[0].y;
  minz=gnodes[0].z;  maxz=gnodes[0].z;
  
  for (i=1;i<nn;i++){
    x=gnodes[i].x;
    y=gnodes[i].y;
    z=gnodes[i].z;
    
    if (minx>x)  minx=x;
    if (maxx<x)  maxx=x;
    
    if (miny>y)  miny=y;
    if (maxy<y)  maxy=y;
    
    if (minz>z)  minx=z;
    if (maxz<z)  maxx=z;
  }
  
  domsizes[0]=maxx-minx;
  domsizes[1]=maxy-miny;
  domsizes[2]=maxz-minz;
}

/**
   function returns sizes of domain
   
   6.10.2003, JK
*/
void gtopology::give_domain_sizes (double *sizes)
{
  sizes[0]=domsizes[0];
  sizes[1]=domsizes[1];
  sizes[2]=domsizes[2];
}


/**
   function reads integer table functions
   function allocates array leso
   
   @param in - input stream
   
   10.2.2006, JK, TKo
*/
void gtopology::read_gf (XFILE *in)
{
  long i,j;
  xfscanf (in,"%ld",&ngf);
  gf = new gfunct [ngf];
  
  for (i=0;i<ngf;i++){
    xfscanf (in,"%ld",&j);
    gf[i].read (in);
  }
  
}

/**
   function reads integer table functions
   function allocates array leso
   
   @param in - input stream
   
   10.2.2006, JK, TKo
*/
void gtopology::read_gf_prop (FILE *in)
{
  long i;
  fscanf (in,"%ld",&ngf);
  gf = new gfunct [ngf];
  
  for (i=0;i<ngf;i++){
    gf[i].read_prop (in);
  }
  
}

/**
   function prints integer table functions
   
   @param out - output stream
   
   17.7.2006, TKr
*/
void gtopology::print_gf (FILE *out)
{
  long i;
  fprintf (out,"\n\n %ld\n",ngf);
  
  for (i=0;i<ngf;i++){
    fprintf (out,"%ld\n",i+1);
    gf[i].print (out);
  }
}

/**
   function allocates and assigns array of nodes and elements which are switched on
   
   3.3.2006, JK
*/
void gtopology::lneso_init ()
{
  long i;
  
  //  array of elements which are switched on
  leso = new long [ne];
  
  for (i=0;i<ne;i++){
    leso[i]=1;
  }

  //  array of nodes which are switched on
  lnso = new long [nn];
  
  for (i=0;i<nn;i++){
    lnso[i]=1;
  }
}


/**
   function initiates variables auxinf on elements
   
   JK, 3.3.2006
*/
void gtopology::auxinf_init ()
{
  long i;
  
  for (i=0;i<ne;i++){
    gelements[i].auxinf=0;
  }
}

/**
   function switches on appropriate elements
   function is used in problems with changing number of nodes, elements and DOFs
   former elements are not temporarily switched off
   
   leso[i]=0 - the i-th element is switched off
   leso[i]=1 - the i-th element is switched on
   
   @param time - actual time
   
   JK, 28.2.2006, last modification 3.11.2006 
*/
void gtopology::update_elem (double time)
{
  long i,j,k;
  
  for (i=0;i<ne;i++){
    //  id of time function
    j=gelements[i].tgf;
    //  value of time function
    k=gf[j].getval_long (time);

    leso[i]=k;
  }
}

/**
   function switches on appropriate nodes
   function is used in problems with changing number of nodes, elements and DOFs
   former nodes are not temporarily switched off
   
   lnso[i]=0 - the i-th node is switched off
   lnso[i]=1 - the i-th node is switched on
   
   JK, 7.11.2006
*/
void gtopology::update_nodes ()
{
  long i,j,nne;
  ivector nod;
  
  for (i=0;i<nn;i++){
    lnso[i]=0;
  }
  
  for (i=0;i<ne;i++){
    if (leso[i]==1){
      //  only element switched on are investigated
      
      //  number of nodes on element
      nne = give_nne (i);
      allocv (nne,nod);
      //  element nodes
      give_nodes (i,nod);
      for (j=0;j<nne;j++){
	lnso[nod[j]]=1;
      }
      destrv (nod);
    }
  }
}

/**
   function searches for new elements
   function is used in problems with changing number of nodes, elements and DOFs
   function returns number of changed elements
   only new elements are switched on, former elements are temporarily switched off
   
   @param time - actual time
   @param prev_time - previous time
   
   JK, 22.9.2006, last modification 3.11.2006
*/
long gtopology::search_newelem (double time,double prev_time)
{
  long i,j,k,l,nae;
  
  //  number of added elements at this time
  nae=0;
  
  if (time>prev_time){
    for (i=0;i<ne;i++){
      //  id of time function
      j=gelements[i].tgf;
      //  value of time function at actual time
      k=gf[j].getval_long (time);
      //  value of time function at previous time
      l=gf[j].getval_long (prev_time);
      
      if (k==l){
	leso[i]=0;
      }
      else{
	leso[i]=1;
	nae++;
      }
    }
  }
  else{
    //  initial time step
    //  time = starting_time
    for (i=0;i<ne;i++){
      //  id of time function
      j=gelements[i].tgf;
      //  value of time function at actual time
      k=gf[j].getval_long (time);
      
      if (k==0){
	leso[i]=0;
      }
      else{
	leso[i]=1;
	nae++;
      }
    }
  }
  
  return nae;
}

/**
   function searches for new DOFs
   function is used in problems with changing number of nodes, elements and DOFs
   only new DOFs are switched on, former DOFs are temporarily switched off
   
   @param time - actual time
   @param prev_time - previous time
   
   JK, 22.9.2006, last modification 3.11.2006
*/
long gtopology::search_newdofs (double time,double prev_time)
{
  long i,nch;
  
  //  number of changes
  nch=0;
  for (i=0;i<nn;i++){
    nch += gnodes[i].search_newdofs (gf,time,prev_time);
  }
  
  //  state of code numbers is changed
  cnstate=0;
  
  return nch;
}


/**
   function switches on appropriate DOFs
   function is used in problems with changing number of nodes, elements and DOFs
   former DOfs are not temporarily switched off
   
   @param time - actual time
   
   6.2.2006, JK, last modification 3.11.2006
*/
void gtopology::update_dofs (double time)
{
  long i;
  
  for (i=0;i<nn;i++){
    gnodes[i].update_dofs (gf,time);
  }

  //  state of code numbers is changed
  cnstate=0;

}


/**
   function changes nodes and elements
   it is used in problems with growing numbers of nodes and elements
   nodes and elements are updated at particular instance of time
   
   JK, 28.2.2006
*/
/*
void gtopology::update_mesh (double time,double prev_time,long &ncn,long &nce)
{

  //  update of nodes
  //  function returns number of changed DOFs
  ncn = update_dofs (time,prev_time);
  
  //  update of element status
  nce = update_elemstat (time);

}
*/

void gtopology::print_time_functions (FILE *kon,double time)
{
  long i;
  //double time1,time2,time=10800000.0,dt=1123200.000000;
  //FILE *kon;
  //kon = fopen ("casfun","w");
  
  //time1=time;
  //time2=time+dt;
  fprintf (kon,"\n\n vypis v case %le\n",time);
  for (i=0;i<ngf;i++){
    //fprintf (kon,"%ld %ld\n",gf[i].getval_long (time1),gf[i].getval_long (time2));
    fprintf (kon,"%ld %ld\n",i,gf[i].getval_long (time));
  }
  
  //fclose (kon);
}







/**
   function searches for common end nodes on interfaces
   it is used in automatic subdomain detection

   function is used for problems with hemivariational inequalities
   
   JK, 12.10.2008
*/
void gtopology::end_nodes (FILE *out)
{
  long i,j,k,nsn,nene;
  long *nodes,*sn;
  long *av;
  
  nodes = new long[2];
  
  //  auxiliary array
  av = new long [nn];
  for (i=0;i<nn;i++){
    av[i]=0;
  }
  
  //  loop over elements
  for (i=0;i<ne;i++){
    //  number of end nodes on element
    nene=give_nen (i);
    
    if (nene>0){
      //  nodes on required edge
      give_end_nodes (i,nodes);
      
      //  loop over egdes on element
      for (j=0;j<2;j++){
	
	//  positions in array av are filled with respect to end nodes
	//  internal end points are characterized by av[i]=2
	//  boundary or interface end points  are characterized by av[i]=1
	av[nodes[j]]++;
      }
    }
  }
  

  fprintf (out,"\n\n\n kontrola koncovych bodu \n\n");
  for (i=0;i<nn;i++){
    fprintf (out,"\n node %4ld  av %4ld",i,av[i]);
  }
  fprintf (out,"\n\n\n");
  
  //  number of suspicious nodes
  nsn=0;
  for (i=0;i<nn;i++){
    if (av[i]==1)
      nsn++;
  }
  
  //  list of suspicious nodes
  sn = new long [nsn];
  nsn=0;
  for (i=0;i<nn;i++){
    if (av[i]==1){
      sn[nsn]=i;
      nsn++;
    }
  }

  //  number of end nodes
  nen=0;
  double threshold=1.0e-7;
  for (i=0;i<nsn;i++){
    for (j=i+1;j<nsn;j++){
      if (fabs(gnodes[sn[j]].x-gnodes[sn[i]].x)<threshold){
	if (fabs(gnodes[sn[j]].y-gnodes[sn[i]].y)<threshold){
	  if (fabs(gnodes[sn[j]].z-gnodes[sn[i]].z)<threshold){
	    nen++;
	  }
	}
      }
    }
  }
  
  fprintf (out,"\n\n\n pocet zajimavych koncovych bodu  %ld \n\n",nen);
  
  endnodes = new endnode [nen];
  k=0;
  for (i=0;i<nsn;i++){
    for (j=i+1;j<nsn;j++){
      if (fabs(gnodes[sn[j]].x-gnodes[sn[i]].x)<threshold){
	if (fabs(gnodes[sn[j]].y-gnodes[sn[i]].y)<threshold){
	  if (fabs(gnodes[sn[j]].z-gnodes[sn[i]].z)<threshold){
	    endnodes[k].nn=2;
	    endnodes[k].nm=2;
	    endnodes[k].fn=sn[i];
	    endnodes[k].ln=sn[j];
	    k++;
	  }
	}
      }
    }
  }
  
  delete [] nodes;
  delete [] sn;
  delete [] av;
}

/**
   function prints auxiliary data about end nodes
   
   @param out - output file
   
   JK, 13.10.2008
*/
void gtopology::endnodes_auxprint (FILE *out)
{
  long i;
  
  fprintf (out,"\n\n number of end nodes %ld",nen);
  
  for (i=0;i<nen;i++){
    fprintf (out,"\n\n end node number %6ld",i);
    fprintf (out,"\n number of nodes  %ld",endnodes[i].nn);
    fprintf (out,"\n number of Lagrange multipliers  %ld",endnodes[i].nm);
    fprintf (out,"\n node number 1    %ld",endnodes[i].fn);
    fprintf (out,"\n node number 2    %ld",endnodes[i].ln);
  }
}

/**
   function allocates array cnm on end nodes
   cnm contains code numbers of Lagrange multipliers
   
   JK, 13.10.2008
*/
void gtopology::alloc_endnode_cn ()
{
  long i,nid,ndofn;
  
  for (i=0;i<nen;i++){
    //  first node
    nid=endnodes[i].fn;
    //  number of DOFs on the first node
    ndofn=give_ndofn (nid);
    
    //  allocation of array cnm
    endnodes[i].alloc_cnm (ndofn);
  }

}

/**
   function localizes contributions from end nodes to the global %matrix
   
   @param gm - pointer to the general %matrix
   
   JK, 14.10.2008
*/
void gtopology::endnodes_localization (gmatrix *gm)
{
  long i,nm,*ncn1,*ncn2,*mcn;

  for (i=0;i<nen;i++){
    //  number of multipliers between nodes
    nm=endnodes[i].ndofn;
    ncn1 = new long [nm];
    ncn2 = new long [nm];
    mcn = new long [nm];

    //  code numbers on first nodes and appropriate multipliers
    give_endnode_code_numbers (i,ncn1,ncn2,mcn);
    
    //  localization of contributions from edges
    gm->mult_localize (nm,ncn1,ncn2,mcn);
    
    delete [] mcn;
    delete [] ncn2;
    delete [] ncn1;
  }
}

/**
   function assignes numbers of all elements to end nodes
   
   JK, 17.10.2008
*/
void gtopology::enodes_allelem (FILE *out)
{
  long i,k,l,fn,nne,stop;
  ivector nod;
  
  //  loop over all end nodes
  for (i=0;i<nen;i++){
    //  allocation of array adjel
    endnodes[i].adjel = new long [2];

    //  first node
    fn=endnodes[i].fn;
    
    //  element containing node number fn has to be searched for
    //  it is necessary to go through all elements until appropriate element is found
    stop=0;
    for (k=0;k<ne;k++){
      //  number of nodes on element
      nne = give_nne (k);
      //  array allocation
      allocv (nne,nod);
      //  element nodes
      give_nodes (k,nod);
      
      //  loop over nodes on element
      for (l=0;l<nne;l++){
	if (nod[l]==fn){
	  endnodes[i].adjel[0]=k;
	  stop=1;
	  break;
	}
      }
      destrv (nod);
      if (stop==1)
	break;
    }


    //  second node
    fn=endnodes[i].ln;
    
    //  element containing node number fn has to be searched for
    //  it is necessary to go through all elements until appropriate element is found
    stop=0;
    for (k=0;k<ne;k++){
      //  number of nodes on element
      nne = give_nne (k);
      //  array allocation
      allocv (nne,nod);
      //  element nodes
      give_nodes (k,nod);
      
      //  loop over nodes on element
      for (l=0;l<nne;l++){
	if (nod[l]==fn){
	  endnodes[i].adjel[1]=k;
	  stop=1;
	  break;
	}
      }
      destrv (nod);
      if (stop==1)
	break;
    }
  }
}










































/**
   function searches for common edges on interfaces
   it is used in automatic subdomain detection

   function is used for problems with hemivariational inequalities
   
   JK, 11.7.2007
*/
void gtopology::edges (FILE *out)
{
  long i,j,k,ned,nned,nsed;
  long *nodes,*fn,*ln;
  long **av;

  //  auxiliary array
  //  it has the same structure as the array adjnodnod
  av = new long* [nn];
  for (i=0;i<nn;i++){
    av[i] = new long [nadjnodnod[i]];
    for (j=0;j<nadjnodnod[i];j++){
      av[i][j]=0;
    }
  }
  
  //  loop over elements
  for (i=0;i<ne;i++){
    //  number of edges on element
    ned=give_ned (i);
    //  number of nodes on one edge
    nned=give_nned (i);
    //  nodes on required edge
    nodes=new long [nned];
    
    //  loop over egdes on element
    for (j=0;j<ned;j++){
      //  nodes on required edge
      give_edge_nodes (i,j,nodes);
      
      //  positions in array av are filled with respect to edge nodes
      //  internal edges (between the i-th and the j-th node) are characterized by av[i][j]=2
      //  boundary or interface edges (between the i-th and the j-th node) are characterized by av[i][j]=1
      for (k=0;k<nadjnodnod[nodes[0]];k++){
	if (adjnodnod[nodes[0]][k]==nodes[nned-1]){
	  av[nodes[0]][k]++;
	  break;
	}
      }
      for (k=0;k<nadjnodnod[nodes[nned-1]];k++){
	if (adjnodnod[nodes[nned-1]][k]==nodes[0]){
	  av[nodes[nned-1]][k]++;
	  break;
	}
      }
    }

    delete [] nodes;
  }
  

  fprintf (out,"\n\n\n kontrola hran \n\n");
  for (i=0;i<nn;i++){
    fprintf (out,"\n node %4ld  ",i);
    for (j=0;j<nadjnodnod[i];j++){
      fprintf (out,"    %4ld : %2ld",adjnodnod[i][j],av[i][j]);
    }
  }
  fprintf (out,"\n\n\n");

  
  //  number of suspicious edges
  nsed=0;
  for (i=0;i<nn;i++){
    for (j=0;j<nadjnodnod[i];j++){
      if (av[i][j]==1)
	nsed++;
    }
  }
  nsed/=2;
  
  //  list of first nodes on edges
  fn = new long [nsed];
  //  list of last nodes on edges
  ln = new long [nsed];
  
  k=0;
  for (i=0;i<nn;i++){
    for (j=0;j<nadjnodnod[i];j++){
      if (av[i][j]==1){
	if (i<adjnodnod[i][j]){
	  fn[k]=i;
	  ln[k]=adjnodnod[i][j];
	  k++;
	}
      }
    }
  }
  
  fprintf (out,"\n\n\n druha kontrola hran \n\n");
  for (i=0;i<nsed;i++){
    fprintf (out,"\n fn  %5ld  ln %5ld",fn[i],ln[i]);
  }
  
  //  number of generalized edges
  nged=0;
  double threshold=1.0e-7;
  for (i=0;i<nsed;i++){
    for (j=i+1;j<nsed;j++){
      if (fabs(gnodes[fn[j]].x-gnodes[fn[i]].x)<threshold && fabs(gnodes[ln[j]].x-gnodes[ln[i]].x)<threshold){
	if (fabs(gnodes[fn[j]].y-gnodes[fn[i]].y)<threshold && fabs(gnodes[ln[j]].y-gnodes[ln[i]].y)<threshold){
	  if (fabs(gnodes[fn[j]].z-gnodes[fn[i]].z)<threshold && fabs(gnodes[ln[j]].z-gnodes[ln[i]].z)<threshold){
	    nged++;
	  }
	}
      }
      if (fabs(gnodes[fn[j]].x-gnodes[ln[i]].x)<threshold && fabs(gnodes[ln[j]].x-gnodes[fn[i]].x)<threshold){
	if (fabs(gnodes[fn[j]].y-gnodes[ln[i]].y)<threshold && fabs(gnodes[ln[j]].y-gnodes[fn[i]].y)<threshold){
	  if (fabs(gnodes[fn[j]].z-gnodes[ln[i]].z)<threshold && fabs(gnodes[ln[j]].z-gnodes[fn[i]].z)<threshold){
	    nged++;
	  }
	}
      }
    }
  }
  
  fprintf (out,"\n\n\n pocet zajimavych hran  %ld \n\n",nged);
  
  gedges = new gedge [nged];
  k=0;
  for (i=0;i<nsed;i++){
    for (j=i+1;j<nsed;j++){
      if (fabs(gnodes[fn[j]].x-gnodes[fn[i]].x)<threshold && fabs(gnodes[ln[j]].x-gnodes[ln[i]].x)<threshold){
	if (fabs(gnodes[fn[j]].y-gnodes[fn[i]].y)<threshold && fabs(gnodes[ln[j]].y-gnodes[ln[i]].y)<threshold){
	  if (fabs(gnodes[fn[j]].z-gnodes[fn[i]].z)<threshold && fabs(gnodes[ln[j]].z-gnodes[ln[i]].z)<threshold){
	    gedges[k].nn=4;
	    gedges[k].nm=2;
	    gedges[k].fn=fn[i];
	    gedges[k].ln=ln[i];
	    gedges[k].nlist = new long [4];
	    gedges[k].nlist[0]=fn[i];
	    gedges[k].nlist[1]=fn[j];
	    gedges[k].nlist[2]=ln[i];
	    gedges[k].nlist[3]=ln[j];
	    k++;
	  }
	}
      }
      if (fabs(gnodes[fn[j]].x-gnodes[ln[i]].x)<threshold && fabs(gnodes[ln[j]].x-gnodes[fn[i]].x)<threshold){
	if (fabs(gnodes[fn[j]].y-gnodes[ln[i]].y)<threshold && fabs(gnodes[ln[j]].y-gnodes[fn[i]].y)<threshold){
	  if (fabs(gnodes[fn[j]].z-gnodes[ln[i]].z)<threshold && fabs(gnodes[ln[j]].z-gnodes[fn[i]].z)<threshold){
	    gedges[k].nn=4;
	    gedges[k].nm=2;
	    gedges[k].fn=fn[i];
	    gedges[k].ln=ln[i];
	    gedges[k].nlist = new long [4];
	    gedges[k].nlist[0]=fn[i];
	    gedges[k].nlist[1]=ln[j];
	    gedges[k].nlist[2]=ln[i];
	    gedges[k].nlist[3]=fn[j];
	    k++;
	  }
	}
      }
    }
  }
  
  //for (i=0;i<nged;i++){
  //gedges[i].print (out);
  //}
  
  delete [] ln;
  delete [] fn;
  for (i=0;i<nn;i++){
    delete [] av[i];
  }
  delete [] av;

}

/**
   function computes direction vectors of edges
   
   JK, 11.7.2007
*/
void gtopology::edge_dirvect ()
{
  long i;
  
  for (i=0;i<nged;i++){
    gedges[i].direction_vector (gnodes);
  }
}

/**
   function computes direction vectors of edges
   
   JK, 11.7.2007
*/
void gtopology::edge_normvect ()
{
  long i;
  
  for (i=0;i<nged;i++){
    gedges[i].normal_vector (gnodes);
  }
}

/**
   function sorts first and last nodes on edges
   
   JK, 12.7.2007
*/
void gtopology::edgenode_sorting ()
{
  long i,j,nc;
  
  nc=0;
  for (i=0;i<nged;i++){
    if (gedges[i].nm != 2 || gedges[i].nn != 4){
      fprintf (stderr,"\n\n wrong number of node multiplicity on edge (file %s, line %d)\n",__FILE__,__LINE__);
      abort ();
    }
    if (gedges[i].nlist[0]>gedges[i].nlist[1]){
      j=gedges[i].nlist[1];
      gedges[i].nlist[1]=gedges[i].nlist[0];
      gedges[i].nlist[0]=j;
      nc++;
    }
    if (gedges[i].nlist[2]>gedges[i].nlist[3]){
      j=gedges[i].nlist[3];
      gedges[i].nlist[3]=gedges[i].nlist[2];
      gedges[i].nlist[2]=j;
      nc++;
    }
  }
  
  fprintf (stdout,"\n number of corrections in function edgenode_sorting is %ld",nc);
}


/**
   function searches for previous and next edges for each edge
   
   JK, 12.7.2007
*/
void gtopology::prev_next_edges ()
{
  long i,j;
  long ifn1,ifn2,iln1,iln2,jfn1,jfn2,jln1,jln2;
  
  for (i=0;i<nged;i++){
    if (gedges[i].nm != 2 || gedges[i].nn != 4){
      fprintf (stderr,"\n\n wrong number of node multiplicity on edge (file %s, line %d)\n",__FILE__,__LINE__);
      abort ();
    }
    
    ifn1=gedges[i].nlist[0];
    ifn2=gedges[i].nlist[1];
    iln1=gedges[i].nlist[2];
    iln2=gedges[i].nlist[3];

    for (j=i+1;j<nged;j++){
      jfn1=gedges[j].nlist[0];
      jfn2=gedges[j].nlist[1];
      jln1=gedges[j].nlist[2];
      jln2=gedges[j].nlist[3];
      
      if (iln1==jfn1 && iln2==jfn2){
	gedges[i].next=j;
	gedges[j].prev=i;
      }
      if (ifn1==jln1 && ifn2==jln2){
	gedges[i].prev=j;
	gedges[j].next=i;
      }
      if (ifn1==jfn1 && ifn2==jfn2){
	gedges[i].prev=j;
	gedges[j].prev=i;
      }
      if (iln1==jln1 && iln2==jln2){
	gedges[i].next=j;
	gedges[j].next=i;
      }
      
    }
  }
  
}



/**
   function searches for series of edges
   
   JK, 12.7.2007
*/
void gtopology::edge_series (FILE *out)
{
  long i,j,k,n,c,e,a,m;
  long *av;

  //  number of series
  nser=0;

  //  list of correspondence between edges and series
  edgeser = new long [nged];
  for (i=0;i<nged;i++){
    edgeser[i]=-1;
  }
  
  //  series of egdes which contain only one edge
  c=0;
  for (i=0;i<nged;i++){
    if (gedges[i].prev==-1 && gedges[i].next==-1){
      edgeser[i]=nser;
      nser++;
      c++;
    }
  }
  
  a=-2; n=-2;
  for (i=0;i<nged;i++){
    if (gedges[i].prev==-1 && edgeser[i]==-1){
      //  untouched edge with no predecessor
      a=i;
      n=gedges[i].next;
      edgeser[i]=nser;
      break;
    }
  }
  if (a==-2 && n==-2){
    //  there is no edge with no predecessor
    for (i=0;i<nged;i++){
      if (gedges[i].next==-1 && edgeser[i]==-1){
	a=i;
	n=gedges[i].prev;
	edgeser[i]=nser;

	gedges[i].next=gedges[i].prev;
	gedges[i].prev=-1;
	m=gedges[i].nlist[0];
	gedges[i].nlist[0]=gedges[i].nlist[2];
	gedges[i].nlist[2]=m;
	m=gedges[i].nlist[1];
	gedges[i].nlist[1]=gedges[i].nlist[3];
	gedges[i].nlist[3]=m;
	
	m=gedges[i].fn;
	gedges[i].fn=gedges[i].ln;
	gedges[i].ln=m;
	
	break;
      }
    }
  }
  c++;
  
  if (c<nged){
    do{
      e=gedges[n].next;
      if (a==e){
	gedges[n].next=gedges[n].prev;
	gedges[n].prev=a;
	m=gedges[n].nlist[0];
	gedges[n].nlist[0]=gedges[n].nlist[2];
	gedges[n].nlist[2]=m;
	m=gedges[n].nlist[1];
	gedges[n].nlist[1]=gedges[n].nlist[3];
	gedges[n].nlist[3]=m;
	
	m=gedges[n].fn;
	gedges[n].fn=gedges[n].ln;
	gedges[n].ln=m;
	
	a=n;
	n=gedges[n].next;
      }
      else{
	a=n;
	n=e;
      }
      edgeser[a]=nser;
      c++;
      
      if (n<0){
	j=0;
	for (i=0;i<nged;i++){
	  if (gedges[i].prev==-1 && edgeser[i]==-1){
	    a=i;
	    n=gedges[i].next;
	    nser++;
	    edgeser[i]=nser;
	    c++;
	    j=1;
	    break;
	  }
	}
	if (j==0){
	  for (i=0;i<nged;i++){
	    if (gedges[i].next==-1 && edgeser[i]==-1){
	      
	      a=i;
	      n=gedges[i].prev;
	      edgeser[i]=nser;
	      
	      gedges[i].next=gedges[i].prev;
	      gedges[i].prev=-1;
	      m=gedges[i].nlist[0];
	      gedges[i].nlist[0]=gedges[i].nlist[2];
	      gedges[i].nlist[2]=m;
	      m=gedges[i].nlist[1];
	      gedges[i].nlist[1]=gedges[i].nlist[3];
	      gedges[i].nlist[3]=m;
	      
	      m=gedges[i].fn;
	      gedges[i].fn=gedges[i].ln;
	      gedges[i].ln=m;
	      
	      nser++;
	      edgeser[i]=nser;
	      c++;
	      break;
	    }
	  }
	}
      }
    }while(c<nged);
    nser++;
  }
  
  fprintf (stdout,"\n\n number of series %ld",nser);
  for (i=0;i<nser;i++){
    fprintf (out,"%ld\n",edgeser[i]);
  }
  fprintf (out,"\n");

  //  number of edges in series
  nedser = new long [nser];
  for (i=0;i<nser;i++){
    nedser[i]=0;
    for (j=0;j<nged;j++){
      if (edgeser[j]==i)
	nedser[i]++;
    }
  }
  
  
  
  
  
  edgelist = new long* [nser];
  for (i=0;i<nser;i++){
    edgelist[i] = new long [nedser[i]];
  }
  av = new long [nged];
  for (i=0;i<nged;i++){
    av[i]=-1;
  }
  
  for (i=0;i<nser;i++){

    for (j=0;j<nged;j++){
      if (av[j]>-1)
	continue;
      else{
	if (gedges[j].prev==-1){
	  a=j;
	  break;
	}
      }
    }
    
    k=0;
    for (j=0;j<nged;j++){
      edgelist[i][k]=a;
      k++;
      av[a]=1;
      n=gedges[a].next;
      a=n;
      if (n==-1)
	break;
    }
  }
  
  
  fprintf (out,"\n\n\n kontrola pole edgelist \n");
  for (i=0;i<nser;i++){
    fprintf (out,"\n serie %5ld   ",i);
    for (j=0;j<nedser[i];j++){
      fprintf (out,"  %ld",edgelist[i][j]);
    }
  }
  fprintf (out,"\n\n");
  
  delete [] av;
}

/**
   function assignes numbers of element to edges
   
   JK, 21.10.2007
*/
void gtopology::edge_elem (FILE *out)
{
  long i,j,k,l,m,fn,ln,ef,nne,nc,eid;
  ivector nod;

  //  list of adjacent elements to elements
  adjacelem (out);
  
  //  loop over number of series
  for (i=0;i<nser;i++){
    //  element not found
    ef=-1;
    
    //  loop over all general edges
    for (j=0;j<nedser[i];j++){
      
      m=edgelist[i][j];
      //  first node
      fn=gedges[m].fn;
      //  last node
      ln=gedges[m].ln;
      
      if (ef==-1){
	//  element containing node number fn and ln has to be searched for
	//  it is necessary to go through all elements until appropriate element is found
	for (k=0;k<ne;k++){
	  //  number of nodes on element
	  nne = give_nne (k);
	  //  array allocation
	  allocv (nne,nod);
	  //  element nodes
	  give_nodes (k,nod);
	  
	  nc=0;
	  //  loop over nodes on element
	  for (l=0;l<nne;l++){
	    if (nod[l]==fn || nod[l]==ln)
	      nc++;
	  }
	  destrv (nod);
	  
	  if (nc==2){
	    gedges[m].re=k;
	    ef=k;
	    break;
	  }
	}
      }
      else{
	//  element containing node number fn and ln has to be searched for
	//  only adjacent elements are investigated
	
	//  loop over adjacent elements
	for (k=0;k<nadjelel[ef];k++){
	  //  element id
	  eid=adjelel[ef][k];
	  //  number of nodes on element
	  nne = give_nne (eid);
	  //  array allocation
	  allocv (nne,nod);
	  //  element nodes
	  give_nodes (eid,nod);
	  
	  nc=0;
	  //  loop over nodes on element
	  for (l=0;l<nne;l++){
	    if (nod[l]==fn || nod[l]==ln)
	      nc++;
	  }
	  destrv (nod);
	  
	  if (nc==2){
	    gedges[m].re=eid;
	    ef=eid;
	    break;
	  }
	  
	}
      }
      
    }
  }
}


/**
   function assignes numbers of all elements to edges
   
   JK, 6.8.2008
*/
void gtopology::edge_allelem (FILE *out)
{
  long i,j,k,l,m,fn,ln,ef,nne,nc,eid;
  ivector nod;
  
  for (i=0;i<nged;i++){
    gedges[i].adjel = new long [2];
    gedges[i].adjel[0]=gedges[i].re;
  }
  
  //  loop over number of series
  for (i=0;i<nser;i++){
    //  element not found
    ef=-1;
    
    //  loop over all general edges
    for (j=0;j<nedser[i];j++){
      
      m=edgelist[i][j];
      //  first node
      fn=gedges[m].nlist[1];
      //  last node
      ln=gedges[m].nlist[3];
      
      if (ef==-1){
	//  element containing node number fn and ln has to be searched for
	//  it is necessary to go through all elements until appropriate element is found
	for (k=0;k<ne;k++){
	  //  number of nodes on element
	  nne = give_nne (k);
	  //  array allocation
	  allocv (nne,nod);
	  //  element nodes
	  give_nodes (k,nod);
	  
	  nc=0;
	  //  loop over nodes on element
	  for (l=0;l<nne;l++){
	    if (nod[l]==fn || nod[l]==ln)
	      nc++;
	  }
	  destrv (nod);
	  
	  if (nc==2){
	    gedges[m].adjel[1]=k;
	    ef=k;
	    break;
	  }
	}
      }
      else{
	//  element containing node number fn and ln has to be searched for
	//  only adjacent elements are investigated
	
	//  loop over adjacent elements
	for (k=0;k<nadjelel[ef];k++){
	  //  element id
	  eid=adjelel[ef][k];
	  //  number of nodes on element
	  nne = give_nne (eid);
	  //  array allocation
	  allocv (nne,nod);
	  //  element nodes
	  give_nodes (eid,nod);
	  
	  nc=0;
	  //  loop over nodes on element
	  for (l=0;l<nne;l++){
	    if (nod[l]==fn || nod[l]==ln)
	      nc++;
	  }
	  destrv (nod);
	  
	  if (nc==2){
	    gedges[m].adjel[1]=eid;
	    ef=eid;
	    break;
	  }
	  
	}
      }
      
    }
  }
  
}



/**
   function checkes orientation of normal vectors
   
   JK, 21.10.2007
*/
void gtopology::normvectorient ()
{
  long i,eid,nne;
  ivector nod;
  vector x,y;
  
  for (i=0;i<nged;i++){
    //  number of reference element
    eid=gedges[i].re;
    //  number of nodes on element
    nne = give_nne (eid);
    //  allocation of vectors
    allocv (nne,x);
    allocv (nne,y);
    allocv (nne,nod);
    
    //  coordinates of element node
    give_node_coord2d (x,y,eid);
    //  element nodes
    give_nodes (eid,nod);
    
    //  check of normal vectors
    gedges[i].check_normal (x,y,nod);
    
    //  deallocation of vectors
    destrv (nod);
    destrv (x);
    destrv (y);
  }

}


/**
   function modifies indicators of code numbers
   in problems of hemivariational inequalities
   
   indicators have to be modified with respect to
   
*/
/*
void gtopology::cnmodif (FILE *out)
{
  long i,j,k,l,max,acn,ndofn;
  long **av=NULL;
  
  fprintf (out,"\n\n kontrola indikatoru kodovych cisel \n");
  for (i=0;i<nn;i++){
    fprintf (out,"\n node %6ld",i+1);
    ndofn = give_ndofn (i);
    for (j=0;j<ndofn;j++){
      fprintf (out,"  %ld",give_dof (i,j));
    }
  }
  
  max=0;
  acn=0;
  for (i=0;i<nn;i++){
    if (max<stop->ltg1[i])
      max=stop->ltg1[i];
    ndofn = give_ndofn (i);
    for (j=0;j<ndofn;j++){
      k=give_dof (i,j);
      if (k>acn)
	acn=k;
    }
  }
  max++;
  acn++;
  
  av = new long* [max];
  
  for (i=0;i<nn;i++){
    l=stop->ltg1[i];
    if (l>-1){
      ndofn = give_ndofn (i);
      av[l] = new long [ndofn];
    }
  }
  
  
  
  for (i=0;i<nn;i++){
    l=stop->ltg1[i];
    if (l>-1){
      ndofn = give_ndofn (i);
      for (j=0;j<ndofn;j++){
	k = give_dof (i,j);
	if (k==1){
	  av[l][j]=acn;
	  acn++;
	}
	else{
	  av[l][j]=k;
	}
      }
    }
  }
  
  for (i=0;i<nn;i++){
    l=stop->ltg1[i];
    if (l>-1){
      ndofn = give_ndofn (i);
      for (j=0;j<ndofn;j++){
	save_dof (i,j,av[l][j]);
      }
    }
  }
  
  for (i=0;i<max;i++){
    delete [] av[i];
  }
  delete [] av;

  fprintf (out,"\n\n kontrola indikatoru kodovych cisel \n");
  for (i=0;i<nn;i++){
    fprintf (out,"\n node %6ld",i+1);
    ndofn = give_ndofn (i);
    for (j=0;j<ndofn;j++){
      fprintf (out,"  %ld",give_dof (i,j));
    }
  }

}
*/

void gtopology::edges_auxprint (FILE *out)
{
  long i,j,a;
  
  fprintf (out,"\n\n\n kontrola pole nlist na jednotlivych hranach \n");
  for (i=0;i<nged;i++){
    fprintf (out,"\n %6ld %6ld       %6ld %6ld",gedges[i].nlist[0]+1,gedges[i].nlist[1]+1,gedges[i].nlist[2]+1,gedges[i].nlist[3]+1);
  }
  
  
  
  fprintf (out,"\n\n\n kontrola usporadni hran \n");
  for (i=0;i<nged;i++){
    fprintf (out,"\n %6ld       %6ld",gedges[i].nlist[0]+1,gedges[i].nlist[1]+1);
  }
  i=nged-1;
  if (i>=0)
    fprintf (out,"\n %6ld       %6ld",gedges[i].nlist[2]+1,gedges[i].nlist[3]+1);
  
  
  
  fprintf (out,"\n\n\n kontrola prev a next \n");
  for (i=0;i<nged;i++){
    fprintf (out,"\n hrana %6ld   prev %6ld  next %6ld",i,gedges[i].prev+1,gedges[i].next+1);
  }
  
  
  fprintf (out,"\n\n\n kontrola souslednosti \n");
  for (i=0;i<nser;i++){
    fprintf (out,"\n\n serie %5ld",i);
    for (j=0;j<nedser[i];j++){
      a=edgelist[i][j];
      fprintf (out,"\n %5ld  hrana %5ld   %5ld %5ld     %lf %lf %lf      %lf %lf %lf",
	       j,a,gedges[a].fn,gedges[a].ln,
	       gedges[a].dv[0],gedges[a].dv[1],gedges[a].dv[2],
	       gedges[a].nv[0],gedges[a].nv[1],gedges[a].nv[2]);
    }
  }
  fprintf (out,"\n\n");

  fprintf (out,"\n\n\n kontrola prilehlych prvku \n");
  for (i=0;i<nser;i++){
    fprintf (out,"\n\n serie %5ld",i);
    for (j=0;j<nedser[i];j++){
      a=edgelist[i][j];
      fprintf (out,"\n hrana %5ld  %6ld %6ld",j,gedges[a].adjel[0],gedges[a].adjel[1]);
    }
  }
  fprintf (out,"\n\n");
  
  
  /*
  fprintf (out,"\n\n\n kontrola souslednosti \n");
  long *av;
  av = new long [Gtm->nged];
  
  for (i=0;i<Gtm->nged;i++){
    av[i]=-1;
  }
  for (i=0;i<Gtm->nged;i++){
    if (Gtm->gedges[i].prev==-1){
      a=i;
      n=Gtm->gedges[i].next;
      
      fprintf (out,"\n %6ld -> %6ld   %lf %lf %lf      %lf %lf %lf",a,n,Gtm->gedges[a].dv[0],Gtm->gedges[a].dv[1],Gtm->gedges[a].dv[2],Gtm->gedges[a].nv[0],Gtm->gedges[a].nv[1],Gtm->gedges[a].nv[2]);
      fprintf (out,"\n %6ld %6ld",Gtm->gedges[a].nlist[0],Gtm->gedges[a].nlist[2]);
      //fprintf (Out,"\n %6ld",Gtm->gedges[a].nlist[0]);
      
      av[i]=1;
      break;
    }
  }
  long ui=Gtm->nged;
  
  for (i=1;i<ui;i++){
    a=n;
    n=Gtm->gedges[n].next;
    av[a]=1;
    fprintf (out,"\n %6ld",Gtm->gedges[a].nlist[0]);
    //fprintf (Out,"\n %6ld %6ld",Gtm->gedges[a].nlist[0],Gtm->gedges[a].nlist[2]);
    fprintf (out,"\n %6ld -> %6ld   %lf %lf %lf      %lf %lf %lf",a,n,Gtm->gedges[a].dv[0],Gtm->gedges[a].dv[1],Gtm->gedges[a].dv[2],Gtm->gedges[a].nv[0],Gtm->gedges[a].nv[1],Gtm->gedges[a].nv[2]);
    if (n<0){
      for (j=0;j<Gtm->nged;j++){
	if (Gtm->gedges[j].prev==-1 && av[j]==-1){
	  a=j;
	  n=Gtm->gedges[j].next;
	  av[j]=1;
	  ui--;
	  fprintf (out,"\n %6ld",Gtm->gedges[a].nlist[0]);
	  fprintf (Out,"\n %6ld -> %6ld   %lf %lf %lf      %lf %lf %lf",a,n,Gtm->gedges[a].dv[0],Gtm->gedges[a].dv[1],Gtm->gedges[a].dv[2],Gtm->gedges[a].nv[0],Gtm->gedges[a].nv[1],Gtm->gedges[a].nv[2]);
	  break;
	}
      }
    }
  }
  fprintf (out,"\n %6ld",Gtm->gedges[a].nlist[2]);
  */
  
  /*
  fprintf (Out,"\n");


  for (i=0;i<Gtm->nged;i++){
    av[i]=-1;
  }
  for (i=0;i<Gtm->nged;i++){
    if (Gtm->gedges[i].prev==-1){
      a=i;
      n=Gtm->gedges[i].next;

      //fprintf (Out,"\n %6ld -> %6ld   %lf %lf %lf      %lf %lf %lf",a,n,Gtm->gedges[a].dv[0],Gtm->gedges[a].dv[1],Gtm->gedges[a].dv[2],Gtm->gedges[a].nv[0],Gtm->gedges[a].nv[1],Gtm->gedges[a].nv[2]);
      //fprintf (Out,"\n %6ld %6ld ",Gtm->gedges[a].nlist[1]-205,Gtm->gedges[a].nlist[3]-205);
      fprintf (Out,"\n %6ld",Gtm->gedges[a].nlist[1]-205);

      av[i]=1;
      break;
    }
  }
  ui=Gtm->nged;
  
  for (i=1;i<ui;i++){
    a=n;
    n=Gtm->gedges[n].next;
    av[a]=1;
    fprintf (Out,"\n %6ld",Gtm->gedges[a].nlist[1]-205);
    //fprintf (Out,"\n %6ld -> %6ld   %lf %lf %lf      %lf %lf %lf",a,n,Gtm->gedges[a].dv[0],Gtm->gedges[a].dv[1],Gtm->gedges[a].dv[2],Gtm->gedges[a].nv[0],Gtm->gedges[a].nv[1],Gtm->gedges[a].nv[2]);
    if (n<0){
      for (j=0;j<Gtm->nged;j++){
	if (Gtm->gedges[j].prev==-1 && av[j]==-1){
	  a=j;
	  n=Gtm->gedges[j].next;
	  av[j]=1;
	  ui--;
	  fprintf (Out,"\n %6ld",Gtm->gedges[a].nlist[1]-205);
	  //fprintf (Out,"\n %6ld -> %6ld   %lf %lf %lf      %lf %lf %lf",a,n,Gtm->gedges[a].dv[0],Gtm->gedges[a].dv[1],Gtm->gedges[a].dv[2],Gtm->gedges[a].nv[0],Gtm->gedges[a].nv[1],Gtm->gedges[a].nv[2]);
	  break;
	}
      }
    }
  }
  fprintf (Out,"\n %6ld",Gtm->gedges[a].nlist[3]-205);
  */


  
  /*
  fprintf (Out,"\n");


  for (i=0;i<Gtm->nged;i++){
    av[i]=-1;
  }
  for (i=0;i<Gtm->nged;i++){
    if (Gtm->gedges[i].prev==-1){
      a=i;
      n=Gtm->gedges[i].next;

      fprintf (Out,"\n %lf %lf   %lf %lf ",Gtm->gedges[a].nv[0],Gtm->gedges[a].nv[1],Gtm->gedges[a].dv[0],Gtm->gedges[a].dv[1]);
      //fprintf (Out,"\n %6ld -> %6ld   %lf %lf %lf      %lf %lf %lf",a,n,Gtm->gedges[a].dv[0],Gtm->gedges[a].dv[1],Gtm->gedges[a].dv[2],Gtm->gedges[a].nv[0],Gtm->gedges[a].nv[1],Gtm->gedges[a].nv[2]);
      //fprintf (Out,"\n %6ld %6ld ",Gtm->gedges[a].nlist[1]-205,Gtm->gedges[a].nlist[3]-205);

      av[i]=1;
      break;
    }
  }
  ui=Gtm->nged;
  
  for (i=1;i<ui;i++){
    a=n;
    n=Gtm->gedges[n].next;
    av[a]=1;
    //fprintf (Out,"\n %6ld %6ld",Gtm->gedges[a].nlist[1]-205,Gtm->gedges[a].nlist[3]-205);
    //fprintf (Out,"\n %6ld -> %6ld   %lf %lf %lf      %lf %lf %lf",a,n,Gtm->gedges[a].dv[0],Gtm->gedges[a].dv[1],Gtm->gedges[a].dv[2],Gtm->gedges[a].nv[0],Gtm->gedges[a].nv[1],Gtm->gedges[a].nv[2]);
    fprintf (Out,"\n %lf %lf   %lf %lf",Gtm->gedges[a].nv[0],Gtm->gedges[a].nv[1],Gtm->gedges[a].dv[0],Gtm->gedges[a].dv[1]);
    if (n<0){
      for (j=0;j<Gtm->nged;j++){
	if (Gtm->gedges[j].prev==-1 && av[j]==-1){
	  a=j;
	  n=Gtm->gedges[j].next;
	  av[j]=1;
	  ui--;
	  fprintf (Out,"\n %6ld %6ld",Gtm->gedges[a].nlist[1]-205,Gtm->gedges[a].nlist[3]-205);
	  //fprintf (Out,"\n %6ld -> %6ld   %lf %lf %lf      %lf %lf %lf",a,n,Gtm->gedges[a].dv[0],Gtm->gedges[a].dv[1],Gtm->gedges[a].dv[2],Gtm->gedges[a].nv[0],Gtm->gedges[a].nv[1],Gtm->gedges[a].nv[2]);
	  fprintf (Out,"\n %lf %lf   %lf %lf",Gtm->gedges[a].nv[0],Gtm->gedges[a].nv[1],Gtm->gedges[a].dv[0],Gtm->gedges[a].dv[1]);
	  break;
	}
      }
    }
  }

  */
  
  /*
  a=0; n=Gtm->gedges[a].next;
  fprintf (out,"\n\n\n kontrola serie \n");
  for (i=0;i<Gtm->nged;i++){
    fprintf (out,"\n%ld -> %ld",a,n);
    a=n;
    n=Gtm->gedges[a].next;
  }
  */
  
  /*
  fprintf (out,"\n\n kontrola souslednosti\n");
  long *av;
  av = new long [Gtm->nged];
  
  for (i=0;i<Gtm->nged;i++){
    av[i]=-1;
  }
  for (i=0;i<Gtm->nged;i++){
    if (Gtm->gedges[i].prev==-1){
      a=i;
      n=Gtm->gedges[i].next;
      
      fprintf (Out,"\n %6ld %6ld     %6ld %6ld",Gtm->gedges[a].nlist[0],Gtm->gedges[a].nlist[2],Gtm->gedges[a].nlist[1],Gtm->gedges[a].nlist[3]);
      fprintf (Out,"         %lf %lf %lf      %lf %lf %lf",Gtm->gedges[a].dv[0],Gtm->gedges[a].dv[1],Gtm->gedges[a].dv[2],Gtm->gedges[a].nv[0],Gtm->gedges[a].nv[1],Gtm->gedges[a].nv[2]);
      
      av[i]=1;
      break;
    }
  }
  long ui=Gtm->nged;
  
  for (i=1;i<ui;i++){
    a=n;
    n=Gtm->gedges[n].next;
    av[a]=1;
    fprintf (Out,"\n %6ld %6ld     %6ld %6ld",Gtm->gedges[a].nlist[0],Gtm->gedges[a].nlist[2],Gtm->gedges[a].nlist[1],Gtm->gedges[a].nlist[3]);
    fprintf (Out,"         %lf %lf %lf      %lf %lf %lf",Gtm->gedges[a].dv[0],Gtm->gedges[a].dv[1],Gtm->gedges[a].dv[2],Gtm->gedges[a].nv[0],Gtm->gedges[a].nv[1],Gtm->gedges[a].nv[2]);
    if (n<0){
      for (j=0;j<Gtm->nged;j++){
	if (Gtm->gedges[j].prev==-1 && av[j]==-1){
	  a=j;
	  n=Gtm->gedges[j].next;
	  av[j]=1;
	  ui--;
	  fprintf (Out,"\n %6ld %6ld     %6ld %6ld",Gtm->gedges[a].nlist[0],Gtm->gedges[a].nlist[2],Gtm->gedges[a].nlist[1],Gtm->gedges[a].nlist[3]);
	  fprintf (Out,"         %lf %lf %lf      %lf %lf %lf",Gtm->gedges[a].dv[0],Gtm->gedges[a].dv[1],Gtm->gedges[a].dv[2],Gtm->gedges[a].nv[0],Gtm->gedges[a].nv[1],Gtm->gedges[a].nv[2]);
	  break;
	}
      }
    }
  }
  

  fprintf (Out,"\n\n\n");

  */



  /*
  for (i=0;i<Gtm->nged;i++){
    fprintf (out,"\n\n hrana cislo  %ld",i);
    Gtm->gedges[i].print (out);
  }
  */
  
  /*
  fprintf (out,"\n\n\n kontrola serii \n");
  for (i=0;i<Gtm->nged;i++){
    fprintf (out,"\n edge %6ld je v serii  %6ld",i,Gtm->edgeser[i]);
  }
  */

}


/**
   function reads informations necessary for sequential FETI method
   problem is solved by FETI method on one processor
   
   @param ns - number of subdomains
   @param in - input file
   
   JK, 6.10.2007
*/
/*
void gtopology::read_feti (long ns,XFILE *in)
{
  meshdescription md;
  
  //  type of mesh description - see GEFEL/galias.h
  xfscanf (in,"%k%m","meshdescript",&meshdescription_kwdset,(int*)&md);
  
  //  instance of the class seqtop (sequential topology)
  stop = new seqtop (ns,md);
  
  //  number of nodes on subdomains
  stop->read_nnsd (in);
  
  //  reading of auxiliary array ltg (local to global map)
  stop->read_ltg (in);
}
*/
/**
   function reads array ltg1
   
   @param in - input file
   
   JK, 22.11.2007
*/
/*
void gtopology::read_ltg1 (XFILE *in)
{
  //  instance of the class seqtop (sequential topology)
  //  ns=1 - number of subdomains
  //  md=1 - all_nodes
  stop = new seqtop (1,all_nodes);
  
  //  reading of auxiliary array ltg1 (local to global map)
  stop->readltg1 (nn,in);
}
*/



/**
   function allocates arrays cn on edges
   cn contains code numbers of Lagrange multipliers
   
   JK, 6.1.2008
*/
void gtopology::alloc_edge_cn ()
{
  long i,nid,ndofnfn,ndofnln;
  
  for (i=0;i<nged;i++){
    //  first node on edge
    nid=gedges[i].fn;
    //  number of DOFs on the first node
    ndofnfn=give_ndofn (nid);
    //  last node on edge
    nid=gedges[i].ln;
    //  number of DOFs on the last node
    ndofnln=give_ndofn (nid);
    
    //  allocation of arrays cnfn and cnln
    gedges[i].alloc_cn (ndofnfn,ndofnln);
  }

}


/**
   function localizes contributions from edges to the global %matrix
   
   @param gm - pointer to the general %matrix
   
   JK, 8.8.2008
*/
void gtopology::edge_localization (gmatrix *gm)
{
  long i,nm,*ncn1,*ncn2,*mcn;

  for (i=0;i<nged;i++){
    //  number of multipliers between first nodes
    nm=gedges[i].ndofnf;
    ncn1 = new long [nm];
    ncn2 = new long [nm];
    mcn = new long [nm];

    //  code numbers on first nodes and appropriate multipliers
    give_edge_code_numbers (i,1,ncn1,ncn2,mcn);
    
    //  localization of contributions from edges
    gm->mult_localize (nm,ncn1,ncn2,mcn);
    
    delete [] mcn;
    delete [] ncn2;
    delete [] ncn1;
    
    
    //  number of multipliers between second nodes
    nm=gedges[i].ndofnl;
    ncn1 = new long [nm];
    ncn2 = new long [nm];
    mcn = new long [nm];

    //  code numbers on first nodes and appropriate multipliers
    give_edge_code_numbers (i,2,ncn1,ncn2,mcn);
    
    //  localization of contributions from edges
    gm->mult_localize (nm,ncn1,ncn2,mcn);
    
    delete [] mcn;
    delete [] ncn2;
    delete [] ncn1;
  }

}






















/**
   function automatically detects subdomains in conforming meshes
   
   it is used e.g. in problems with hemivariational inequalities
   
   @param out - output file
   
   JK, 12.11.2007
*/
void gtopology::auto_subdom_detection (FILE *out)
{
  //  list of adjacent nodes to each node
  adjacnodes (out);
  
  // ***********************
  //  contact end points  **
  // ***********************
  end_nodes (out);

  enodes_allelem (out);

  endnodes_auxprint (out);
  
  
  // ******************
  //  contact edges  **
  // ******************
  //  assembling of interface edges - edges on contact between domains
  //  list of nodes on edges is created
  edges (out);
  //  edge nodes are sorted
  edgenode_sorting ();
  //  number of previous and next edges (index of previous and next edges)
  prev_next_edges ();
  //  assembling of edge series
  edge_series (out);
  //  edge-element correspondence
  edge_elem (out);


  //  edge-all elements correspondence
  edge_allelem (out);

  //  computation of direction/tangential vectors
  edge_dirvect ();
  //  computation of normal vectors
  edge_normvect ();
  //  check of normal vector orientation
  normvectorient ();
  
  
  
  domdecomp ();
  //  creation of array ltg (local to global ordering)
  //  ltg array is used for Schur ordering of unknowns
  create_ltg (out);

  edges_auxprint (out);


  //  code numbers are generated with respect to saddle point problems
  cngen=3;

}

/**
   function generates code numbers of Lagrange multipliers
   
   JK, 6.1.2008
*/
long gtopology::codenum_multip ()
{
  long i,j,k,l,ii,nid,nm,ndofn;
  long *cn;
  
  //  auxiliary variable
  nm=ndof;

  // ***********************************
  //  multipliers defined on end nodes
  // ***********************************
  
  //  allocation of array cnm on end nodes
  alloc_endnode_cn ();
  
  for (i=0;i<nen;i++){
    //  first node id
    nid=endnodes[i].fn;
    //  number of DOFs on the node
    ndofn=give_ndofn (nid);
    cn = new long [ndofn];
    for (l=0;l<ndofn;l++){
      ii=give_dof (nid,l);
      if (ii!=0){
	cn[l]=nm;
	nm++;
      }
      else
	cn[l]=0;
    }
    for (l=0;l<ndofn;l++){
      endnodes[i].cnm[l]=cn[l];
    }
    delete [] cn;
  }
  
  // *******************************
  //  multipliers defined on edges
  // *******************************
  
  //  allocation of arrays cn on edges
  alloc_edge_cn ();
  
  for (i=0;i<nser;i++){
    //  loop over series of edges
    for (j=0;j<nedser[i];j++){
      //  loop over all edges in the serie
      
      //  number of actual edge
      k=edgelist[i][j];
      
      // **************
      //  first nodes
      // **************
      if (j==0){
	//  node id of the first node on the edge
	nid=gedges[k].fn;
	//  number of DOFs on the node
	ndofn=give_ndofn (nid);
	cn = new long [ndofn];
	for (l=0;l<ndofn;l++){
	  ii=give_dof (nid,l);
	  if (ii!=0){
	    cn[l]=nm;
	    nm++;
	  }
	  else
	    cn[l]=0;
	}
      }
      
      for (l=0;l<ndofn;l++){
	gedges[k].cnfn[l]=cn[l];
      }
      delete [] cn;
      
      // *************
      //  last nodes
      // *************
      //  node id of the first node on the edge
      nid=gedges[k].ln;
      //  number of DOFs on the node
      ndofn=give_ndofn (nid);
      cn = new long [ndofn];
      for (l=0;l<ndofn;l++){
	ii=give_dof (nid,l);
	if (ii!=0){
	  cn[l]=nm;
	  nm++;
	}
	else
	  cn[l]=0;
      }
      
      for (l=0;l<ndofn;l++){
	gedges[k].cnln[l]=cn[l];
      }
      
      if (j==nedser[i]-1)
	delete [] cn;
    }
  }
  
  return nm;
}

/**
   function localizes %matrix entries connected with
   Lagrange multipliers to the general %matrix
   there are contributions from end nodes (1D),
   edges (2D)
   
   @param gm - general %matrix
   
   JK, 14.10.2008
*/
void gtopology::mult_localization (gmatrix *gm)
{
  //  contributions from end nodes
  endnodes_localization (gm);
  //  contributions from edges
  edge_localization (gm);
}

/**
   function creates lists of nodes in subdomains
   mesh must be described by the glued ordering
   
   1.6.2008, JK
*/
void gtopology::domdecomp ()
{
  long i,j,ii,jj,nvn,nan,nnf,nnnf,anode;
  long *aux,*oldlist,*newlist;
  
  aux = new long [nn];
  for (i=0;i<nn;i++){
    aux[i]=-1;
  }
  
  //  number of subdomain
  ns=0;
  //  number of visited nodes
  nvn=0;
  do{
    for (i=0;i<nn;i++){
      if (aux[i]==-1)
	anode=i;
    }
    
    oldlist=new long[1];
    oldlist[0]=anode;
    nnf=1;
    
    do{
      //  number of nodes on a new front
      nnnf=0;
      //  loop over nodes on actual front
      //  nnf - number of nodes on actual front
      for (i=0;i<nnf;i++){
	//  number of actual node on the front
	ii=oldlist[i];
	//  number of adjacent nodes to actual node
	nan=nadjnodnod[ii];
	//  loop over adjacent nodes to actual node
	for (j=0;j<nan;j++){
	  //  number of adjacent node
	  jj=adjnodnod[ii][j];
	  if (aux[jj]==-1){
	    //  node has not been touched yet
	    nnnf++;
	  }
	}
      }
      //  list of nodes on new front
      newlist=new long [nnnf];
      //  number of nodes on a new front
      nnnf=0;
      //  loop over nodes on actual front
      //  nnf - number of nodes on actual front
      for (i=0;i<nnf;i++){
	//  number of actual node on the front
	ii=oldlist[i];
	//  number of adjacent nodes to actual node
	nan=nadjnodnod[ii];
	//  loop over adjacent nodes to actual node
	for (j=0;j<nan;j++){
	  //  number of adjacent node
	  jj=adjnodnod[ii][j];
	  if (aux[jj]==-1){
	    //  node has not been touched yet
	    newlist[nnnf]=jj;
	    nnnf++;
	    aux[jj]=ns;
	    nvn++;
	  }
	}
      }
      delete [] oldlist;
      oldlist=new long [nnnf];
      for (i=0;i<nnnf;i++){
	oldlist[i]=newlist[i];
      }
      delete [] newlist;
      nnf=nnnf;
    }while(nnf>0);

    delete [] oldlist;
    
    ns++;
  }while(nvn<nn);
  
  
  //  instance of the class seqtop (sequential topology)
  stop = new seqtop (ns,all_nodes);
  
  //  allocation 
  stop->nnsd = new long [ns];
  for (i=0;i<ns;i++){
    stop->nnsd[i]=0;
  }
  
  for (i=0;i<nn;i++){
    stop->nnsd[aux[i]]++;
  }
  
  delete [] aux;
}

/**
   function assembles array ltg for sequential topology
   array is constructed from the knowledge of interfaces
   among subdomains
   function is intended for hemivariational inequalities
   
   JK, 6.11.2007
*/
void gtopology::create_ltg (FILE *out)
{
  long i,j,k,nin;
  long *aux,*v;

  aux = new long [nn];
  for (i=0;i<nn;i++){
    aux[i]=nn;
  }
  
  //  auxiliary array assembling
  //  aux[i]=j - the i-th node has label j
  //  if j=nn - noninterface node
  //  if j!=nn - interface node, j is equal to minimum number of node connected to this interface node
  //
  
  //  loop over end nodes
  for (i=0;i<nen;i++){
    k=endnodes[i].fn;
    j=endnodes[i].ln;
    if (j<k){
      if (aux[j]>j)
	aux[j]=j;
      if (aux[k]>j)
	aux[k]=j;

      if (aux[j]<aux[k])
	aux[k]=aux[j];
      else
	aux[j]=aux[k];
    }
    else{
      if (aux[j]>k)
	aux[j]=k;
      if (aux[k]>k)
	aux[k]=k;

      if (aux[j]<aux[k])
	aux[k]=aux[j];
      else
	aux[j]=aux[k];
    }
    
  }

  //  loop over general edges
  for (i=0;i<nged;i++){
    j=gedges[i].nlist[0];
    k=gedges[i].nlist[1];
    if (j<k){
      if (aux[j]>j)
	aux[j]=j;
      if (aux[k]>j)
	aux[k]=j;

      if (aux[j]<aux[k])
	aux[k]=aux[j];
      else
	aux[j]=aux[k];
    }
    else{
      if (aux[j]>k)
	aux[j]=k;
      if (aux[k]>k)
	aux[k]=k;

      if (aux[j]<aux[k])
	aux[k]=aux[j];
      else
	aux[j]=aux[k];
    }
    
    j=gedges[i].nlist[2];
    k=gedges[i].nlist[3];
    if (j<k){
      if (aux[j]>j)
	aux[j]=j;
      if (aux[k]>j)
	aux[k]=j;

      if (aux[j]<aux[k])
	aux[k]=aux[j];
      else
	aux[j]=aux[k];
    }
    else{
      if (aux[j]>k)
	aux[j]=k;
      if (aux[k]>k)
	aux[k]=k;

      if (aux[j]<aux[k])
	aux[k]=aux[j];
      else
	aux[j]=aux[k];
    }

  }
  
  
  
  v = new long [nn];
  for (i=0;i<nn;i++){
    v[i]=-1;
  }
  
  
  for (i=0;i<nn;i++){
    j=aux[i];
    if (j!=nn)
      v[j]++;
  }
  
  //  generation of interface node numbers
  //  nin - number of interface nodes
  nin=0;
  for (i=0;i<nn;i++){
    if (v[i]>-1){
      v[i]=nin;
      nin++;
    }
  }
  
  for (i=0;i<nn;i++){
    j=aux[i];
    if (j!=nn)
      aux[i]=v[j];
    else
      aux[i]=-1;
  }

  
  //
  //  array ltg
  //
  if (stop->ltg==NULL){
    stop->ltg = new long* [stop->ns];
    for (i=0;i<stop->ns;i++){
      stop->ltg[i] = new long [stop->nnsd[i]];
    }
  }
  
  k=0;
  for (i=0;i<stop->ns;i++){
    for (j=0;j<stop->nnsd[i];j++){
      stop->ltg[i][j]=aux[k];
      k++;
    }
  }
  
  
  fprintf (out,"\n\n\n funkce seqtop::create_ltg   pole ltg");
  for (i=0;i<stop->ns;i++){
    fprintf (out,"\n");
    for (j=0;j<stop->nnsd[i];j++){
      fprintf (out,"%ld  ",stop->ltg[i][j]);
    }
  }
  fprintf (out,"\n");


  delete [] aux;
  delete [] v;
  
}

/**
   function allocates object of the class sequential topology
   function read data about meshes

   this function is used in sequential (one-processor) implementation
   of domain decomposition methods

   @param ns - number of subdomains, it is read in the part which
               deals with solvers
   @param in - input file
   
   JK, 23.12.2007
*/
void gtopology::read_seq_top (XFILE *in)
{
  meshdescription md;
  
  //  type of mesh description - see GEFEL/galias.h
  //  md=1 - all_nodes
  //  md=2 - boundary nodes
  //  md=11 - metis
  xfscanf (in,"%k%m","meshdescript",&meshdescription_kwdset,(int*)&md);
  
  //  number of subdomains/aggregates
  xfscanf (in,"%ld",&ns);
  
  //  instance of the class seqtop (sequential topology)
  stop = new seqtop (ns,md);
  
  //  number of nodes on subdomains
  stop->read_nnsd (in);
  
  //  reading of auxiliary array ltg (local to global map)
  stop->read_ltg (in);
}


/**
   function allocates object of the class sequential topology
   function read data about meshes

   this function is used in sequential (one-processor) implementation
   of domain decomposition methods

   @param ns - number of subdomains, it is read in the part which
               deals with solvers
   @param in - input file
   
   JK, 23.12.2007
*/
void gtopology::read_loadbal_top (XFILE *in)
{
  meshdescription md;
  
  //  type of mesh description - see GEFEL/galias.h
  //  md=1 - all_nodes
  //  md=2 - boundary nodes
  //  md=11 - metis
  xfscanf (in,"%k%m","meshdescript",&meshdescription_kwdset,(int*)&md);
  
  ns=1;
  
  //  instance of the class seqtop (sequential topology)
  stop = new seqtop (ns,md);
  
  //  number of nodes on subdomains
  stop->read_nnsd (in);
  
  //  reading of auxiliary array ltg (local to global map)
  stop->read_loadbal_ltg (in);
}




















/**
   function reorders node numbering with respect to Cuthill-McKee algorithm
   
   JB, 27.2.2007
   changed JB, 24.10.2008
*/

void gtopology::cuthill_mckee_renumb (FILE *out)
{
  long i,j,k;
  long minnode,n,a,b,m,stop;
  int startnode,endnode;
  long *cmk,**auxadjnodnod,*auxnadjnodnod,*aux; 
  
  //fprintf (stdout,"\n\n\n CUTHILL MCKEE\n\n");
  // nalezeni sousednich uzlu k uzlu a jejich poctu
  adjacnodes (out);
  
  // pole precislovanych uzlu a zaroven navstivenych uzlu
  cmk = new long [nn];
  for(i = 0; i<nn; i++)  cmk[i] = 0;
  
  
  findPseudoPheripheral(&startnode,&endnode);
  minnode = long(startnode);
  //fprintf (stdout,"\n\n\nminnode = %ld\n\n",minnode+1);
  
  // nalezni minima v nadjnodnod pro prvni uzel
  //min = LONG_MAX;
  //for(i = 0; i<nn; i++){
    //fprintf (stdout,"%ld %ld\n",i+1, nadjnodnod[i]);
    //if(min >= nadjnodnod[i]){
  //min = nadjnodnod[i];
  //  minnode = i;
  //}
  //}
  //fprintf (stdout,"\n\n\nminnode = %ld\n\n",minnode); 
  
  
  //precislovani prvniho node - cislovani od 1
  cmk[minnode] = 1;
  
  //serazeni pole adjnodnod podle velikosti sousedu
  // plneni pomocneho pole
  // pomocne pole
  auxadjnodnod = new long* [nn];
  auxnadjnodnod = new long [nn];
  
  for(i = 0; i < nn; i++){
    auxnadjnodnod[i] = nadjnodnod[i]-1;
  }
  
  for(i = 0; i < nn; i++){
    auxadjnodnod[i] = new long [auxnadjnodnod[i]];
    k = 0;
    //fprintf (stdout,"%ld  %ld -> ",i,auxnadjnodnod[i]);
    for(j = 0; j < nadjnodnod[i]; j++){
      if( i != adjnodnod[i][j]){
	auxadjnodnod[i][k] =  adjnodnod[i][j];
	//fprintf (stdout,"%ld-%ld   ",auxadjnodnod[i][k],auxnadjnodnod[auxadjnodnod[i][k]]);
	k++;
      }
    }
    //fprintf (stdout,"\n");
    
    //buble sorting pole auxadjnodnod podle velikosti sousedu
    for (m =auxnadjnodnod[i] - 1; m >= 0;  m--){
      for (j = 1; j <= m; j++){
	a = auxadjnodnod[i][j];
	b = auxadjnodnod[i][j-1];
	if (auxnadjnodnod[b] > auxnadjnodnod[a]){
	  n = auxadjnodnod[i][j-1];
	  auxadjnodnod[i][j-1] = auxadjnodnod[i][j];
	  auxadjnodnod[i][j] = n;
	}
      }
    }
    
    // for(j = 0; j < auxnadjnodnod[i]; j++){
    // fprintf (stdout,"      %ld-%ld",auxadjnodnod[i][j],auxnadjnodnod[auxadjnodnod[i][j]]);
    //}
    //fprintf (stdout,"\n");
  }

  
  aux = new long[nn];
  for(i = 0; i < nn; i ++){
    aux[i]=-1;
  }
  cmk = new long[nn];
  startnode = 0;
  aux[minnode] = 0;
  stop = 0;
  m = 1;
  for(k = 0; k < nn; k++){
    //fprintf (stdout,"startnode-%ld\n",startnode+1);
    for(i = 0; i < nn; i ++){ 
      if(aux[i] == startnode){
	//fprintf (stdout,"jsem na %ld\n",i+1);
	for(j = 0; j < auxnadjnodnod[i];j++){
	  if( aux[auxadjnodnod[i][j]] == -1){
	    cmk[auxadjnodnod[i][j]] = m;
	    aux[auxadjnodnod[i][j]] = m;
	    //fprintf (stdout,"%ld-%ld\n",auxadjnodnod[i][j]+1,m+1);
	    m++;
	  }
	  if(m == nn){
	    stop = 1;
	  }
	}
	break;
      }
    }
    if(stop == 1){
      break;
    }
    startnode++;
  }


  
  
  // mazani pomocneho pole 
  for(i = 0; i < nn; i++) delete[] auxadjnodnod[i];
  delete []auxadjnodnod;
  delete []auxnadjnodnod;
  delete []aux;
  
  //  cmk[i]=j - the i-th node of the original ordering has number j in the new ordering
  
  if (nodren == cuthill_mckee){
    for (i=0;i<nn;i++){
      ordering[cmk[i]]=i;
    }
  }

  if (nodren == rev_cuthill_mckee){
    for (i=0;i<nn;i++){
      ordering[cmk[i]]=i;
    }
    j=nn-1;
    for (i=0;i<nn;i++){
      cmk[j]=ordering[i];
      j--;
    }
    for (i=0;i<nn;i++){
      ordering[i]=cmk[i];
    }
    
  }
  
  //fprintf (stdout,"Kontrolni tisk ordering\n"); 
  //for(i = 0; i < nn; i++)  fprintf (stdout,"%ld %ld\n",i+1,ordering[i]+1);

}





void gtopology::writePriorityQueue(int *fronta){
  for (int i = 0; i < (nn); i++){
    //printf( "%d. hodnota %d\n", i, fronta[i]);
  }
}

void gtopology::shell_sort( int *array, int arrayLength)
{
  int flag = 1, d = arrayLength, i, temp;
  while( flag || (d>1))      // boolean flag (true when not equal to 0)
    {
      flag = 0;           // reset flag to 0 to check for future swaps
      d = (d+1) / 2;
      for (i = 0; i < (arrayLength - d); i++)
	{
	  if (nadjnodnod[array[i + d]] < nadjnodnod[array[i]])
	    {
	      temp = array[i + d];      // swap items at positions i+d and d
	      array[i + d] = array[i];
	      array[i] = temp;
	      flag = 1;                  // indicate that a swap has occurred
	    }
	}
    }
  return;
}



void gtopology::lastLevelOfRootedStructure(int start, int *velikost, int *maxSirka, int *hloubka, int *vzdalenost, int **lastLevel){
  int *aktUzlyStart;
  int i,j,k; // pomocne promenne cyklu
  aktUzlyStart = new int[nn];
  for(i = 0; i< nn; i++){	
    aktUzlyStart[i] = 0;
  }
  int pocetObjevenych = 0; // pocet neaktivnich uzlu nalezenych v okoli aktualne zpracovaneho uzlu
  int *hranice;
  int velikostHranice;
  int aktualniUzel,pomUzel,posl;
  
  *maxSirka = -1;
  *hloubka = 0;
  aktUzlyStart[start]=1;
  velikostHranice = 1;
  hranice = new int[1];
  hranice[0] = start;
  int *pomocnePole;
  pomocnePole = new int[nn];
  do{
    
    
    //vynulujeme pomocne pole
    for( j=0; j<nn; j++){
      pomocnePole[j] = 0;
      
    }
    posl =0;
    pocetObjevenych = 0;
    //projdeme vsechny prvky na hranici
    for( j=0; j<velikostHranice; j++){
      
      aktualniUzel = hranice[j];
      //printf("Zpracovavam uzel %d\n",aktualniUzel);
      //projdem vsechny neaktivni uzly na hranici a pridame je do hranice
      for(i = 0; i< nadjnodnod[aktualniUzel]; i++){
	pomUzel = adjnodnod[aktualniUzel][i];
	vzdalenost[aktualniUzel] = *hloubka;
	//printf("\tZpracovavam souseda %d ",pomUzel);
	if(aktUzlyStart[pomUzel] == 0){
	  //printf("platny uzel");
	  
	  aktUzlyStart[pomUzel] = 1;
	  pomocnePole[posl] = pomUzel;
	  pocetObjevenych++;
	  posl++;
	  
	}
	//printf("\n");
      }
    }
    if(pocetObjevenych > *maxSirka){
      *maxSirka = pocetObjevenych;
    }
    (*hloubka)++; // v kazdem pruchodu se hloubka zvetsuje
    //getchar();
    //zkopirujeme pomocne pole do hranice
    if(pocetObjevenych != 0){
      delete[] hranice;
      hranice = new int[posl + 1];
      for( k=0; k<posl; k++){
	hranice[k]= pomocnePole[k];
      }
      velikostHranice = posl ;
      //printf("Pocet objevenych %d " , pocetObjevenych);
    }
  }while(pocetObjevenych != 0);
  //
  //for(k=0; k<velikostHranice; k++){
  //		printf("Prvek %d a stupen %d\n" , hranice[k],nadjnodnod[hranice[k]]);
  //	}
  ///printf("velikostHranice =- %d", velikostHranice);
  if(*lastLevel != NULL){
	  delete [] *lastLevel;
  }
    *lastLevel = hranice;
    *velikost = velikostHranice;
    
}

void gtopology::generateInitialList(int start, int **seznam, int * velikostSeznamu, int *hloubkaSeznamu, int *maxSirka){
  
  //int minUzel;
  int sirka, hloubka,velikost;
  int *lastLevel=NULL;
  int *vzdalenosti;
  
  
  vzdalenosti = new int[nn];
  //int i,j,k; // pomocne promenne cyklu
  // nalezeni sousednich uzlu k uzlu a jejich poctu
  
  lastLevelOfRootedStructure(start,&velikost,&sirka , &hloubka, vzdalenosti , &lastLevel);
  
  
  
  
  shell_sort(lastLevel,velikost);
  removeRepetitiousDegrees(&lastLevel, &velikost);
  
  *velikostSeznamu = velikost;
  *hloubkaSeznamu = hloubka;
  *maxSirka = sirka;
  *seznam = lastLevel ;
}


// funkce najde uzel s minimalnim stupnem
int gtopology::findMinimumDegree(){
  int minUzel,minimum = 99999999;
  for(int i = 0; i< nn; i++){
    
    //printf("MINIMAL Prvek %d a stupen je %d\n" , i,nadjnodnod[i]);
    
    if(nadjnodnod[i]< minimum){
      minUzel = i;
      minimum = nadjnodnod[i];
    }
  }
  return minUzel;
}



void gtopology::removeRepetitiousDegrees(int **pole, int *velikost){
  int aktStupen= -1,k,posl;
  int *pomocnePole;
  int *pracovniPole;
  pracovniPole= *pole;
  pomocnePole = new int[nn];
  //vynulujeme pomocne pole
  for( k=0; k<nn; k++){
    pomocnePole[k] = 0;
    
  }
  posl =0;
  
  // odstranim ostatni stupne
  for( k=0; k<*velikost; k++){
    if(aktStupen != nadjnodnod[pracovniPole[k]]){
      
      aktStupen = nadjnodnod[pracovniPole[k]];
      //printf("akt stupen %d\n",aktStupen);
      pomocnePole[posl] = pracovniPole[k];
      //printf("zmena %d",hranice[k]);
      posl++;
    }
  }
  delete[] pracovniPole;
  pracovniPole = new int[posl + 1];
  for( k=0; k<posl; k++){
    pracovniPole[k]= pomocnePole[k];
  }
  *velikost = posl ;
  *pole = pracovniPole;
  
}

void gtopology::findPseudoPheripheral(int *startUzel, int *cilUzel){
  int *pracovniSeznam;
  int velikostSeznamu;
  int hloubkaSeznamu;
  int maxSirkaSeznamu;
  int start,cil ;
  int aktualniSirka = 99999999;
  int zpracSirka,zpracHloubka,zpracVelikost,*zpracVzdalenosti,*zpracLastLevel;
  bool nalezenCil = false;
  bool novyStart = false;
  zpracLastLevel = NULL; 
  start = findMinimumDegree();
  // tvorim startovni seznam
  //printf("NalezenCil %d, novyStart %d\n", nalezenCil,novyStart);
  zpracVzdalenosti = new int[nn];
  while(!((nalezenCil)&(!novyStart))){
    printf(" %d a %d \n", nalezenCil, novyStart);
    //printf("iterace");
    //printf("Tvorim startovni seznam start %d\n", start);
    generateInitialList(start, &pracovniSeznam, &velikostSeznamu, &hloubkaSeznamu, &maxSirkaSeznamu);
    /*for(int k=0; k<velikostSeznamu; k++){
      printf("Prvek %d  \n" , pracovniSeznam[k]);
      }*/
    ///printf("velikost %d, hloubka %d, sirka %d  \n" , velikostSeznamu, hloubkaSeznamu,maxSirkaSeznamu);
      novyStart = false;
      nalezenCil = false;
      
      for(int k=0; k<velikostSeznamu; k++){
	
	///printf("NalezenCil %d, novyStart %d\n", nalezenCil,novyStart);
	///printf("Zpracovavam uzel %d\n",pracovniSeznam[k]);
	lastLevelOfRootedStructure(pracovniSeznam[k], &zpracVelikost, &zpracSirka, &zpracHloubka, zpracVzdalenosti, &zpracLastLevel)	;
	//	for(int j=0; j<zpracVelikost; j++){
	//	//printf("Zprac Prvek %d  \n" , zpracLastLevel[j]);
	//}
	//printf("velikost %d, hloubka %d, sirka %d  \n" , zpracVelikost, zpracHloubka,zpracSirka);
	if((zpracHloubka > hloubkaSeznamu) && (zpracSirka < aktualniSirka)){
	  // nastavuji novy pocatecni uzel
	  printf("novy start\n");
	  start = pracovniSeznam[k];
	  novyStart = true;
	  
	  delete[] pracovniSeznam;
	  break;
	  
	}
	else if(zpracSirka < aktualniSirka){
	  // nastavuju koncovy uzel
	  cil = pracovniSeznam[k];
	  printf("novy cil\n");
	  aktualniSirka = zpracSirka;
	  nalezenCil = true;
	  break;
	}
	
	
      }
      
  }
  *startUzel = start;
  *cilUzel = cil;
  
}

void gtopology::sloan_renumb (FILE *out)
{
  adjacnodes (out);
  
  int zpracSirka,zpracHloubka,zpracVelikost,*zpracLastLevel;
  //int *zpracVzdalenosti;
  zpracLastLevel = NULL;
  int start,cil;
  int oznacenych=-1;
  int *vzdalenosti, *priority, *statusy, *priorFronta,*noveCislovani;
  int vrcholFronty;
  int zpracUzel;
  int maxPriorita;
  int pomUzel,pomActive, poziceUzlu;
  int j,k;
  //1.entry
  //printf("ACTIVE = %d, 	INACTIVE = %d, 	PREACTIVE=%d,	POSTACTIVE = %d\n",ACTIVE, 	INACTIVE , 	PREACTIVE,	POSTACTIVE);
  findPseudoPheripheral(&start, &cil);
  //start = 0;
  
  //printf("Cislo startovniho uzlu je %d \n souradnice [ %e, %e, %e ]", start, gnodes[start].x, gnodes[start].y, gnodes[start].z);
  //cil = 3;
  //printf("\nCislo ciloveho uzlu je %d \n souradnice [ %e, %e, %e ]", cil, gnodes[cil].x, gnodes[cil].y, gnodes[cil].z);
  //printf("\nstart %d cil %d \n",start,cil);
  
  vzdalenosti = new int[nn];
  priority = new int[nn];
  statusy = new int[nn];
  priorFronta = new int[nn];
  noveCislovani = new int[nn];
  int W1 = 1, W2 = 2;
  //2.compute distances
  lastLevelOfRootedStructure(cil, &zpracVelikost, &zpracSirka, &zpracHloubka, vzdalenosti, &zpracLastLevel)	;
  //3.assign initial status and priority
  
  for(k=0; k<nn; k++){
    //printf(" uzel %d vzdalenost k cily %d a pocet sousedu %ld ", k, vzdalenosti[k]+1, nadjnodnod[k]-1);
    
    for (j = 0; j< (nadjnodnod[k]);j++){
      
      //printf(" %ld", adjnodnod[k][j]);
    }
    //printf("\n");
    priority[k] = W1*(vzdalenosti[k]+1) - W2*(nadjnodnod[k]-1);
    statusy[k] = INACTIVE;
    priorFronta[k] = 0;
    
    /*printf(" vzdalenost uzlu %d od uzlu %d je %d a priorita je %d\n", k,cil,vzdalenosti[k],priority[k] );*/
  }
  //printf("za cyklem\n\n\n");
  //writePriorityQueue(priority);
  //4.initizlize node count and priority queue
  oznacenych=0;
  statusy[start] = PREACTIVE;
  vrcholFronty = 0;
  priorFronta[vrcholFronty] = start;
  // 5.test for termination
  int f=0;
  
  
  while(vrcholFronty > -1){
    
    //writePriorityQueue(priorFronta);
    //printf("\n");
    
    for( k=0; k<nn; k++){
      //printf(" %d",statusy[k]);
      //if( ( (k+1)%3)  == 0)printf("\n");
    }
    //getchar();
    //printf("Nova iterace vrchol fronty je %d\n\n\n",vrcholFronty); 
    f=0;
    while(f<=vrcholFronty){
      //printf("Prioritni fronta pozice %d je prvek %d\n",f,priorFronta[f]);
      f++;
    }
    
    //hledam maximalni prioritu
    //6. select node to be labeled
    maxPriorita = -9999999;
    
    for(k=0; k<=vrcholFronty; k++){// opravit pokud je vrcholFronty +1 mimo meze spadne to!!!
      //printf("Prioritni fronta pozice %d je uzel %d\n",k,priorFronta[k]);
      
      if(priority[priorFronta[k]] > maxPriorita){
	maxPriorita = priority[priorFronta[k]];
	zpracUzel = priorFronta[k];
	poziceUzlu = k;
	//
	
      }
      //printf("Maximalni prioritu ma uzel na pozici %d uzel %d status %d\n",k,  priorFronta[k],statusy[zpracUzel]);
      //printf("Budu zpracovavat uzel %d\n", zpracUzel);
    }
    
    //smazu zpracUzel z fronty
    //7. update queue and priorities
    priorFronta[poziceUzlu] = priorFronta[vrcholFronty];
    vrcholFronty--;
    if(statusy[zpracUzel] == PREACTIVE){
      //printf("Je preactive\n");
      for(j = 0; j < nadjnodnod[zpracUzel]; j++){
	
	pomUzel = adjnodnod[zpracUzel][j];
	if(pomUzel == zpracUzel)continue;
	//printf("Prohledavam okoli %d uzel %d se statusem %d \n",zpracUzel,pomUzel,statusy[pomUzel] );
	priority[pomUzel] = priority[pomUzel] + W2;
	if(statusy[pomUzel] == INACTIVE){
	  
	  vrcholFronty++;
	  priorFronta[vrcholFronty] = pomUzel;
	  statusy[pomUzel] = PREACTIVE;
	  //printf("vkladam do fronty na pozici %d uzel %d\n",vrcholFronty,pomUzel);
	  //printf("davam status %d\n",PREACTIVE);
	}
      }
      
    }
    
    // 8.label the next node
    //printf("oznacuji uzel %d\n",zpracUzel);
    oznacenych++;
    noveCislovani[zpracUzel] = oznacenych;
    statusy[zpracUzel] = POSTACTIVE;
    //9. update queue and priorities
    for(j = 0; j < nadjnodnod[zpracUzel]; j++){
      
      //printf("pred cyklem\n\n\n");
      pomUzel = adjnodnod[zpracUzel][j];
      //printf("pred cyklem\n\n\n");
      if(pomUzel == zpracUzel)continue;
      if(statusy[pomUzel] == PREACTIVE){
	
	
	//printf("Nastavuji uzel %d na ACTIVE\n",pomUzel);
	statusy[pomUzel] = ACTIVE;
	priority[pomUzel] = priority[pomUzel] + W2;
	//printf("Prochazim okoli uzlu %d\n",pomUzel);
	
	for(k = 0; k < nadjnodnod[pomUzel]; k++){
	  
	  pomActive = adjnodnod[pomUzel][k];
	  if(pomUzel == pomActive)continue;
	  //printf("Prohledavam krok 9. okoli %d uzel %d se statusem %d \n",pomUzel,pomActive,statusy[pomActive] );
	  if(statusy[pomActive] != POSTACTIVE){
	    priority[pomActive] = priority[pomActive] + W2;	
	  }
	  if(statusy[pomActive] == INACTIVE){
	    
	    vrcholFronty++;
	    priorFronta[vrcholFronty] = pomActive;
	    statusy[pomActive] = PREACTIVE;
	    //printf("vkladam do fronty na pozici %d uzel %d\n",vrcholFronty,pomActive);
	    //printf("davam status %d\n",PREACTIVE);
	  }
	}
      }
      
    }
  }

  //FILE *outt;
  //outt=fopen ("sloan.txt","w");
  //for( int k=0; k<nn; k++){
  //ordering[noveCislovani[k]-1] = k;
    //	printf(" Puvodni %d nove %d\n", k, noveCislovani[k]);
    //fprintf(outt," Puvodni %d nove %d\n", k, noveCislovani[k]);
  //}
  //fclose (outt);
  
  //printf("--------------------------");
  //for( int k=0; k<nn; k++){
  //	
  //	printf(" Puvodni %d nove %d\n", k, ordering[k] );
  //}
  //printf("--------------------------");
  //printf("--------------------------");
}











