
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <ctype.h>
#include "iotools.h"



/**
  Function writes error on standard error output device.

  @param emsg - string with error message, one terminating character \n will automatically added
  @param errfile - string with source file name where the error was generated
  @param errln - line number in source file  where the error was generated
  @param errfunc - string with function name where the error was generated

  Function does not return anything.
  
  created  12.2006,  Tomas Koudelka, koudelka@cml.fsv.cvut.cz
*/
void print_err(const char *emsg, const char *errfile, int errln, const char *errfunc)
{
  fflush(stdout); // force standard output to be written
  fflush(stderr); // force standard error output to be written
  fprintf(stderr, "\n\nError : %s\n", emsg);
  fprintf(stderr, "in file %s, line %d, function %s\n", errfile, errln, errfunc);
  fflush(stderr);
}



/**
  Function writes error on standard error output device.

  @param n - number of processor
  @param emsg - string with error message, one terminating character \n will automatically added
  @param errfile - string with source file name where the error was generated
  @param errln - line number in source file  where the error was generated
  @param errfunc - string with function name where the error was generated

  Function does not return anything.
  
  created  12.2006,  Tomas Koudelka, koudelka@cml.fsv.cvut.cz
*/
void par_print_err(int n,const char *emsg, const char *errfile, int errln, const char *errfunc)
{
  fflush(stdout); // force standard output to be written
  fflush(stderr); // force standard error output to be written
  fprintf(stderr, "\n\nError on processor %d:\n %s\n",n,emsg);
  fprintf(stderr, "in file %s, line %d, function %s\n", errfile, errln, errfunc);
  fflush(stderr);
}



/**
  Function opens file given by parameter name with given mode. This function replaces standard fopen
  function for XFILE type. Default, the kwdmode is set to 'ignore' and warrnings are switched off

  @param name - string with file name
  @param mode - string with file handling mode. Only "r" or "a" reading mode is supported

  @retval The function returns pointer to the newly allocated structure XFILE with opened file or NULL
          in case failure of file opening.

  created  12.2006,  Tomas Koudelka, koudelka@cml.fsv.cvut.cz
*/
XFILE *xfopen(const char *name, const char *mode)
{
  XFILE *ifile;
  char *emsg;
  if (strchr(mode,'w') || strchr(mode,'a'))
  {
    print_err("unsupported mode for opening of XFILE", __FILE__, __LINE__, __func__);
    abort();
  }
  if (name == NULL)
  {
    return NULL;
  }
  ifile           = new XFILE;
  ifile->file     = fopen(name, mode);
  ifile->line     = 1;
  ifile->col      = 1;
  ifile->lnfpos   = 0;
  ifile->warrning = 0;

  ifile->fname = new char[strlen(name)+1];
  memcpy(ifile->fname, name, strlen(name)+1);

  ifile->kwdmode    = ignore;
  ifile->ignorecase = 0;

  ifile->index_created =  0;
  ifile->id_sec        = -1;
  ifile->asect         =  NULL;
  ifile->num_sec       =  0;
  ifile->sect          =  NULL;

  if (ifile->file == NULL)
  {
    emsg = new char[50+strlen(name)];
    sprintf(emsg, "cannot open file %s for reading", name);
    print_err(emsg, __FILE__, __LINE__, __func__);
    delete [] emsg;
    delete [] ifile->fname;
    delete ifile;
    ifile = NULL;
  }
  return ifile;
}



/**
  Function closes file given by parameter f. This function replaces standard fclose
  function for XFILE type.

  @param f - pointer to opened XFILE structure

  @retval The function returns the same value as fclose function

  created  12.2006,  Tomas Koudelka, koudelka@cml.fsv.cvut.cz
*/
int xfclose(XFILE* &f)
{
  int ret;
  ret = fclose(f->file);
  delete [] f->fname;
  delete [] f->sect;
  f->sect = NULL;
  delete f;
  f=NULL;
  return ret;
}



/**
  Function loactes the first occurence of keyword given by kwd 
  in the string (line) given by lnstr. Total number of occurences
  is returned via noc.

  @param lnstr - pointer to scanned string
  @param kwd   - pointer to searched string
  @param ignorecase - flag for case insensitivity (=1)/ case sensitivity (=0)
  @param noc   - number of occurences searched keyword in lnstr

  @retval The function returns either pointer to the first occurence of 
          keyword in the lnstr or NULL in case that the keyword was not found.

  created 02.2008, Tomas Koudelka, koudelka@cml.fsv.cvut.cz
*/
char *locate_kwd(char *lnstr,const char *kwd, long ignorecase, long &noc)
{
  char *ret;
  char *astr;
  size_t kwdl = strlen(kwd);
  size_t strl = strlen(lnstr);
  char *kwd_beg;
  noc = 0;
  ret = NULL;
  astr = lnstr;
  do
  {
    kwd_beg = strstr(astr, kwd, ignorecase);
    if (kwd_beg)
    {
      // testing that the keyword is not part of another word
      if (kwd_beg-lnstr+kwdl < strl)  
      {
        if (isspace(kwd_beg[kwdl]) == 0) 
        // something is joined to the end of located keyword
        {
          astr = kwd_beg+kwdl;
          kwd_beg = NULL;
          continue;
        }
      }
      if (kwd_beg-lnstr > 0)
      {
        if (isspace(kwd_beg[-1]) == 0) 
        // something is prepended to the beginning of located keyword
        {
          if (kwd_beg-lnstr+kwdl < strl)  
          {
            astr = kwd_beg+kwdl;
            kwd_beg = NULL;
            continue;
          }
          else
          {
            kwd_beg = NULL;
            break;
          }
        }
      }
      // located keyword is correct
      if (ret == NULL)
        ret = kwd_beg;
      noc++;
      if (kwd_beg-lnstr+kwdl < strl)  
      // searching continues for other possible keyword occurences
      // in the rest of lnstr
      {
        astr = kwd_beg+kwdl;
        kwd_beg = NULL;
      }
      else
      // end of lnstr was reached -> searching stops
        break;
    }
    else
    // keyword was not found in the lnstr
      break;
  } while(1);

  return ret;
}



/**
  Function detects particular sections in file given by parameter f. Number of 
  searched sections is given by data member n of kwdb and kwde (they must be the same). 
  These structures contain array with base keywords indicating  beginnings or ends 
  of searched sections. Not all of the searched sections have to be detected and 
  the function stores data about detected sections in coresponding data mambers in 
  structure of f (see description of XFILE).

  @param f - pointer to opened XFILE structure
  @param kwdb    - structure containing array with keywords used for indicating
                   beginning of sections
  @param kwde    - structure containing array with keywords used for indicating
                   end of sections

  @retval The function returns number of successfully detected sections

  created  02.2008,  Tomas Koudelka, koudelka@cml.fsv.cvut.cz
*/
long xfdetect_sect(XFILE* f, const kwdset &kwdb, const kwdset &kwde)
{
  char *lnstr;
  long i, l;
  long n = kwdb.n;
  long num_sec;
  long line;
  long col;
  long lnfpos = 0;
  size_t kwdl;
  long noc;
  char *kwd_beg;
  char *kwd_end;
  char errmsg[1001];
  xfsection **sect;
  xfsection *tsect;

  switch (f->kwdmode)
  {
    case sect_mode_seq:
    case sect_mode_full:
      break;
    default:
      return 0;
  }

  lnstr = new char[f->maxlnsize];
  sect = new xfsection*[n];
  memset(sect, 0, sizeof(*sect)*n);

  fseek(f->file, 0, SEEK_SET);
  line = 1;
  if (kwdb.n != kwde.n)
  {
    sprintf(errmsg, "Number keywords for beginning of sections is not the same\n"
                    " as number of keywords for end of sections\n");
    print_err(errmsg, __FILE__, __LINE__, __func__);
    abort();
  }
  do
  {
    memset(lnstr, 0, sizeof(*lnstr)*f->maxlnsize);
    fscanf(f->file, "%[^\n]",lnstr); //read new line without terminating \n
    clearcomments(lnstr);   // clear possible comments
    // removing trailing \r form the last character of string (due to DOS \r\n newline)
    l = strlen(lnstr);
    if (l > 0)
    {
      if (lnstr[l-1] == '\r')
        lnstr[l-1] = '\0';
    }
    if (strlen(lnstr) == 0) // actual line have zero length
    {
      fscanf(f->file, "%1[\n]", lnstr); // read terminating \n
      line++;
      col=1;
      lnfpos = ftell(f->file);// backup of begining line position
      continue;
    }
    for (i=0; i<n; i++)
    {
      // searching beginnings of sections
      kwdl = strlen(kwdb.set[i].alias);
      kwd_beg = locate_kwd(lnstr, kwdb.set[i].alias, f->ignorecase, noc);
      if (kwd_beg)
      {
        if ((sect[i] == NULL) && (noc < 2))  // first keyword occurence
        {         
          tsect = new xfsection;
          tsect->beg_secpos = lnfpos + kwd_beg - lnstr + kwdl;
          tsect->beg_secln  = line;
          tsect->beg_seccol = kwd_beg - lnstr + kwdl + 1;
          tsect->name.id = kwdb.set[i].id;
          tsect->name.alias = new char [kwdl+1];
          //sprintf(tsect->name.alias, "%s",kwdb.set[i].alias);
	  tsect->name.alias = kwdb.set[i].alias;
          tsect->end_secpos = -1;
          sect[i] = tsect;
        }
        else  // multiple keyword occurence
        {
          if (sect[i] == NULL)
            sprintf(errmsg, "Multiple occurence of keyword %s on line %ld\n", 
                    kwdb.set[i].alias, line);
          else
            sprintf(errmsg, "Multiple occurence of keyword %s(line %ld and line %ld)\n", 
                    kwdb.set[i].alias, sect[i]->beg_secln, line);
          print_err(errmsg, __FILE__, __LINE__, __func__);
          abort();
        }
      }

      // searching ends of sections
      kwdl = strlen(kwde.set[i].alias);
      kwd_end = locate_kwd(lnstr, kwde.set[i].alias, f->ignorecase, noc);
      if (kwd_end)
      // keyword for end of section was located
      {
        if (sect[i] == NULL)  
        // error - keyword for end of section preceds keyword for beginning
        {         
          sprintf(errmsg, "Keyword for end of section %s(line %ld) precedes keyword for beginning of section\n", 
                  tsect->name.alias, line);
          print_err(errmsg, __FILE__, __LINE__, __func__);
          abort();
        }
        if ((noc > 1) || (sect[i]->end_secpos >= 0))  // multiple keyword occurence
        {
          if (sect[i]->end_secpos >= 0)
            sprintf(errmsg, "Multiple occurence of keyword %s on line %ld\n", 
                    kwde.set[i].alias, line);
          else
            sprintf(errmsg, "Multiple occurence of keyword %s(line %ld and line %ld)\n", 
                    kwde.set[i].alias, sect[i]->end_secln, line);


          print_err(errmsg, __FILE__, __LINE__, __func__);
          abort();
        }
        sect[i]->end_secpos = lnfpos + kwd_end - lnstr;
        sect[i]->end_secln  = line;
        sect[i]->end_seccol = kwd_end - lnstr + 1;
      }
    }
  } while (!feof(f->file));

  // Counting total number of detected sections
  for(i=0, num_sec=0; i<n; i++)
  {
    if (sect[i])
      num_sec++;
  }
  if (num_sec)
  {
    // Creating index of sections in xfile structure and copying results into it
    f->sect = new xfsection[num_sec];
    for(i=0; i<n; i++)
    {
      if (sect[i])
      {
        if (sect[i]->end_secpos < 0)
        {
          sprintf(errmsg, "Beginning keyword %s was found (line %ld) but end keyword %s was not found\n", 
                  sect[i]->name.alias, sect[i]->beg_secln, kwde.set[i].alias);
          print_err(errmsg, __FILE__, __LINE__, __func__);
          abort();
        }
        f->sect[i].beg_secpos = sect[i]->beg_secpos; 
        f->sect[i].end_secpos = sect[i]->end_secpos; 
        f->sect[i].beg_secln  = sect[i]->beg_secln; 
        f->sect[i].end_secln  = sect[i]->end_secln; 
        f->sect[i].beg_seccol = sect[i]->beg_seccol; 
        f->sect[i].end_seccol = sect[i]->end_seccol; 
        f->sect[i].name = sect[i]->name;
      }
      delete sect[i];
    }
    f->index_created = 1;
    f->id_sec = 0;
    f->asect = &f->sect[0];
    f->num_sec = num_sec;
  }
  delete [] sect;  
  delete [] lnstr;

  fseek(f->file, f->lnfpos, SEEK_SET);
  return num_sec;
}



/**
  Function resets internal variables of given file so that they 
  points to the beginning of the actual section.

  @param f - pointer to opened XFILE structure

  @retval The function returns nothing

  created  02.2008,  Tomas Koudelka, koudelka@cml.fsv.cvut.cz
*/
void xf_resetsec(XFILE *f)
{
  if ((f->index_created == 0) || (f->id_sec < 0) || (f->id_sec > f->num_sec-1))
  {
    char errmsg[1001];
    if (f->index_created == 0)
      sprintf(errmsg, "Error - index of sections in file %s was not created\n", f->fname);
    if ((f->id_sec < 0) || (f->id_sec > f->num_sec-1))
      sprintf(errmsg, "Error - index %ld of section in file %s is out of range (0, %ld)\n", f->id_sec, f->fname, f->num_sec-1);
     print_err(errmsg, __FILE__, __LINE__, __func__);
     abort();
  }
  fseek(f->file, f->asect->beg_secpos, SEEK_SET);
  f->line   = f->asect->beg_secln;
  f->col    = f->asect->beg_seccol;
  f->lnfpos = f->asect->beg_secpos-f->col+1;
  return;
}



/**
  Function sets internal file pointers to the beginning of section with given alias.

  @param f         - pointer to opened XFILE structure
  @param sec_alias - structure containing string with section name and enum with section alias

  @retval The function returns following values:
          0 - on success
          1 - index was not found
          2 - section index was not built

  created  02.2008,  Tomas Koudelka, koudelka@cml.fsv.cvut.cz
*/
long xf_setsec(XFILE *f, enumstr sec_alias)
{
  long i, id = -1;

  if (f->index_created == 0)
  // index of sections was not built
    return 2;

  for (i = 0; i < f->num_sec; i++)
  {
    if (strstr(f->sect[i].name.alias, sec_alias.alias) &&
        f->sect[i].name.id == sec_alias.id)
    {
      // required alias of section was found
      id = i;
      break;
    }
  }
  if (id < 0)
  // required alias was not found
    return 1;

  // setup of internal file pointers  
  f->id_sec = id;
  f->asect  = &f->sect[id];
  xf_resetsec(f);
  return 0;
}



/**
  Function sets internal file pointers to the beginning of section of given id

  @param f      - pointer to opened XFILE structure
  @param sec_id - index of required section in the array sect of XFILE

  @retval The function returns following values:
          0 - on success
          1 - index is out of range
          2 - index was not built

  created  02.2008,  Tomas Koudelka, koudelka@cml.fsv.cvut.cz
*/
long xf_setsec(XFILE *f, long sec_id)
{
  if (f->index_created == 0)
  // index of sections was not built
    return 2;

  if ((sec_id < 0) || (sec_id > f->num_sec))
  // required index is out of range
    return 1;

  // setup of internal file pointers  
  f->id_sec = sec_id;
  f->asect = &f->sect[sec_id];
  xf_resetsec(f);
  return 0;
}



/**
  Function extends standard fscanf function about error checking, input type checking,
  and keywords. Several new conversions are defined or redefined :
  %a - reads string including whitespaces until the newline chracter is reached
       argument of char* type is expected and it has to have sufficient memory allocated to hold
       whole scanned string including trailing \0.
  %k - keyword stored in the corresponding argument will be located in the input file. Keyword locating and handling with keyword 
       is controlled by the XFILE kwdmode option. Argument should be string with required keyword, no conversion is performed.
       Following conversion should perform conversion and assignment of keyword value to the argument which follows argument with 
       keyword string.
  %m - reads enum value given either by string or corresponding integer value. This conversion processes two arguments. The first argument
       should be pointer to class kwdset, which contains array with enumstr structures containing 
       predefined alias strings and integer values. This set of aliases (string or integer values) are searched in the input file.
       Integer value of located alias is assigned to the second argument which is supposed to be pointer to integer type.

  @param in - pointer to the opened XFILE structure
  @param fmt - format string
  @param ... - arguments used for conversions prescribed in the fmt string

  @retval returns number of successfully scanned and assigned items
  
  created  12.2006,  Tomas Koudelka, koudelka@cml.fsv.cvut.cz
*/
int xfscanf(XFILE *in, char *fmt, ...)
{
  long  nvar;      // number of required variables by fmt string
  char *types;     // array with identfied conversion types in fmt
  char **ptypes;   // array with pointers to % before each identfied conversion types in fmt
  long  idt;       // acual id number of processed conversion in types array
  char *subfmt;    // string with whole actually processed conversion
  unsigned int br; // total number of read bytes in sscanf call
  int tbr;         // total bytes read in xfscanf call
//  unsigned int br1; // number of read bytes in sscanf call for assignment
//  unsigned int br2; // number of read bytes in sscanf call for other characters
  va_list params;  // list of passed parameters where the data will be read to
  void *apar;      // pointer to the acutual read parameter
  char *lnstr;     // string with acutal read line
  char *aptr;      // actual position in lnstr where the data are read from
  long  errbrk;    // break flag for error in sscanf
  char *errmsg;    // auxiliary string for error messages
  long  read;      // flag for reading of new line
  long  ret;       // return value from sscanf or get... functions
  long war;        // warning flag
  long readkwdval; // flaf for reading keyword value
  kwdset *akwdset; // keyword set for enums 
  long l;          // lenght of string - auxiliary variable
  long start_pos;  // starting position in the file
  
  // detection of total number of required conversions
  nvar = checkfmt(fmt);
  if (nvar == 0)
    return 0;
    
  types  = new char[nvar];
  ptypes = new char*[nvar];
  subfmt = new char[strlen(fmt)+5];
  lnstr  = new char[in->maxlnsize];
  errmsg = new char[1001];
  
  // identification of each conversions and
  // checking of their correct types 
  gettypes(fmt, types, ptypes, nvar);

  va_start(params, fmt);
  idt = 0;
  errbrk = 0;
  read = 1;
  readkwdval = 0;
  tbr = 0;
  start_pos = ftell(in->file);
  do
  {
    if (read && readkwdval && (in->kwdmode == line_mode))
    {  
      // reading of new line is required and the keyword search mode is 
      // line_mode => error keyword value was not found in the actual line
      sprintf(errmsg, "unable to read keyword value (fmt '%s', par id=%ld)\n"
                      "input file : %s, line=%ld, col=%ld", fmt, idt+1, in->fname, in->line, in->col);
      print_err(errmsg, __FILE__, __LINE__, __func__);
      abort();      
    }
    if (read && readkwdval && (in->kwdmode == sect_mode_seq) && (in->line == in->asect->end_secln))
    {  
      // reading of new line is required and the keyword search mode is 
      // line_mode => error keyword value was not found in the actual line
      sprintf(errmsg, "unable to read keyword value (fmt '%s', par id=%ld)\n"
                      "input file : %s, line=%ld, col=%ld", fmt, idt+1, in->fname, in->line, in->col);
      print_err(errmsg, __FILE__, __LINE__, __func__);
      abort();      
    }
    if (read) // read new line
    {
      memset(lnstr, 0, sizeof(*lnstr)*in->maxlnsize);
      checkfeof(in, fmt, idt, ptypes[idt], subfmt, errmsg, __FILE__, __LINE__, __func__);
      checkeos(in, fmt, idt, ptypes[idt], subfmt, errmsg, __FILE__, __LINE__, __func__);
      fscanf(in->file, "%[^\n]",lnstr); //read new line without terminating \n
      // cut string before end of actual section
      cut_str_sec(in, lnstr);
      // removing trailing \r form the last character of string (due to DOS \r\n newline)
      l = strlen(lnstr);
      if (l > 0)
      {
        if (lnstr[l-1] == '\r')
          lnstr[l-1] = '\0';
      }
      clearcomments(lnstr);   // clear possible comments
      aptr = lnstr;       // setup actual position in lnstr
      read = 0;
    }
    if (strlen(aptr) == 0) // actual line have zero length
    {
      checkfeof(in, fmt, idt, ptypes[idt], subfmt, errmsg, __FILE__, __LINE__, __func__);
      checkeos(in, fmt, idt, ptypes[idt], subfmt, errmsg, __FILE__, __LINE__, __func__);
      fscanf(in->file, "%1[\n]", lnstr); // read terminating \n
      in->line++;
      in->col=1;
      in->lnfpos = ftell(in->file);// backup of begining line position
      read = 1;
      continue;
    }
    // for usual conversions (real and integer), a new line is required in case of empty string
    if (strchr("efgdikmopsux",types[idt]) && isemptystr(aptr, " \t\n\r"))
    {
      checkfeof(in, fmt, idt, ptypes[idt], subfmt, errmsg, __FILE__, __LINE__, __func__);
      checkeos(in, fmt, idt, ptypes[idt], subfmt, errmsg, __FILE__, __LINE__, __func__);
      fscanf(in->file, "%1[\n]", lnstr);  // read terminating \n
      in->line++;
      in->col=1;
      in->lnfpos = ftell(in->file);// backup of begining line position
      read = 1;
      continue;
    }
    // for extended string conversion, a new line is required in case of empty string (due to DOS \r\n newline)
    if (strchr("a",types[idt]) && isemptystr(aptr, "\r"))
    {
      checkfeof(in, fmt, idt, ptypes[idt], subfmt, errmsg, __FILE__, __LINE__, __func__);
      checkeos(in, fmt, idt, ptypes[idt], subfmt, errmsg, __FILE__, __LINE__, __func__);
      fscanf(in->file, "%1[\n]", lnstr);  // read terminating \n
      in->line++;
      in->col=1;
      in->lnfpos = ftell(in->file);// backup of begining line position
      read = 1;
      continue;
    }
    // copy format for actual conversion to the subfmt and append %n due to 
    // column counter
    getsubfmt(ptypes[idt], subfmt,1);
    ret = 0;
    war = 0;
    if ((check_asterisk(subfmt)==1) || (types[idt] == ' '))
    // assignment suppression indicator ("%*...") was detected -> no parameter should be passed
      apar = NULL;
    else
    // otherwise selection of newly read parameter
      apar = va_arg(params, void*);
    
    switch (tolower(types[idt]))
    {
      case 'a': // %a conversion = reading of actual line to string
        l = strlen(aptr);
        memcpy(apar, aptr, sizeof(char)*l);
        ((char *)apar)[l] = '\0';
        br = l;
        errbrk = 0;
        readkwdval = 0;
        break;
      case 'k': // %k conversion = locating of keyword depending on the in->kwdmode and in->ignorecase
        // locates keyword from the actual line position, 
        // in the whole line or in the given section
        ret = getkwd(in, lnstr, aptr, (char *)apar, br);
        if (ret)    
        {
          // in case of error reading keyword
          switch (ret)
          {
            case 1 :
              sprintf(errmsg, "unable to read keyword %s (fmt '%s', par id=%ld)\n"
                              "input file : %s, line=%ld, col=%ld", (char *)apar, fmt, idt+1, in->fname, in->line, in->col);
              break;
            case 2 :
              sprintf(errmsg, "multiple occurence of keyword %s on line (fmt '%s', par id=%ld)\n"
                              "input file : %s, line=%ld, col=%ld", (char *)apar, fmt, idt+1, in->fname, in->line, in->col);
              break;
            case 3 :
              sprintf(errmsg, "zero length of keyword %s (fmt '%s', par id=%ld)\n"
                              "input file : %s, line=%ld, col=%ld", (char *)apar, fmt, idt+1, in->fname, in->line, in->col);
              break;
              
          }
          print_err(errmsg, __FILE__, __LINE__, __func__);
          errbrk = 1;
        }
        else
        {
          // setup flag for reading of keyword value
          readkwdval = 1;
        }
        break;
      case 'm': // %m conversion = reading of enum aliases either by the string alias or by the integer value
        akwdset = (kwdset *)apar; // the first argument is the keyword set
        // selection of newly read enum parameter
        apar = va_arg(params, void*); // the second argument is real 
        ret = getenum(aptr, akwdset,(int *)apar, br, in->kwdmode, in->ignorecase);
        errbrk = checkenumerr(in, ret, 1, war, akwdset, errmsg, fmt, subfmt, idt, 
                              ptypes[idt], __FILE__, __LINE__, __func__);
        readkwdval = 0;
        break;
      case 'f':  // real number conversions %f, %e, %g, %E
      case 'e':
      case 'g':
        war = checkreal(aptr); // checking whether the actual scanned string contains real number
        ret = sscanf(aptr, subfmt,apar,&br);
        errbrk = checkscanferr(in, ret, 1, war, errmsg, "unable to read real number",
                               fmt, subfmt, idt, ptypes[idt], __FILE__, __LINE__, __func__);
        readkwdval = 0;
        break;
      case 'd':   // integer number conversions %d, %i, %o, %p, %x, %D, %X
      case 'i':
      case 'o':
      case 'p':
      case 'x':
        war = checkint(aptr); // checking whether the actual scanned string contains integer number
        ret = sscanf(aptr, subfmt,apar,&br);
        errbrk = checkscanferr(in, ret, 1, war, errmsg, "unable to read integer number",
                               fmt, subfmt, idt, ptypes[idt], __FILE__, __LINE__, __func__);
        readkwdval = 0;
        break;
      case 'u':  // unsigned integer number conversion %u
        war = checkuint(aptr);  // checking whether the actual scanned string contains unsigned integer number
        ret = sscanf(aptr, subfmt,apar,&br);
        errbrk = checkscanferr(in, ret, 1, war, errmsg, "unable to read unsigned integer number",
                               fmt, subfmt, idt, ptypes[idt], __FILE__, __LINE__, __func__);
        readkwdval = 0;
        break;
      case 'n':  // number of read characters conversion %n
        *((int *)apar) = tbr;
        errbrk = 0;
        break;
      case ' ': // without conversion which will not be checked
        ret = sscanf(aptr, subfmt,&br);
        errbrk = 0;
        readkwdval = 0;
	break;
      default: // default unscpecified conversion which will not be checked
        ret = sscanf(aptr, subfmt,apar,&br);
        errbrk = checkscanferr(in, ret, 1, 0, errmsg, "unable to read required conversion",
                               fmt, subfmt, idt, ptypes[idt], __FILE__, __LINE__, __func__);
        readkwdval = 0;
    }
    if (errbrk) // an error was encountered leave the loop
      break;
    else
    {
      // conversion was successfully scanned and assigned to the argument
      idt++; // index of next conversion
      if (((in->kwdmode == sect_mode_seq) || (in->kwdmode == sect_mode_full)) && (readkwdval == 1))
        tbr = in->lnfpos - in->asect->beg_secpos + in->col - 1;
      else
      {
        in->col += br; // increase column number with read bytes
        aptr += br;    // change actual position pointer in the scanned line
        tbr = in->lnfpos - start_pos + in->col - 1;
      }
      //tbr += br; 
    }
  } while (idt < nvar); // repeat while the index of actual conversion is less then number of required variables in the format string
  va_end(params);
  ret = idt;
  if (types[0] == ' ')
    ret--;
  delete [] types;
  delete [] ptypes;
  delete [] subfmt;
  delete [] lnstr;
  delete [] errmsg;
  if (errbrk) 
    abort();
    
  // setting actual position in file depending on the number of 
  // processed characters of line (i.e. column number-1)
  fseek(in->file, in->lnfpos+in->col-1,SEEK_SET);
  return ret;
}



/**
  Function removes commented characters from the string. The '#' character
  is assumed to be begining of commented part of string.

  @param lnstr - pointer to opened XFILE structure

  @retval The function does not return anything.

  created  12.2006,  Tomas Koudelka, koudelka@cml.fsv.cvut.cz
*/
void clearcomments(char *lnstr)
{
  long i = 0;
  do
  {
    if (lnstr[i] == '#')
    {
      lnstr[i] = '\0';
      return;
    }
    i++;
  } while (lnstr[i] != '\0');
  return;
}



/**
  Function scans for format conversions in the given string. 
  The number of conversion is increased by one in case that the
  first character of the format string does not contain a conversion
  indicator (%). In such the case the begining of the format string
  up to the first conversion indicator is treated as an individual conversion
  (It can be used for skipping initial spaces in the input file - adding initial space befor
   conversion indicator - " %a", simple keyword management "keyword %ld" etc.; see description 
   of scanf function for handling of nonconversion charcaters.)
  Syntax of format string is same as in the standard scanf function.

  @param fmt - pointer to opened XFILE structure

  @retval The function returns number of located conversions.

  created  12.2006,  Tomas Koudelka, koudelka@cml.fsv.cvut.cz
*/
long checkfmt(const char *fmt)
{
  long i, l, nt, idp;
  l = strlen(fmt);
  nt = 0;
  idp = 0;
  for(i=0; i<l; i++)
  {
    if ((fmt[i] == '%') && (idp == 0))
    {
      idp = 1;
      continue;
    }
    if ((fmt[i] == '%') && (idp == 1))
    {
      if (i==1) // initial chracters are treated as an individual conversion
        nt++;
      idp = 0;
      continue;
    }
    if ((i == 0) && (fmt[i] != '%'))  // initial chracters before the first % are treated as an individual conversion
      nt++;
    if ((fmt[i] != '%') && (idp == 1))
    {
      nt++;
      idp = 0;
    }
  }
  return nt;
}



/**
  Function assembles one begining conversion from the format string and copies it to the subfmt string. 
  If the bc flag is set up than the byte count conversion (%n) is appended to subformat string.

  @param fmt - format string
  @param subfmt - subformat string
  @param bc - byte count flag whether %n is appended to the subformat string (bc=1) or not (bc=0).

  @retval The function does not return anything.

  created  12.2006,  Tomas Koudelka, koudelka@cml.fsv.cvut.cz
*/
void getsubfmt(const char *fmt, char *subfmt,long bc)
{
  long i, l;
  
  l = strlen(fmt);
  i = 0;
  memset(subfmt, 0, sizeof(*subfmt)*(l+5));
  while (i < l)
  {
    if ((fmt[i] == '%') && (i != 0))
      break;
    subfmt[i] = fmt[i];
    i++;
  }
  if (bc)
    strcpy(subfmt+i, "%n");
  else
    subfmt[i] = '\0';
/*  if (bc)
  {
    l = strlen(subfmt);
    j = -1;
    k = 0;
    while ((j < 0) && (k < l))
    {
      switch(subfmt[k])
      {
        case 'f':
        case 'e':
        case 'g':
          j = k;
          break;
        case 'd':
        case 'i':
        case 'o':
        case 'p':
        case 'x':
          j = k;
          break;
        case 'u':
          j = k;
          break;
      }
      k++;
    }
    if ((j >= 0) && ((l-k)>0))
    {
      tmp = new char[l-k+1];
      strcpy(tmp, subfmt+k);
      strcpy(subfmt+k, "%n");
      strcpy(subfmt+k+2, tmp);
      strcpy(subfmt+l+2, "%n");
      delete [] tmp;
    }
  }
  else
    subfmt[i] = '\0';*/
}



/**
  Function extracts particular conversions from the format string fmt and copies their base charctyer to the array types. 
  In addition, the function stores pointers to the beginign of each conversion in the ptypes array.

  @param fmt    - format string
  @param types  - array with base characters of particular conversions
  @param ptypes - array of pointers to begining of particular conversions in the format string fmt
  @param nvar   - number of required variables

  @retval The function does not return anything.

  created  12.2006,  Tomas Koudelka, koudelka@cml.fsv.cvut.cz
*/
void gettypes(char *fmt, char *types,char** ptypes, long nvar)
{
  long i, j, l, idp;
  l = strlen(fmt); // length of format string
  idp = 0;         // flag for percent character (%) occurence
  char *subfmt;    // subformat string
  char *afmt;      // auxiliary format string
  
  subfmt = new char[l+5];
  j=0;  // conversion counter
  for(i=0; i<l; i++)
  {
    if ((fmt[i] != '%') && (i == 0))
    {
      ptypes[j] = fmt;
      j++;
      continue;
    }
    if ((fmt[i] == '%') && (idp == 0)) // check for double percent character - the first occurence of precent character
    {
      idp = 1;
      continue;
    }
    if ((fmt[i] == '%') && (idp == 1)) // check for double percent character - the second occurence of precent character
    {
      if (i==1)
      {
        ptypes[j] = fmt;
        j++;
      }
      idp = 0;
      continue;
    }
    if ((fmt[i] != '%') && (idp == 1)) 
    // check for double percent character - the second occurence of precent character was not found
    // i.e. normal conversion was located
    {
      ptypes[j] = &fmt[i-1]; // pointer to the begining of conversion (%)
      j++;
      idp = 0;
    }
  }
  for(i=0; i<nvar; i++)
  {
    getsubfmt(ptypes[i], subfmt, 0); // assemble subformat form the format string
    afmt = strpbrk(subfmt, "acCdeEfFgGikmnopsSuxX[");  // check for known conversions
    if (afmt != NULL)
      types[i] = *afmt;
    else
    {
      if (i==0)
        types[i] = ' ';
      else
      {
        print_err("unknown type of conversion is required in xfscanf", __FILE__, __LINE__, __func__);
        abort();
      }
    }
  }
  delete [] subfmt;
}



/**
  Function checks for asterisk chracter which suppresses following conversion assignment

  @param fmt    - format string

  @retval 0 - no asignment suppression was found
  @retval 1 - assignment suppression was detected

  created  12.2006,  Tomas Koudelka, koudelka@cml.fsv.cvut.cz
*/
long check_asterisk(const char *fmt)
{
  long l = strlen(fmt);
  long i;
  long fb = 0; // non-whitespace character flag
  for(i=1; i<l; i++)
  {
    if ((fmt[i] == '*') && (fb == 0))
      return 1;
    if ((fmt[i] == '*') && (fb == 1))
    {
      print_err("invalid position of assignment suppresion ""*"" in xfscanf", __FILE__, __LINE__, __func__);
      abort();
    }
    if (fb == 0)
    {
      if (isspace(fmt[i]) != 0)
        fb = 1;
    } 
  }
  return 0;
}




/**
  Function checks string lnstr for integer value

  @param lnstr - input string

  @retval 0 - only permited characters was found for integer
  @retval 1 - nonpermited character was found 

  created  12.2006,  Tomas Koudelka, koudelka@cml.fsv.cvut.cz
*/
long checkint(const char *lnstr)
{
  long i=0, l;
  l = strlen(lnstr);
  while (isspace(lnstr[i]) && (i < l)) // search the first non-whitespace character
    i++;
  while((isspace(lnstr[i])==0) && (i < l)) 
  {
    if (strchr("-+0123456789", lnstr[i])) // check each character whether is permitted for integer number
      i++;
    else
      return 1;
  }
  return 0;
}



/**
  Function checks string lnstr for unsigned integer value

  @param lnstr - input string

  @retval 0 - only permited characters was found for integer
  @retval 1 - nonpermited character was found 

  created  12.2006,  Tomas Koudelka, koudelka@cml.fsv.cvut.cz
*/
long checkuint(const char *lnstr)
{
  long i=0, l;
  l = strlen(lnstr);
  while (isspace(lnstr[i]) && (i < l)) // search the first non-whitespace character
    i++;
  while((isspace(lnstr[i])==0) && (i < l))
  {
    if (strchr("+0123456789", lnstr[i])) // check each character whether is permitted for unsigned integer number
      i++;
    else
      return 1;
  }
  return 0;
}



/**
  Function checks string lnstr for real number value

  @param lnstr - input string

  @retval 0 - only permited characters was found for integer
  @retval 1 - nonpermited character was found 

  created  12.2006,  Tomas Koudelka, koudelka@cml.fsv.cvut.cz
*/
long checkreal(const char *lnstr)
{
  long i=0, l;
  l = strlen(lnstr);
  while (isspace(lnstr[i]) && (i < l)) // search the first non-whitespace character
    i++;
  while((isspace(lnstr[i])==0) && (i < l))
  {
    if (strchr("-+.0123456789eE", lnstr[i])) // check each character whether is permitted for real number
      i++;
    else
      return 1;
  }
  return 0;
}



/**
  Function checks string lnstr contains only set of chracters defined by the string filter
  Usually in this file, the filter contains whitespace characters and the function detects 'empty' string.

  @param lnstr - input string

  @retval 1 - string contains only characters from the filter set
  @retval 0 - string contains not only characters from the filter set

  created  12.2006,  Tomas Koudelka, koudelka@cml.fsv.cvut.cz
*/
long isemptystr(const char *lnstr,const char *filter)
{
  long i = 0;
  while(lnstr[i] != '\0')
  {
    if (strchr(filter, lnstr[i]) == NULL) // detect whether the actual character is in the filter set
      return 0;
    i++;
  }
  return 1;
}



/**
  Function checks opened file in for end of file (EOF).

  @param in - pointer to the structure XFILE with opened text file for reading
  Following parameters are used only for error message 
  @param fmt - format string
  @param idt - index of actually processed conversion in fmt
  @param ptype   - pointer to the actual conversion
  @param subfmt  - string for subformat
  @param errmsg  - string for error message
  @param errfile - string with source file name where the error was detected
  @param errline - line number in source file where the error was detected
  @param errfunc - string with function name where the error was detected

  @retval Function does not return anything.

  created  12.2006,  Tomas Koudelka, koudelka@cml.fsv.cvut.cz
*/
void checkfeof(XFILE *in, const char *fmt, long idt, const char *ptype, char *subfmt,
               char *errmsg, const char *errfile, int errline, const char *errfunc)
{
  if (feof(in->file))
  {
    getsubfmt(ptype, subfmt, 0);
    sprintf(errmsg, "end of file was reached - reading cannot be continued\n"
                    "(format '%s', subformat '%s', parameter id=%ld)\n"
                    "input file : %s, line=%ld, column=%ld", fmt, subfmt, idt+1, 
                    in->fname, in->line, in->col);
    print_err(errmsg, errfile, errline, errfunc);
    abort();
  }
  return;
}



/**
  Function checks opened file in for end of section (only for sect_mode_seq and sect_mode_full).

  @param in - pointer to the structure XFILE with opened text file for reading
  Following parameters are used only for error message 
  @param fmt - format string
  @param idt - index of actually processed conversion in fmt
  @param ptype   - pointer to the actual conversion
  @param subfmt  - string for subformat
  @param errmsg  - string for error message
  @param errfile - string with source file name where the error was detected
  @param errline - line number in source file where the error was detected
  @param errfunc - string with function name where the error was detected

  @retval Function does not return anything.

  created  12.2006,  Tomas Koudelka, koudelka@cml.fsv.cvut.cz
*/
void checkeos(XFILE *in,char *fmt, long idt,char *ptype,char *subfmt,
              char *errmsg, char *errfile, int errline, const char *errfunc)
{
  switch (in->kwdmode)
  {
    case ignore:
    case sequent_mode:
    case line_mode:
      return;
    default:
      break;
  }
  long pos = ftell(in->file);
  if ((in->line == in->asect->end_secln) && 
      (pos >= in->asect->end_secpos))
  {
    getsubfmt(ptype, subfmt, 0);
    sprintf(errmsg, "end of section %s was reached - reading cannot be continued\n"
                    "(format '%s', subformat '%s', parameter id=%ld)\n"
            "input file : %s, line=%ld, column=%ld", in->asect->name.alias, fmt, subfmt, idt+1, 
                    in->fname, in->line, in->col);
    print_err(errmsg, errfile, errline, errfunc);
    abort();
  }
  return;
}



/**
  Function cuts string lnstr depending on the end of actually selected section.
  
  @param in - pointer to the structure XFILE with opened text file for reading
  @param lnstr - input string

  @retval Function does not return anything.

  created  2.2008,  Tomas Koudelka, koudelka@cml.fsv.cvut.cz
*/
void cut_str_sec(XFILE *in, char *lnstr)
{
  switch (in->kwdmode)
  {
    case ignore:
    case sequent_mode:
    case line_mode:
      return;
    default:
      break;
  }
  if (in->asect->end_secln == in->line)
    lnstr[in->asect->end_seccol]='\0';
  return;
}



/**
  Function checks opened file in for error generated by scanf function.

  @param in - pointer to the structure XFILE with opened text file for reading
  @param ret - returned value from the scanf function
  @param reqret - required returned value from the scanf function which was excepted
  @param war - warning indicator (0=warning ignored/ 1=warning / 2=warning treated as error)
  @param errmsg  - string for error message
  @param partmsg  - string with part of error message

  Following parameters are used only for error message 
  @param fmt - format string
  @param idt - index of actually processed conversion in fmt
  @param ptype   - pointer to the actual conversion
  @param subfmt  - string for subformat
  @param errmsg  - string for error message
  @param errfile - string with source file name where the error was detected
  @param errline - line number in source file where the error was detected
  @param errfunc - string with function name where the error was detected

  @retval 0 - no error was encountered
  @retval 1 - error was encountered

  created  12.2006,  Tomas Koudelka, koudelka@cml.fsv.cvut.cz
*/
long checkscanferr(XFILE *in, long ret, long reqret, long war, char *errmsg, const char *partmsg,
                   const char *fmt, char *subfmt, long idt, const char *ptype, const char *errfile,
                   int errline, const char *errfunc)
{
  if (ret != reqret) // required return value was not obtained
  {
    getsubfmt(ptype, subfmt, 0);
    sprintf(errmsg, "%s\n(format '%s', subformat '%s', parameter id=%ld)\n"
                    "input file : %s, line=%ld, column=%ld", 
                     partmsg, fmt, subfmt, idt+1, in->fname, in->line, in->col);
    print_err(errmsg, errfile, errline, errfunc);
    return 1;
  }
  if (war && in->warrning) // warning was encountered and reporting of warning is switched on
  {
    getsubfmt(ptype, subfmt, 0);
    if (in->warrning == 2) // treat warning as error
    {   
      sprintf(errmsg, "%s\n(format '%s', subformat '%s', parameter id=%ld)\n"
                      "input file : %s, line=%ld, column=%ld", 
                       partmsg, fmt, subfmt, idt+1, in->fname, in->line, in->col);
      print_err(errmsg, errfile, errline, errfunc);
      return 1;
    }
    if (in->warrning == 1) // only report warning
    {   
      sprintf(errmsg, "\rWarrning : %s\n(format '%s', subformat '%s', parameter id=%ld)\n"
                      "input file : %s, line=%ld, column=%ld", 
                       partmsg, fmt, subfmt, idt+1, in->fname, in->line, in->col);
      print_err(errmsg, errfile, errline, errfunc);
    }
  }
  return 0;
}





/**
  Function scans lnstr for keyword. Required keyword is given by apar string. 

  @param in - pointer to the structure XFILE with opened text file for reading
  @param lnstr - string with line scanned for keyword
  @param aptr  - actual position in lnstr string where should scanning start
  @param apar  - string with required keyword
  @param br    - output parameter in which the read bytes are stored

  @retval 0 - keyword was successfully found
  @retval 1 - keyword was not found
  @retval 2 - multiple keyword detection in lnstr in case line_mode handling
  @retval 3 - invalid keyword is required (zero length of keyword string)

  created  12.2006,  Tomas Koudelka, koudelka@cml.fsv.cvut.cz
*/
long getkwd(XFILE *in,char *lnstr, char *&aptr,char *apar, unsigned int &br)
{
  long l, tl, lkwd;
  char *kwd_beg;
  char *kwd_beg2;
  char *tmpstr;
  switch (in->kwdmode)
  {
    case ignore : // ignore keywords
      br = 0;
      break;
    case sequent_mode : // sequentially scan for keyword
      lkwd = strlen(apar);
      if (lkwd < 1)  // error - zero length of keyword string
        return 3;
      kwd_beg = NULL;
      l = strlen(aptr);
      tmpstr = new char[l+1];
      memset(tmpstr, 0, sizeof(*tmpstr)*l);
      sscanf(aptr, "%s", tmpstr);                // selection of the first substring in lnstr
      tl = strlen(tmpstr);                        // length of scanned substring, it should be equal to length of keyword string
      if (tl == lkwd)
        kwd_beg = strstr(tmpstr, apar, in->ignorecase); // detection of keyword in the first substring
      delete [] tmpstr;
      if (kwd_beg == NULL)
        return 1;
      else
      {
        kwd_beg = strstr(aptr, apar, in->ignorecase); // detection of keyword in the lnstr -> br
        br = kwd_beg-aptr+lkwd;
      }
      break;
    case line_mode :
      lkwd = strlen(apar);
      if (lkwd < 1)  // error - zero length of keyword string
        return 3;
      kwd_beg = strstr(lnstr, apar, in->ignorecase); // detection of keyword
      if (kwd_beg == NULL)
        return 1;
      l = strlen(lnstr);
      tmpstr = new char[l+1];
      memset(tmpstr, 0, sizeof(*tmpstr)*l);
      sscanf(kwd_beg, "%s", tmpstr);                // selection of the first substring in lnstr
      tl = strlen(tmpstr);                        // length of scanned substring, it should be equal to length of keyword string
      if (tl != lkwd)
      {
        delete [] tmpstr;
        return 1;
      }
      // detection of multiple occurence of keyword
      kwd_beg2 = strstr(kwd_beg+lkwd, apar, in->ignorecase);
      if (kwd_beg2) 
      {
        memset(tmpstr, 0, sizeof(*tmpstr)*l);
        sscanf(kwd_beg2, "%s", tmpstr);             // selection of the first substring in lnstr
        tl = strlen(tmpstr);                        // length of scanned substring, it should be equal to length of keyword string
        if (tl != lkwd)
        {
          delete [] tmpstr;
          return 2;
        }
      }
      delete [] tmpstr;
      br = kwd_beg - lnstr + lkwd;
      break;
    case sect_mode_seq:
    case sect_mode_full:
      getkwd_sect(in, lnstr, aptr, apar, 0);
      break;      
    default :
      print_err("Unknown type of keyword handling mode is required\n", __FILE__, __LINE__, __func__);
      abort();
  }
  return 0;
}



/**
  Function scans acutal section of file in for given keyword
  and return number of located keywords. If the lnstr == NULL,
  the function is supposed tobe called outside from the xfscanf,
  otherwise it is supposed to be called by the xfscanf. 
  At the function entry it is supposed :
  - in->line in->col have to be correct actual line and column number
  - in->lnfpos has to be correct file position of the actual line beginning
  - in case lnstr == NULL, file poition(ftell) corresponds to in->line and in->col
  - in case lnstr != NULL, file position(ftell) corresponds to the last character on the actual line before \n
  if called from xfscanf (lnstr != NULL) :
  - content of lnstr is changed so that it contains line with the first keyword occurence
  - aptr is changed so that it points OVER first located keyword
  - in->line, in->col are changed so that it points OVER first located keyword
  - in->lnfpos file position of the actual line beginning
  - position in file is set (fseek) to the end of actual line before \n
  if called externally (lnstr == NULL) :
  - lnstr and aptr are unchanged
  - in->line, in->col are changed so that it points TO THE FIRST character of first located keyword
  - in->lnfpos file position of the actual line beginning
  - position in file is set (fseek) to the end of actual line before \n


  @param in - pointer to the structure XFILE with opened text file for reading
  @param lnstr - string scanned for keyword used in the xfscanf
  @param aptr  - actual position in lnstr string where scanning should start
  @param apar  - string with required keyword
  @param detect_allkwd - flag for detection of all keyword occurences(=1) or 
                         only the firstkeyword occurence (=0).
 
  @retval number of suceffully located keywords

  created  12.2006,  Tomas Koudelka, koudelka@cml.fsv.cvut.cz

*/
long getkwd_sect(XFILE *in,char *lnstr, char *&aptr,char * apar, long detect_allkwd)
{
  switch (in->kwdmode)
  {
    case sect_mode_full:
      // for this mode, searching starts always from the beginning of the section
      xf_resetsec(in); 
      break;
    case sect_mode_seq:
      break;
    default:
      print_err("Unknown type of keyword handling mode is required\n", __FILE__, __LINE__, __func__);
      abort();
  }

  if (in->index_created == 0)
  {
    print_err("Section keyword mode is required but section index is not built\n"
              "Use xfdetect_sect function before calling xfscanf function\n", __FILE__, __LINE__, __func__);
    abort();
  }
  if ((in->id_sec < 0) || (in->id_sec > in->num_sec))
  {
    print_err("Section id is out of range\n", __FILE__, __LINE__, __func__);
    abort();
  }  

  long tnoc = 0;   // total number of keyword occurences
  long noc  = 0;   // number of keyword occurences in one string
  char *kwd_loc = NULL;       // pointer to the first keyword occurence
  size_t kwdl = strlen(apar); // length of keyword
  long l;           // length of astr(line)
  char *loc = NULL; // pointer to the other keyword occurences
  char *astr;       // string with line which is read actually
  long br = 0;      // flag for read bytes;
  long beg_fpos;    // actual position in file before the function starts reading 
  long end_fkwdpos; // file position over the last character of first located keyword
  long fkwd_line;   // line number with the first keyword occurence
  long fkwd_col;    // column number with the first keyword occurence
  long fkwd_lnfpos; // position of the fkwd_line beginning in the file 

  beg_fpos = in->lnfpos + in->col - 1;
  end_fkwdpos = 0;
  astr = new char[in->maxlnsize+1];
  astr[0] = '\0';
  if ((lnstr == NULL) || (in->kwdmode == sect_mode_full))
  // getkwd_sect is called out of xfscanf function or 
  // for sect_mode_full, the first line of the section has to be read
  {
    if (in->line <= in->asect->end_secln)
    {          
      astr[0] = '\0'; // setting string to zero length
      fscanf(in->file, "%[^\n]",astr); //read new line without terminating \n
      br = 1;
      // cut string before end of actual section
      cut_str_sec(in, astr);
      // removing trailing \r form the last character of string (due to DOS \r\n newline)
      l = strlen(astr);
      if (l > 0)
      {
        if (astr[l-1] == '\r')
          astr[l-1] = '\0';
      }
      clearcomments(astr);   // clear possible comments
    }
    else
      return 0;
  }
  else  // getkwd_sect is called from the xfscanf function
  {
    // astr should begin on the actual position on the given line
    strcpy(astr, aptr);
  }
  do 
  {
    if (kwd_loc == NULL) // no keyword was located
      kwd_loc = locate_kwd(astr, apar, in->ignorecase, noc);
    else // at least one occurence of keyword was located and multiple occerences are required to be detected
      loc = locate_kwd(astr, apar, in->ignorecase, noc);

    if (kwd_loc && (tnoc==0)) // the first keyword occurence is detected
    {
      if ((br == 0) && lnstr)
        strcpy(lnstr, astr);
      fkwd_col    = in->col + kwd_loc - astr;
      fkwd_line   = in->line;
      fkwd_lnfpos = in->lnfpos;
      if (lnstr)
      {
        // for internal call from the xfscanf the position over the first 
        // located keyword is required
        fkwd_col += kwdl;
        if (br == 0)
          aptr += kwd_loc - astr + kwdl;
        else
          aptr = lnstr + fkwd_col - 1;
      }
      end_fkwdpos = in->lnfpos + fkwd_col - 1;  // file position of the end of first located keyword 
      if (detect_allkwd == 0)
        break;
      else
        tnoc += noc;
    }
    else  // keyword was not detected or other occurences of keyword are searched
    {
      if (loc)  // other occurence of keyword was detected
        tnoc += noc;
    }
    loc = NULL;
    // reading of new line
    if (in->line < in->asect->end_secln)
    {          
      astr[0] = '\0'; // setting string to zero length
      fscanf(in->file, "%*1[\n]"); //read new line terminating character
      in->line++;
      in->col = 1;
      in->lnfpos = ftell(in->file);
      fscanf(in->file, "%[^\n]",astr); //read new line
      br = 1;
      // cut string before end of actual section
      cut_str_sec(in, astr);
      // removing trailing \r form the last character of string (due to DOS \r\n newline)
      l = strlen(astr);
      if (l > 0)
      {
        if (astr[l-1] == '\r')
          astr[l-1] = '\0';
      }
      clearcomments(astr);   // clear possible comments
    }
    else
      break;

  } while (1); 

  if (kwd_loc)  // setting of file position over the end of first located keyword
  {
    in->line   = fkwd_line;
    in->col    = fkwd_col;
    in->lnfpos = fkwd_lnfpos;
    if (lnstr == NULL) 
      // external call requires setting up file position to the beginning of first located keyword
      fseek(in->file, fkwd_lnfpos+fkwd_col-1, SEEK_SET);
  }
  delete [] astr;
  return tnoc;
}




/**
  Function scans lnstr for enum type. Possible aliases for given enum are stored in the akwdset. 

  @param lnstr - string scanned for keyword
  @param akwdset - pointer to class containing keyword set and set of corresponding integer values which are perimtted for given enum
  @param apar  - pointer to the integer argument to which the detected enum integer value will be assigned
  @param br    - output parameter in which the read bytes are stored
  @param handling - this parameter controls scanning for keyword
                    = 0 - ignore akwdset -> do not scan for aliases keywords
                    != 0 - scan either for possible keywords or for the integer id
  @param ignorecase - flag for case insensitivity (=1)/ case sensitivity (=0)

  @retval 1 - enum was successfully scanned
  @retval 2 - cannot read integer value of enum
  @retval 3 - scanned integer value of enum was not found in akwdset
  @retval 4 - nor enum keyword nor integer value of enum was read

  created  12.2006,  Tomas Koudelka, koudelka@cml.fsv.cvut.cz
*/
long getenum(const char *lnstr, kwdset *akwdset, int *apar, unsigned int &br, int handling, long ignorecase)
{
  long l, n, tbr;
  char *ekwd;
  long ret = 1;
  long war;

  switch (handling)
  {
    case ignore :
      // try to read int value of enum 
      war = checkint(lnstr);
      ret = sscanf(lnstr, "%d%n",apar,&br);
      if (ret != 1)
        ret = 2; // cannot read integer value of enum
      else
      {
        if (akwdset)
        {
          if (matchekwdint(apar, akwdset))
            ret = 1; // O.K. integer value was found and checked
          else
            ret = 3; // enum integer value was not found in kwdset
        }
      }
      break;
    default :
      // For all cases of keyword handling, the enum keyword have to be read by the sequential mode
      if (akwdset == NULL)
        return 5; // the keyword set was not passed
      l = strlen(lnstr);
      ekwd = new char[l+1];
      n = sscanf(lnstr, "%s%n", ekwd, &br);
      l = strlen(ekwd);
      tbr = br-l; // number of whitespaces suppressed in above sscanf %s
      if (n != 1)
      {
        ret=4; // cannot read enum
      }
      else
      {
        if (matchekwd(ekwd, akwdset, apar, ignorecase))
          ret = 1; // O.K. kwd was found and checked
        else
        {
          // try to read int value of enum 
          war = checkint(ekwd);
          ret = sscanf(ekwd, "%d%n",apar,&br);
          br += tbr;
          if (ret != 1)
            ret = 4; // nor enum keyword nor integer value of enum was read
          else
          {
            if (matchekwdint(apar, akwdset))
              ret = 1; // O.K. integer value was found and checked
            else
              ret = 3; // enum integer value was not found in kwdset
          }
        }
      }
      delete [] ekwd;
  }
  return ret;
}




/**
  Function searches alieses keyword in the string. (used for enum conversion)
  
  @param ekwd - string with required keyword 
  @param akwdset - pointer to class which contains set of aliases keywords and corresponding integer values
  @param apar - output parameter (pointer to integer) to which the integer value of detected enum keyword will be assigned
  @param ignorecase - flag for case insensitivity (=1)/ case sensitivity (=0)

  @retval 0 - enum keyword was not found
  @retval 1 - enum keyword was successfully scanned

  created  12.2006,  Tomas Koudelka, koudelka@cml.fsv.cvut.cz
*/
long matchekwd(const char *ekwd, kwdset *akwdset, int *apar, long ignorecase)
{
  long i, lkwd, tl, l;
  const char *kwd_beg; 
  long ret=0;
  long pos_ok;
 
  l = strlen(ekwd);
  for(i=0; i<akwdset->n; i++)
  {
    kwd_beg = strstr(ekwd, akwdset->set[i].alias, ignorecase);
    if (kwd_beg)
    {
      tl   = strlen(kwd_beg);
      lkwd = strlen(akwdset->set[i].alias);
      
      // check position of found keyword string 
      // (corectly found keyword cannot be part of another nonwhitespace part of searched string)
      pos_ok = 0;
      if (kwd_beg != ekwd)
      {
        if (isspace(*(kwd_beg-1))) // only space can precede of the found keyword 
          pos_ok = 1;
      }
      else
        pos_ok = 1;

      if ((tl == lkwd) && pos_ok)
      {
        *apar = akwdset->set[i].id;
        ret   = 1;
        break;
      }
    }
  }
  return ret;
}



/**
  Function searches alieses integer in the string. (used for enum conversion)
  
  @param apar - output parameter (pointer to integer) to which the integer value of detected enum keyword will be assigned
  @param akwdset - pointer to class which contains set of aliases keywords and corresponding integer values

  @retval 0 - enum integer was not found
  @retval 1 - enum integer was successfully scanned

  created  12.2006,  Tomas Koudelka, koudelka@cml.fsv.cvut.cz
*/
long matchekwdint(int *apar, kwdset *akwdset)
{
  long i;
  long ret=0;
  for(i=0; i<akwdset->n; i++)
  {
    if (*apar == akwdset->set[i].id)
    {
      ret = 1;
      break;
    }
  }
  return ret;
}



/**
  Function checks for errors during enum conversion.
  

  @param in - pointer to the structure XFILE with opened text file for reading
  @param ret - returned value from the scanf function
  @param reqret - required returned value from the scanf function which was excepted
  @param war - warning indicator (0=warning ignored/ 1=warning / 2=warning treated as error)
  @param akwdset - pointer to class which contains set of aliases keywords and corresponding integer values
  @param errmsg  - string for error message

  Following parameters are used only for error message 
  @param fmt - format string
  @param idt - index of actually processed conversion in fmt
  @param ptype   - pointer to the actual conversion
  @param subfmt  - string for subformat
  @param errmsg  - string for error message
  @param errfile - string with source file name where the error was detected
  @param errline - line number in source file where the error was detected
  @param errfunc - string with function name where the error was detected

  @retval 0 - enum integer was not found
  @retval 1 - enum integer was successfully scanned

  created  12.2006,  Tomas Koudelka, koudelka@cml.fsv.cvut.cz
*/
long checkenumerr(XFILE *in, long ret, long reqret, long war, kwdset *akwdset, char *errmsg,
                  const char *fmt, char *subfmt, long idt, const char *ptype, const char *errfile,
                  int errline, const char *errfunc)
{
  long i;
  char *partmsg;

  if (ret != reqret)
  {
    switch (ret)
    {
      case 2:
        partmsg = "cannot read integer id of enum";
        break;
      case 3:
        partmsg = "enum integer id was not found in keyword set";
        break;
      case 4:
        partmsg = "cannot read enum keyword nor integer id";
        break;
      case 5:
        partmsg = "keyword set was not specified in source code file";
        break;
      default :
        partmsg = "cannot read enum";
    }
    getsubfmt(ptype, subfmt, 0);
    sprintf(errmsg, "%s\n(format '%s', subformat '%s', parameter id=%ld)\n"
                    "input file : %s, line=%ld, column=%ld", 
                     partmsg, fmt, subfmt, idt+1, in->fname, in->line, in->col);
    print_err(errmsg, errfile, errline, errfunc);
    if (akwdset == NULL)
      fprintf(stderr, "No keyword set was specified\n");
    else
    {
      fprintf(stderr, "Possible values for this enum are :\n");       
      for (i=0; i < akwdset->n; i++)
        fprintf(stderr, "%s (%d)\n", akwdset->set[i].alias, akwdset->set[i].id);
    }
    return 1;
  }
  if (war && in->warrning)
  {
    getsubfmt(ptype, subfmt, 0);
    if (in->warrning == 2)
    {   
      sprintf(errmsg, "unable to integer\n(format '%s', subformat '%s', parameter id=%ld)\n"
                      "input file : %s, line=%ld, column=%ld", 
                       fmt, subfmt, idt+1, in->fname, in->line, in->col);
      print_err(errmsg, errfile, errline, errfunc);
      return 1;
    }
    if (in->warrning == 1)
    {   
      sprintf(errmsg, "\rWarrning : unable to integer\n(format '%s', subformat '%s', parameter id=%ld)\n"
                      "input file : %s, line=%ld, column=%ld", 
                       fmt, subfmt, idt+1, in->fname, in->line, in->col);
      print_err(errmsg, errfile, errline, errfunc);
    }
  }
  return 0;
}

/**
  Function performs case sensistive/insensitive searching of substring 'needle' in the string 'haystack', 
  The terminating '\0' characters are not compared. The case sensitivity is controlled by the parameter
  ignorecase : 0 = case sensitive / 1 = case insensitive.

  @param haystack - searched in string
  @param needle - searched for string
  @param ignorecase - flag for case insensitivity (=1)/ case sensitivity (=0)

  @retval The function returns a pointer to the begining of the substring, or NULL if the substring is not found
*/
char *strstr(char *haystack, const char *needle, long ignorecase)
{
  char *ret;
  if (ignorecase==1)
    ret = strstrcis(haystack, needle);
  else
    ret = strstr(haystack, needle);
  return ret;
}



/**
  Function performs case sensistive/insensitive searching of substring 'needle' in the string 'haystack', 
  The terminating '\0' characters are not compared. The case sensitivity is controlled by the parameter
  ignorecase : 0 = case sensitive / 1 = case insensitive.

  @param haystack - searched in string literal
  @param needle - searched for string
  @param ignorecase - flag for case insensitivity (=1)/ case sensitivity (=0)

  @retval The function returns a pointer to the begining of the substring, or NULL if the substring is not found
*/
const char *strstr(const char *haystack, const char *needle, long ignorecase)
{
  const char *ret;
  if (ignorecase==1)
    ret = strstrcis(haystack, needle);
  else
    ret = strstr(haystack, needle);
  return ret;
}



/**
  Function performs case insensitive searching of substring 'needle' in the string 'haystack', The terminating
  '\0' characters are not compared.

  @param haystack - searched in string
  @param needle - searched for string

  @retval The function returns a pointer to the begning of the substring, or NULL if the substring is not found
*/
char *strstrcis(char *haystack, const char *needle)
{
  long i, j, lh, ln;
  lh = strlen(haystack);
  ln = strlen(needle);
  j = 0;

  if ((ln == 0) || (needle == NULL) || (haystack == NULL))
    return NULL;

  for (i = 0; i<lh; i++)
  {
    if (tolower(haystack[i]) == tolower(needle[j]))
    {
      j++;
      if (j == ln)
        return (char *)(haystack+i-(j-1));
    }
    else
      j = 0;  
  }

  return NULL;
}



/**
  Function performs case insensitive searching of substring 'needle' in the string 'haystack', The terminating
  '\0' characters are not compared.

  @param haystack - searched in string literal
  @param needle - searched for string

  @retval The function returns a pointer to the begning of the substring, or NULL if the substring is not found
*/
const char *strstrcis(const char *haystack, const char *needle)
{
  long i, j, lh, ln;
  lh = strlen(haystack);
  ln = strlen(needle);
  j = 0;

  if ((ln == 0) || (needle == NULL) || (haystack == NULL))
    return NULL;

  for (i = 0; i<lh; i++)
  {
    if (tolower(haystack[i]) == tolower(needle[j]))
    {
      j++;
      if (j == ln)
        return (haystack+i-(j-1));
    }
    else
      j = 0;  
  }

  return NULL;
}
