/*
    File:                    tablefunct.cpp
    Author:                  Martin Kulhavy
    Purpose:                 Class computes function value at exact point, where the function is given by the table x,y
                             Implemented interpolations: dirac <), lagrange, piecewiselinear 
    Revision history:        Martin Kulhavy 20.11.2001 - file was created
*/

#include <stdlib.h>
#include "tablefunct.h"
#include "intools.h"


/**
   The constructor defines arrays for interpolation.
   Parameters:
   @param asize is the size of vectors x, y
   @param x is the vector where time values are stored
   @param y is the vector where the function values are stored
 
   created  20.11.2001 by Martin Kulhavy, mkulhavy@cml.fsv.cvut.cz
*/  
tablefunct::tablefunct()
{
  asize=0; x=NULL; y=NULL;
}


/**
   The destructor deallocates the memory where the vectors x, y are stored
   Parameters:

   created  20.11.2001 by Martin Kulhavy, mkulhavy@cml.fsv.cvut.cz
*/
tablefunct::~tablefunct()
{
  delete[] x; delete[] y;
}


/**
   The function returns the interpolated value at exact point (temp) by Dirac interpolation
   Parameters:
   @param temp is x-coordinate, where the function value will be found
   
   created  20.11.2001 by Martin Kulhavy, mkulhavy@cml.fsv.cvut.cz
*/
double 
tablefunct::diracinterpol(double temp)
{
  //this function is priceless for instance, for assembling load value at current time etc.
 
  long i;
  double iv;
  if (temp<x[0] || temp>x[asize-1]){
    fprintf(stderr,"\ntablefunct::diracinterpol\nValue is out of range (%s, line %d)",__FILE__,__LINE__);
    exit(2);
  }
  
  for (i=0;i<asize;i++){
    if (i==asize-1){
      iv=y[i];
      break;
    }
    else if (temp>x[i] && temp<=x[i+1]){
      iv=y[i+1];
      break;
    }
  }
  return iv;
}


/**
   The function returns the interpolated value at exact point (temp) by Dirac interpolation,
   if the point lies out of the range, extrapolation from the limit intervals takes effect.
   Parameters:
   @param temp is x-coordinate, where the function value will be found
   
   created  8.8.2005 by Tomas Koudelka, koudelka@cml.fsv.cvut.cz
*/
double 
tablefunct::diracinterpol2(double temp, double &k)
{
  long i;
  double iv;
  for (i=0;i<asize;i++){
    if (i==asize-1){
      iv=y[i];
      break;
    }
    if ((i==0) && (x[i] >= temp))
    {
      iv=y[i];
      break;
    }
    if (temp>x[i] && temp<=x[i+1]){
      iv=y[i+1];
      break;
    }
  }
  k = 0.0;
  return iv;
}


/**
   The function returns the interpolated value at exact point (temp) by linear interpolation
   Parameters:
   @param temp is x-coordinate, where the function value will be found
   
   created  20.11.2001 by Martin Kulhavy, mkulhavy@cml.fsv.cvut.cz
*/
double tablefunct::lininterpol (double temp)
{
  double k;
  return lininterpol(temp, k);
}

/**
   The function returns the interpolated value at exact point (temp) by linear interpolation
   Parameters:
   @param temp is x-coordinate, where the function value will be found
   @param k is derivative of linear function
   
   created  20.11.2001 by Martin Kulhavy, mkulhavy@cml.fsv.cvut.cz
*/
double tablefunct::lininterpol (double temp, double &k)
{
  //this function is priceless for instance for assembling load value at current time etc.
  
  double q,iv;

  int i,h=0;
  
  for (i=0;i<asize-1;i++){
    if (temp>x[i] && temp<=x[i+1]){
      k =(y[i+1]-y[i])/(x[i+1]-x[i]);
      q =y[i]-k*x[i];
      iv=k*temp+q;
      h=1;
      break;
    }
  }
  if(h==0){
    fprintf (stderr,"\ntablefunct::lininterpol\nValue is out of range (%s, line %d)",__FILE__,__LINE__);
    exit(2);
  }
  
  return iv;
}




/**
   The function returns the interpolated value at exact point (temp) by linear interpolation,
   if the point lies out of the range, extrapolation from the limit intervals takes effect.
   Parameters:
   @param temp is x-coordinate, where the function value will be found
   @param k is derivative of linear function
   
   created  8.8.2005 by Tomas Koudelka, koudelka@cml.fsv.cvut.cz
*/
double tablefunct::lininterpol2 (double temp, double &k)
{
  //this function is priceless for instance for assembling load value at current time etc.
  
  double q,iv;

  int i,h=0;
  
  for (i=0;i<asize;i++){
    if ((x[i] >= temp) || (i == asize-1)){
      if (i == 0)
      {
        k =(y[i+1]-y[i])/(x[i+1]-x[i]);
        q =y[i]-k*x[i];
        iv=k*temp+q;
        h=1;
        break;
      }     
      k =(y[i]-y[i-1])/(x[i]-x[i-1]);
      q =y[i]-k*x[i];
      iv=k*temp+q;
      h=1;
      break;
    }
  }
  if(h==0){
    fprintf (stderr,"\ntablefunct::lininterpol2\nValue is out of range (%s, line %d)",__FILE__,__LINE__);
    exit(2);
  }
  
  return iv;
}



/**
   The function returns the interpolated value at exact point (temp) by linear interpolation of inverse function!!!,
   if the point lies out of the range, extrapolation from the limit intervals takes effect.
   Parameters:
   @param temp is x-coordinate, where the function value will be found
   @param k is derivative of linear function
   
   modified 11.2.2007 by Tomas Krejci, krejci@cml.fsv.cvut.cz
*/
double tablefunct::lininterpol3 (double temp, double &k)
{
  //this function is priceless for instance for assembling load value at current time etc.
  
  double q,iv;

  int i,h=0;
  
  for (i=0;i<asize;i++){
    if ((y[i] >= temp) || (i == asize-1)){
      if (i == 0)
      {
        k =(x[i+1]-x[i])/(y[i+1]-y[i]);
        q =x[i]-k*y[i];
        iv=k*temp+q;
        h=1;
        break;
      }     
      k =(x[i]-x[i-1])/(y[i]-y[i-1]);
      q =x[i]-k*y[i];
      iv=k*temp+q;
      h=1;
      break;
    }
  }
  if(h==0){
    fprintf (stderr,"\ntablefunct::lininterpol3\nValue is out of range (%s, line %d)",__FILE__,__LINE__);
    exit(2);
  }
  
  return iv;
}



/**
   The function returns the interpolated value at exact point (temp) by Lagrangian n-1 degree polynom
   Parameters:
   @param temp is x-coordinate, where the function value will be found
   
   created  20.11.2001 by Martin Kulhavy, mkulhavy@cml.fsv.cvut.cz
*/
double
tablefunct::laginterpol(double temp)
{
  double y1=0.,nom,denom;
  long i,j;
 
  for(i=0; i<asize; i++){
    nom=y[i];
    denom=1.;
    for(j=0; j<asize; j++){
      if(i==j) continue;
      nom=nom*(temp-x[j]);
      denom=denom*(x[i]-x[j]);
      if(denom<1.0e-8) continue; // skip subsequent multipile data for identical xx
    }
    y1=y1+nom/denom;
  }
 
  return y1;
}
      


/**
   The function returns the interpolated value at exact point (temp), interpolation is chosen according to 
   variable itype. Dirac, linear and Lagrangian n-1 interpolations are supported, see tablefunct.h  
   Parameters:
   @param temp is x-coordinate, where the function value will be found
   
   created  20.11.2001 by Martin Kulhavy, mkulhavy@cml.fsv.cvut.cz
*/
double tablefunct::getval(double temp)
{
    
  double iv;
  if (asize<=1 || x==NULL || y==NULL){
    fprintf (stderr,"\ntablefunct::ivalue\nWrong size of the array (%s, line %d)",__FILE__,__LINE__);
    exit(2);
  }
  
  switch(itype){
  case piecewiselin:{
    iv=lininterpol(temp);
    break;
  }
  case dirac:{
    iv=diracinterpol(temp);
    break;
  }
  case lagrange:{
    iv=laginterpol(temp);
    break;
  }
  case piecewiselin2:{//added by TKr 2.11.2007
    iv=lininterpol(temp);
    break;
  }
  default:{
    fprintf (stderr,"\ntablefunct::ivalue\nUnknown type of interpolation (%s, line %d)",__FILE__,__LINE__);
    exit(2);
  }
  }   
  
  return iv;
}




/**
   The function returns the interpolated value at exact point (temp), interpolation is chosen according to 
   variable itype. Dirac, linear and Lagrangian n-1 interpolations are supported, see tablefunct.h  
   Parameters:
   @param temp is x-coordinate, where the function value will be found
   @param k is derivative of the function at given interval
   
   created  05.2007 by Tomas Koudelka koudelka@cml.fsv.cvut.cz
*/
double tablefunct::getval2(double temp, double &k)
{
    
  double iv;
  if (asize<=1 || x==NULL || y==NULL){
    fprintf (stderr,"\ntablefunct::getval2\nWrong size of the array (%s, line %d)",__FILE__,__LINE__);
    exit(2);
  }
  
  switch(itype){
  case piecewiselin2:{
    iv=lininterpol2(temp, k);
    break;
  }
  case dirac2:{
    iv=diracinterpol2(temp, k);
    break;
  }
  default:{
    fprintf (stderr,"\ntablefunct::getval2\nUnknown type of interpolation (%s, line %d)",__FILE__,__LINE__);
    exit(2);
  }
  }   
  
  return iv;
}



/**
   The function returns the interpolated value at exact point (temp) of inverse function!!!, interpolation is chosen according to 
   variable itype. Dirac, linear and Lagrangian n-1 interpolations are supported, see tablefunct.h  
   Parameters:
   @param temp is x-coordinate, where the function value will be found
   @param k is derivative of the function at given interval
   
   modified  11.2007 by Tomas Krejci krejci@cml.fsv.cvut.cz
*/
double tablefunct::getval3(double temp, double &k)
{
    
  double iv;
  if (asize<=1 || x==NULL || y==NULL){
    fprintf (stderr,"\ntablefunct::getval2\nWrong size of the array (%s, line %d)",__FILE__,__LINE__);
    exit(2);
  }
  
  switch(itype){
  case piecewiselin2:{
    iv=lininterpol3(temp, k);
    break;
  }
  default:{
    fprintf (stderr,"\ntablefunct::getval2\nUnknown type of interpolation (%s, line %d)",__FILE__,__LINE__);
    exit(2);
  }
  }   
  
  return iv;
}




/**
   The function allocates and reads (stores) input parameters (asize, x, y) for interpolation from the file 
   Example: 1 4 0.0 100.0 200.0 300.0 2.4   1.6   8.9  10.0 = itype asize x[0] y[0] x[1] y[1] x[2] y[2] x[3] y[3]
   Parameters: 
   @param in is input file
   
   created  20.11.2001 by Martin Kulhavy, mkulhavy@cml.fsv.cvut.cz
*/
void tablefunct::read(XFILE *in)
{
  long i;
  xfscanf (in,"%k%m%k%ld","approx_type", &interpoltype_kwdset, (int*)&itype, "ntab_items", &asize);
  x=new double[asize];
  y=new double[asize];
  for (i=0;i<asize;i++){
    xfscanf (in,"%lf",&x[i]);
    xfscanf (in,"%lf",&y[i]);
  }
}

/**
   The function writes input parameters (asize, x, y) for interpolation

   Parameters: 
   @param out - output file
   
   created  22.11.2002 by JK
*/
void tablefunct::print(FILE *out)
{
  long i;
  fprintf (out,"\n %d %ld",itype,asize);
  for (i=0;i<asize;i++){
    fprintf (out,"\n %le ",x[i]);
    fprintf (out,"%le ",y[i]);
  }
}

/**
   The function allocates and reads (stores) input parameters (asize, x, y) for interpolation from the file
   Example: 1 4 0.0 100.0 200.0 300.0 2.4   1.6   8.9  10.0 = itype asize x[0] y[0] x[1] y[1] x[2] y[2] x[3] y[3]
   Parameters:
   @param in is input file

   created  20.11.2002 by Tomas Koudelka, koudelka@cml.fsv.cvut.cz
*/
void tablefunct::read_prop(FILE *in)
{
  long i;
  getint(in, (int&)itype);
  getlong(in, asize);
  x=new double[asize];
  y=new double[asize];
  for (i=0;i<asize;i++)
  {
    getdouble (in, x[i]);
    getdouble (in, y[i]);
  }
}


/**
   function copies all data from subroutine argument to the actual object
   
   JK, 15.11.2008
*/
void tablefunct::copy (tablefunct &tf)
{
  long i;
  
  //  type of interpolation
  itype = tf.itype;
  //  number of function values
  asize = tf.asize;
  
  if (x!=NULL)
    delete [] x;
  x = new double [asize];
  
  if (y!=NULL)
    delete [] y;
  y = new double [asize];

  for (i=0;i<asize;i++){
    x[i]=tf.x[i];
    y[i]=tf.y[i];
  }
}
